#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

//AID RF BOX
#include "front_end_ctrl.h" 

#include <gr_file_sink.h>
#include <gr_file_source.h>
#include <gr_firdes.h>
#include <ar_fractional_interpolator_cc.h>
#include <ar_fir_filter_ccf.h>
#include <ar_throughput.h>
#include <ar_mixer_cc.h>
#include <ar_complex_to_xxx.h>
#include <gr_binary_slicer_fb.h>
#include <gr_agc_cc.h>
#include <gr_multiply_vcc.h>
#include <gr_stream_to_vector.h>
#include <gr_fft_vcc.h>
#include <gr_complex_to_xxx.h>
#include <gr_delay.h>
#include <ar_integrate_dump_ff.h>

#include "gnuradio.h"
#include "gnuradio_misc.h"
#include "freq_acq.h"
#include "ar_freq_acq.h"

// Esteban
// USRP
#include <usrp1_source_c.h>
using namespace std;

freq_acq::freq_acq()
{
}

freq_acq::freq_acq(const freq_acq_config &config, bool no_hardware)
{
    setup(config, no_hardware);
}

void freq_acq::setup(const freq_acq_config &config, bool no_hardware) 
{
    d_config = config;
    // Check if the blocks and/or the scheduler are already setup, and if
    // so, stop and clear them.
    if (d_scheduler)
    {
        printf("freq_acq::Setup:Scheduler already exists\n");
        stop();
    }

    d_blocks.clear();

    // USRP
   usrp1_source_c_sptr hardware_source;
   gr_file_source_sptr recorded_data_source;
   
    //Use Recorded Data.
    if(config.rec_data) 
    {
         recorded_data_source = gr_make_file_source(sizeof(gr_complex), config.filename,config.repeat); 
    }  
    //Configure L band board and USRP For Freq Acq. 
    else
    {
         hardware_source = 
              usrp1_make_source_c (    config.usrp_which_board, config.usrp_decim_rate,
              config.usrp_nchan, config.usrp_mux,config.usrp_mode, 
              0,0,"","");
         // Connect all the inputs to appropriate upstream buffers.
         // The syntax is self explanatory: 
         // from_block.from_port -> to_block.to_port, list of blocks	
         //usrp_source->set_rx_freq (0,fc+foff);
         hardware_source->set_pga (config.usrp_which_board,1);	
         hardware_source->set_rx_freq (0,config.usrp_fc);


         //======================================================================
         // AID PARAMETERS   12/03/2008
         /*
         Gain choices are as follows:
         LNA(high) MIXER(low)  -> GAIN = 01; 12dB
         LNA(low)  MIXER(high) -> GAIN = 02; 13dB
         LNA(high) MIXER(high) -> GAIN = 03; 25dB
         LNA(low)  MIXER(low)  -> GAIN = 04; 00dB
         */
         set_frend_gain( hardware_source, config.usrp_which_board, 4 );
         //DFS latches
         /*
         freequency calculation of the DFS
         ADF 4360-3 (refer to data sheet for more information
         fout = [(PxB)+A]xfref/R;
         where:
         P - modulus prescaler (0,1,2) -> 8, 16 and 32 
         B - divide ratio 13 bit (3-8191)
         A - divide ratio 5  bit (0-31)
         R - divide ratio 14 bit (1-16383)
         fref = 27Mhz
         */
         // These options are hard-wired and do not matter. Any value is fine
         //set_frend_dfs(hardware_source, 3, 0, 1, 8 );
         
         //EV: dummy variable for the IF version
         set_frend_dfs(hardware_source, 1837.4e6 );
    }
   

    //======================================================================

	
	//----------------------Downconverter / mixer
	gr_mixer_cc_sptr mixer = 
		gr_make_mixer_cc (config.Fs, -config.Fc);
		
     mixer->set_name("freq_acq::mixer");

	//---------------------- throughput measurement ---------------------
	gr_throughput_sptr throughput =
		gr_make_throughput(sizeof(gr_complex), 2, "cout", "FreqAcq");

         // Decimation 
    // The input sample rate can be reduced for more efficient computation if
    // the frequency uncertainty is small.  Per Nyquist, we must sample at 2x
    // the maximum frequency.  The signal is complex, covering (-Fs/2..+Fs/2).
    // The squaring operation doubles all the frequencies, halving the 
    // coverage range.  For safety, add a margin of 1.5x to avoid edge effects
    int decim_ratio = int(config.Fs / 2.0 / config.freq_range / 1.5);
     //printf("decim ration = %d\n", decim_ratio);
    assert (decim_ratio > 0 && decim_ratio < 64); // Otherwise need higher Fs

    //decim_ratio =1; //JR Debug; 

    gr_block_sptr decim;
    if (decim_ratio > 1)
    {
        std::vector<float> decim_taps = 
            gr_firdes::low_pass(1.0, 1.0, 0.5 / decim_ratio, 0.1);

   	    decim = 
            gr_make_fir_filter_ccf (decim_ratio, decim_taps);
    } else {
        decim = 
            gr_make_delay (sizeof(gr_complex), 0);  
    }

    // Squaring
    gr_multiply_vcc_sptr           square =         gr_make_multiply_vcc (1);
    gr_multiply_vcc_sptr           square2 =        gr_make_multiply_vcc (1);

    gr_delay_sptr decim_delay =  gr_make_delay (sizeof(gr_complex), config.delay);  

    // Serial to parallel
    // Compute the size of the FFT.  The FFT size is dictated by the maximum 
    // acceptable frequency uncertainty.  The uncertainty is equal to the FFT
    // bin size.  FFT bin size = sampling frequency / fft_size.
    float fft_size_f = (config.Fs / decim_ratio) / config.resolution;
    // FFT is computed most efficiently for a power-of-2 size.  Scale to the
    // next power of 2.
    int fft_size = (int) pow(2.0, (int)(log(fft_size_f)/log(2.0)) + 1);
    printf("freq_acq::fft_size = %d\n", fft_size);
    // Save the bin frequency resolution to be used later.  
    d_fft_bin_size = (config.Fs / decim_ratio) / fft_size;

    // The additional factor of 2.0 is due to the squaring operation, which 
    // doubles all the tone frequencies.
    if (config.fourth_law) 
         d_fft_bin_size /= 4.0;
    else if(config.square_law)
         d_fft_bin_size /= 2.0;
    
     //Design a smoother for 1st order Frequency Acquisition Engine [Doesn't yet work]   
  /*std::vector<float> smoother_taps;  //Smoother used when single order freq acq is used.
    int filter_length = (int)(2*config.bandwidth/d_fft_bin_size);  //Design a filter that is 2x as wide as the carrier
    smoother_taps.resize(filter_length, 1); //ones(1,filter_length); 
    gr_fir_filter_ccf_sptr smoother = gr_make_fir_filter_ccf (1, smoother_taps);  //Design smoother block 
     gr_vector_to_stream_sptr par2ser = gr_make_vector_to_stream (sizeof(gr_complex), fft_size);*/

    gr_stream_to_vector_sptr ser2par =
        gr_make_stream_to_vector (sizeof(gr_complex), fft_size);
    
    // FFT
    std::vector<float> fft_window(fft_size, 1.0);
    gr_fft_vcc_sptr fft = 
        gr_make_fft_vcc (fft_size, true, fft_window, true);

    // Take the magnitude squared
    gr_complex_to_mag_squared_sptr           mag_squared=
        gr_make_complex_to_mag_squared (fft_size);

    // Acumulate
    gr_integrate_dump_ff_sptr                acc_fft = 
        gr_make_integrate_dump_ff (config.num_fft_avg, fft_size);

    // Acquisition block
    gr_freq_acq_sptr                         acq = 
        gr_make_freq_acq(fft_size);
    acq->set_name("freq_acq::acq");

	//---------------------- observation points ------------------
    /*
	gr_file_sink_sptr mixer_sink = 
		gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/freq_acq_mixer.dat");
	gr_file_sink_sptr decim_sink = 
		gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/freq_acq_decim.dat");
	gr_file_sink_sptr square_sink = 
		gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/freq_acq_square.dat");
	gr_file_sink_sptr fft_sink = 
		gr_make_file_sink (sizeof(gr_complex)*fft_size, "c:/temp/usrp_log/freq_acq_fft.dat");
	gr_file_sink_sptr acc_sink = 
		gr_make_file_sink (sizeof(float)*fft_size, "c:/temp/usrp_log/freq_acq_acc.dat");
	gr_file_sink_sptr mag_sink = 
		gr_make_file_sink (sizeof(float)*fft_size, "c:/temp/usrp_log/freq_acq_mag.dat");
      */

	// **********************************************************************
	// ************** Connect the blocks together ***************************
	// **********************************************************************
	// The syntax is self explanatory: from_block.from_port -> to_block.to_port

    // Simpler debugging without hardware. Higher level can connect directly
    // to front_end()->get_input_block()

    if (config.rec_data)  
    {
          connect_blocks(recorded_data_source, 0, mixer, 0, d_blocks); 
    }
    else if (!no_hardware){  
         connect_blocks(hardware_source, 0, mixer, 0, d_blocks);
	}

    connect_blocks(mixer,       0, decim,       0, d_blocks);
   // connect_blocks(mixer,       0, throughput,  0, d_blocks);
    int square_law = 0; 
    if(config.square_law || config.fourth_law) {
        //Eliminate Configuration Error. 4th law IMPLIES 2nd order law!
         square_law = 1; 
    }
    if (square_law)
    {
        connect_blocks(decim, 0, decim_delay, 0, d_blocks);     
        connect_blocks(decim,   0, square,      0, d_blocks);
        connect_blocks(decim_delay,   0, square,      1, d_blocks);
        if (config.fourth_law){
               connect_blocks(square,   0, square2,    0, d_blocks);
               connect_blocks(square,   0, square2,    1, d_blocks);
               connect_blocks(square2,  0, ser2par,    0, d_blocks);
        }else{
               connect_blocks(square,  0, ser2par,    0, d_blocks);
        }
    }
    else{  //First Order Law Doesn't work. Can't smooth signal yet
         connect_blocks(decim, 0, ser2par, 0, d_blocks);
    }

    connect_blocks(ser2par,     0, fft,         0, d_blocks);
    connect_blocks(fft,         0, mag_squared, 0, d_blocks);
    connect_blocks(mag_squared, 0, acc_fft,     0, d_blocks);
    connect_blocks(acc_fft,     0, acq,         0, d_blocks);

}

bool freq_acq::get_freqs(int num_peaks, vector<freq_acq_result> &results)
{
    gr_freq_acq *acq = 
        find_block_ptr<gr_freq_acq>(d_blocks, "freq_acq::acq");

    vector<float> freqs,powers;

    bool result_ready = acq->get_freqs(num_peaks, freqs, powers);
    if (!result_ready) return false;

    results.resize(num_peaks);
    for (int i=0; i<num_peaks; ++i) 
    {
        //printf(" Freq_acq_bfsk.cpp     %i BinPeak= %g \tBinSize=%4.3f \n",i,freqs[i] ,d_fft_bin_size);
        results[i].relative_power = powers[i];
        results[i].freq = (freqs[i] - acq->get_fft_size()/2) * d_fft_bin_size;
    }

    return true;
}



bool freq_acq::get_peak_freq(freq_acq_result &result)
{
    gr_freq_acq *acq = 
        find_block_ptr<gr_freq_acq>(d_blocks, "freq_acq::acq");

    float freq, power;
    bool result_ready = acq->get_peak_freq(&freq, &power);
    if (!result_ready) return false;

    result.freq = (freq - acq->get_fft_size()/2) * d_fft_bin_size;
    result.freq += d_config.Fc; // Don't forget to add the original offset
    result.relative_power = power;

    return true; 
}


void freq_acq::run()
{
    if (d_scheduler)
    {
        printf("freq_acq: Run:Scheduler already running");
        assert(0);
    }

   	d_scheduler = start_scheduler_thread(d_blocks, true);
}

void freq_acq::stop()
{
    d_scheduler->stop();
    while (d_scheduler->running()) ; // Wait for it to quit
    d_scheduler = gr_single_threaded_scheduler_sptr(); // Assign to NULL
    //Added EV
    d_blocks.clear();
// printf("d_scheduler stopped\n");

}

bool freq_acq::running()
{
    if (d_scheduler)
        return d_scheduler->running();
    else
        return false;
}


#define CARRIER_WIDTH 4
/*bool freq_acq::get_crosscorr_freq(bpsk_config &config, freq_acq_result &result)
{
    gr_freq_acq *acq = 
        find_block_ptr<gr_freq_acq>(d_blocks, "freq_acq::acq");

    // Build the frequency mask.  For now, set all bins where we expect some
    // energy to 1.
    int fft_size = acq->get_fft_size();
    vector<float> mask(fft_size);
    // center tone
    for (int i=fft_size/2-CARRIER_WIDTH/2; i<fft_size/2+CARRIER_WIDTH/2; i++)
        mask[i] = 1.0;
    // BPSK humps.  Could be on the positive side, negative side, or both
    // Compute how many FFT bins the BPSK signal occupies, and add 25% margin
    int bpsk_width = config.bit_rate / d_fft_bin_size * 1.25;
    // Compute the BPSK center frequency in bins
    int bpsk_center = config.Fc / d_fft_bin_size;
    // BPSK Fc is negative
    if (config.Fc < 0 || config.double_sided)
        for (int i=fft_size/2-bpsk_center-bpsk_width/2; 
                 i<fft_size/2-bpsk_center+bpsk_width/2; i++) mask[i] = 1.0;
    // BPSK Fc is positive
    if (config.Fc > 0 || config.double_sided)
        for (int i=fft_size/2+bpsk_center-bpsk_width/2; 
                 i<fft_size/2+bpsk_center+bpsk_width/2; i++) mask[i] = 1.0;

    // Get the results
    vector<float> fft(fft_size);
    float max_power, mean_power;
    int max_idx;
    bool result_ready = acq->get_crosscorr_freq(mask, &max_idx, &max_power, 
                                               &mean_power, fft);
    if (!result_ready) return false;

    // There are three possible hypothesis:
    // 1. Both the carrier and BPSK are present
    // 2. Only the carrier is present
    // 3. Neither is present, only noise

    // If there's signal, the peak must be much higher than the mean
    if (max_power < 4 * mean_power)
    {
        printf("freq_acq::signal not found\n");
        result.relative_power = -1;
        return true;
    }

    // If the carrier is present, it will be above BPSK.
    // The 'max_idx' returned by the acq is relative to the crosscorrelation
    // output.  The crosscorrelation is computed for offsets from 
    // -fft_size to + fft_size.  Thus, if max_idx is x, the best alignment
    // was achieved for offset x+fft_size.  The tone is set in the middle
    // of the mask, at index fft_size/2.  Thus, to convert from the crosscorr
    // index, to the fft index:
    max_idx = fft_size - max_idx + fft_size/2;
    // Search of the actual peak within the crosscorr mask
    vector<float>::iterator fft_center = fft.begin() + max_idx;

    float carrier_power = *max_element(fft_center - CARRIER_WIDTH/2,
                                       fft_center + CARRIER_WIDTH/2);

    // Find the max of the BPSK region.  Again, search positive offset,
    // negative offset, or both
    // BPSK Fc is negative
    float bpsk_neg_power = 0.0;
    if (config.Fc < 0 || config.double_sided)
        bpsk_neg_power = *max_element(fft_center-bpsk_center-bpsk_width/2, 
                                      fft_center-bpsk_center+bpsk_width/2);
    // BPSK Fc is positive
    float bpsk_pos_power = 0.0;
    if (config.Fc > 0 || config.double_sided)
        bpsk_pos_power = *max_element(fft_center+bpsk_center-bpsk_width/2, 
                                      fft_center+bpsk_center+bpsk_width/2);


    printf("freq_acq::Powers (carrier/pos/neg) = %2.2g/%2.2g/%2.2g\n", 
        carrier_power, bpsk_pos_power, bpsk_neg_power);

    float bpsk_power = max(bpsk_neg_power, bpsk_pos_power);

    if (carrier_power < bpsk_power)
    {
        printf("freq_acq::carrier not present\n");
        result.relative_power = -1;
        return true;
    }

    result.freq = (max_idx - acq->get_fft_size()/2) * d_fft_bin_size;
    result.relative_power = carrier_power / bpsk_power;


    return true;

}*/
