#pragma once

#include "gnuradio_misc.h"
#include <gr_single_threaded_scheduler.h>

using namespace std;

//! This structure defines the search parameters for the freq_acq class
typedef struct {
    float Fs;       //!< Sampling frequency at the input
    int delay; //Number of samples per symbol of the incoming waveform (delay purposes) JR
    float Fc;       //!< Subcarrier frequency
    float resolution; //!< Detection resolution.  Equal to the FFT bin size
    float freq_range; //!< Maximum frequency range to search = F_max-F_min
    float num_fft_avg; //!< Number of FFT blocks to average for each estimate
    bool  square_law; //!< Square the input signal to remove modulation
    //float bandwidth; //Approximate Bandwidth of the signal you are trying to acquire.  Used in Single Order Mode.

    //Added by JR to allow freq acq for recorded files
    char filename[200];
    bool rec_data;
    bool repeat;

    // Added by EV to configure USRP
    int	            usrp_which_board;   // USRP Board
    unsigned int	usrp_decim_rate;    // Deciation Rate
    int			    usrp_nchan;		// No of Channels=1
    int			    usrp_mux;			// MUX = 1;
    int			    usrp_mode;		     // Mode= 0;
    double		    usrp_fc;			// Center freq. for USRP
    // EVint			    lrit_filter;
    //int			    file_sink_on;
    char           *log_dir;
    bool            fourth_law; //!< Square the input signal to remove modulation - QPSK
    //bool            oqpsk;       // Enable Offset OQPSK
    float           samples_per_symbol;

} freq_acq_config;

typedef struct {
    float relative_power;
    float freq;
} freq_acq_result;

/*! This module performs FFT-based frequency acquisition.  It should work for
 *  most linear modulations or for pure tones.  The acquisition is based on
 *  detecting a tone.  The tone can be generated by passing the input signal
 *  through a nonlinearity (squaring).  The resultant tone is detected by
 *  taking multiple FFTs and averaging them. A combination of GNU radio blocks
 *  is used
 */
class freq_acq
{
private:
    vector<gr_block_sptr> d_blocks;
    freq_acq_config d_config;
    float d_fft_bin_size;

    gr_single_threaded_scheduler_sptr d_scheduler;

public:
    //! Create the transmitter.  See documentation on freq_acq_config
    freq_acq();
    freq_acq(const freq_acq_config &config, bool no_hardware);
    //! Setup or re-setup the frequency acquisition block
    void setup(const freq_acq_config &config, bool no_hardware);
    //! Return a vector of all the blocks created in this module
    vector<gr_block_sptr> &get_blocks() {
        return d_blocks;
    }
    //! Get the first block -- mixer
    gr_block_sptr get_input_block()
    {
        return find_block(d_blocks, "freq_acq::mixer");
    }
    //! Get the frequencies of the N highest peaks. Returns false if
    //! the acquisition engine does not have an estimate ready yet.
    //! \param num_peaks number of largest peaks to return
    //! \param results a vector of results.
    bool get_freqs(int num_peaks, vector<freq_acq_result> &results);
    //! Get the center frequency of the peak (interpolated). Returns
    //! false if the acquisition engine does not have an estimate ready yet.
    //! Computes a center of mass to get a precise frequency estimate
    //! \param result Estimated frequency and relative power
    bool get_peak_freq(freq_acq_result &result);

    //! Start the scheduler in a separate thread.  Returns immediately
    void run();
    //! Stop the scheduler.  Blocks until the scheduler has stopped
    void stop();
    //! Check if the scheduler is running
    bool running();
    //! BPSK-specific acquisition
    //bool get_crosscorr_freq(bpsk_config &config, freq_acq_result &result);
};

