/*
GOES Radio Comments
This function is only used for the L-band board design.  The L-band board design has not be proven
to meet GOES-R System specifications.

This program uses the spi interface hooks available via the original USRP design to configure the 
DFS chip. Configuring the DFS chip will control what IF frequency you come down to.  
*/

#include "front_end_ctrl.h"
#include <usrp1_source_c.h>
#include <string>
#include "windows.h"

using namespace std;
bool set_frend_dfs(usrp1_source_c_sptr usrp_source, double freq )
{
/*
Freequency calculation of the DFS
ADF 4360-3 (refer to data sheet for more information
fout = [(PxB)+A]xfref/R;
where:
P - modulus prescaler (0,1,2) -> 8, 16 and 32 
B - divide ratio 13 bit (3-8191)
A - divide ratio 5  bit (0-31)
R - divide ratio 14 bit (1-16383)
fref = 27Mhz

_write_spi function is used to controll the DFS on the board.
The format is as follows:

     _write_spi(header,chip_select,format,information);

	 Where:
    "header"      - for this device set it to ->"0" no header
	"chip_select" - I have selected to use SEN_TX_A wire on the board(see schematics pin 90 on USB chip)
    and it is 0x10 (if need more info look in usrp_spi_defs.h and usrp_commands.h) 
	"format"      - MSB or LSB first 0x00 is MSB and 0x80 for LSB first
	"information" is an array of bytes, and has to contain actual informaion to be sent on the SPI Bus
*/  

	int A,P,B,R;
	double temp = 0;
	bool   spi_good = true;
    int latchR, latchC, latchN;
	//Hardwire to 32
	double ref = 27e6;//ref freq
    P = 2;
	R = 135;// for 200kHz step size
	//R = 270;// for 100kHz step size but you need to increase B and A so that numerator is increased by 2

	//Define Latch "C"***************************************************************************
    latchC = (((P << 22)&0xC00000)|OUT_POW_12dBm|PHASE_DET_POL|MUX_OUT_CTRL_2|CORE_POW_05mA|SELECT_C_L)&BUFFER_MASK;
	//Define Latch "C"***************************************************************************

	//Prescalar is 32-> division by 32
	temp = R*(freq/ref)/32;
	B = temp;
	A = (temp-B)*32;
    if(B > 8191 )
        B = 8191;
    if(B < 3)
        B = 3;
    if(A < 0)
        A = 0;
    if(A > 31)
        A = 31;

	//Define Latch "N"***************************************************************************
    latchN = (CP_GAIN_ON|((B << 8)&0x1FFF00)|((A << 2)&0x00007C)|SELECT_N_L)&BUFFER_MASK;
	//Define Latch "N"***************************************************************************

	//Define Latch "R"***************************************************************************
    latchR = (ANTI_BACK_PW_3ns|((R << 2)&0x00FFFC)|SELECT_R_L)&BUFFER_MASK;
	//Define Latch "R"***************************************************************************
 
    //Initalize buffer string
    string buffC;
	buffC.resize(3);
    string buffR;
	buffR.resize(3);
    string buffN;
	buffN.resize(3);

    buffC[2] = (latchC >> 0) & 0xFF;
    buffC[1] = (latchC >> 8) & 0xFF;
    buffC[0] = (latchC >> 16)& 0xFF;

    buffN[2] = (latchN >> 0) & 0xFF;
    buffN[1] = (latchN >> 8) & 0xFF;
    buffN[0] = (latchN >> 16)& 0xFF;

    buffR[2] = (latchR >> 0) & 0xFF;
    buffR[1] = (latchR >> 8) & 0xFF;
    buffR[0] = (latchR >> 16)& 0xFF;
    
	//For some reason need to do this twice for the settings to take effect
	int count = 0;
	do
	{
		count++;
		//1)R latch is writter
		spi_good = usrp_source->_write_spi(0 ,SPI_ENABLE_DFS ,SPI_DFS_MSB ,buffR);
		if(!spi_good)
			return false;
		//2)Control latch is written
		spi_good = usrp_source->_write_spi(0 ,SPI_ENABLE_DFS ,SPI_DFS_MSB ,buffC);
		if(!spi_good)
			return false;
		//3)Control latch is written
		spi_good = usrp_source->_write_spi(0 ,SPI_ENABLE_DFS ,SPI_DFS_MSB ,buffN);
		if(!spi_good)
			return false;
	}
	while (count<=1);
	return true;
}

float set_frend_gain(usrp1_source_c_sptr usrp_source, int which_board, int gain)
{
    /*
    Gain choices are as follows:
    LNA(high) MIXER(low)  -> GAIN = 01; 12dB
    LNA(low)  MIXER(high) -> GAIN = 02; 13dB
    LNA(high) MIXER(high) -> GAIN = 03; 25dB
    LNA(low)  MIXER(low)  -> GAIN = 04; 00dB
   
    This interface controlls IO_RX_A[15,0] (marked on the schematics) FPGA pins [116-108] & [213-195] 
    Front end gains of the LNA in the Mixer and Mixer itself are connected
    to pins IO_RX_A_11 (pin203)->LNA Gain and IO_RX_A_10(pin202)->MIXER Gain
    */
    usrp_source->_write_oe(which_board, 0x0c00,0x0c00);
		
    switch (gain){
         case 1:
            usrp_source->write_io(which_board, 0x0900, 0x0c00);
            return 12.00;
        case 2:
            usrp_source->write_io(which_board, 0x0400, 0x0c00);
            return 13.00;
        case 3:
            usrp_source->write_io(which_board, 0x0c00, 0x0c00);
            return 25.00;
        case 4:
            usrp_source->write_io(which_board, 0x0000, 0x0c00);
            return 0.00;
        default:
            //return false if unknown combination
            return -1.00;
         }
}