/* GOES Radio Comments 
ar_lrit_image_sink is a gr_sync_block type that takes in inputs, but does not
generate an output for another gr_block.  This particular project creates and instantiates 
the GOESprocessor class with the name lrit_processor.  Upon starting up, this block takes
instantiates Goesprocessor::InitializeProcessor(); and 
GOESProcessor::InitialzeSocket(socket emwin port, socket lrit port, socket_output_only)
During successive calls, this the lrit_image_sink class calls GOESProcessor::processlritDatablock
Process LritDatablock is the data link layer processor for LRIT/HRIT
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_lrit_image_sink.h>
#include <gr_io_signature.h>
#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdexcept>

#define RE_ACQ

// win32 (mingw/msvc) specific
#ifdef HAVE_IO_H
#include <io.h>
#endif
#ifdef O_BINARY
#define	OUR_O_BINARY O_BINARY
#else
#define	OUR_O_BINARY 0
#endif

// should be handled via configure
#ifdef O_LARGEFILE
#define	OUR_O_LARGEFILE	O_LARGEFILE
#else
#define	OUR_O_LARGEFILE 0
#endif


ar_lrit_image_sink::ar_lrit_image_sink(size_t itemsize, const char *filename, int resynch_thresh,char * BERFileName, int socket_flag, int socket_lrit_flag, int only_socket)
  : gr_sync_block ("file_sink",
		   gr_make_io_signature(1, 1, itemsize),
		   gr_make_io_signature(0, 0, 0)),
    d_itemsize(itemsize), d_fp(0), d_new_fp(0), d_resynch_thresh(resynch_thresh), d_updated(false), d_BERFileName(BERFileName), d_socket_flag(socket_flag), d_socket_lrit_flag(socket_lrit_flag), d_only_socket(only_socket)
{
  if (!open(filename))
    throw std::runtime_error ("can't open file");

  if (!lrit_processor.InitializeProcessor())
	  throw std::runtime_error ("can't initialize goes processor");

    lrit_processor.InitializeSocket(d_socket_flag, d_socket_lrit_flag,d_only_socket);
    Synched = TRUE;
       
}


ar_lrit_image_sink::ar_lrit_image_sink(size_t itemsize, const char *filename, int resynch_thresh,char * BERFileName, int socket_flag, int socket_lrit_flag)
  : gr_sync_block ("file_sink",
		   gr_make_io_signature(1, 1, itemsize),
		   gr_make_io_signature(0, 0, 0)),
    d_itemsize(itemsize), d_fp(0), d_new_fp(0), d_resynch_thresh(resynch_thresh), d_updated(false), d_BERFileName(BERFileName), d_socket_flag(socket_flag), d_socket_lrit_flag(socket_lrit_flag), d_only_socket(0)
{
  if (!open(filename))
    throw std::runtime_error ("can't open file");

  if (!lrit_processor.InitializeProcessor())
	  throw std::runtime_error ("can't initialize goes processor");

    lrit_processor.InitializeSocket(d_socket_flag, d_socket_lrit_flag,d_only_socket);
  Synched = TRUE;
       
}


ar_lrit_image_sink_sptr
ar_make_lrit_image_sink (size_t itemsize, const char *filename, int resynch_thresh,char * BERFileName)
{
  return ar_lrit_image_sink_sptr (new ar_lrit_image_sink (itemsize, filename, resynch_thresh,BERFileName, 0, 0));
}


ar_lrit_image_sink_sptr
ar_make_lrit_image_sink (size_t itemsize, const char *filename, int resynch_thresh,char * BERFileName, int socket_flag, int socket_lrit_flag)
{
  return ar_lrit_image_sink_sptr (new ar_lrit_image_sink (itemsize, filename, resynch_thresh,BERFileName, socket_flag, socket_lrit_flag));
}

ar_lrit_image_sink_sptr
ar_make_lrit_image_sink (size_t itemsize, const char *filename, int resynch_thresh,char * BERFileName, int socket_flag, int socket_lrit_flag, int only_socket)
{
  return ar_lrit_image_sink_sptr (new ar_lrit_image_sink (itemsize, filename, resynch_thresh,BERFileName, socket_flag, socket_lrit_flag,only_socket));
}

ar_lrit_image_sink::~ar_lrit_image_sink ()
{
  close();
  if (d_fp){
    fclose((FILE *) d_fp);
    d_fp = 0;
  }
}

bool
ar_lrit_image_sink::open(const char *filename)
{
  omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this function

  // we use the open system call to get access to the O_LARGEFILE flag.
  int fd;
  if ((fd = ::open (filename,
		    O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY, 0664)) < 0){
    perror (filename);
    return false;
  }

  if (d_new_fp){		// if we've already got a new one open, close it
    fclose((FILE *) d_new_fp);
    d_new_fp = 0;
  }

  if ((d_new_fp = fdopen (fd, "wb")) == NULL){
    perror (filename);
    ::close(fd);		// don't leak file descriptor if fdopen fails.
  }

  d_updated = true;
  return d_new_fp != 0;
}

void
ar_lrit_image_sink::close()
{
  omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this function
  if (d_new_fp){
    fclose((FILE *) d_new_fp);
    d_new_fp = 0;
  }
  d_updated = true;
}

bool ar_lrit_image_sink::get_synch(){

    return Synched;
}

int 
ar_lrit_image_sink::work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items)
{
  unsigned char *inbuf = (unsigned char *) input_items[0];
  int  nwritten = 0;

  if (d_updated){
    omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this block
    if (d_fp)
      fclose((FILE *)d_fp);
    d_fp = d_new_fp;			// install new file pointer
    d_new_fp = 0;
    d_updated = false;
  }
  
  if (!d_fp)
    return noutput_items;		// drop output on the floor

#ifdef RE_ACQ
  Synched=TRUE;
  Synched = lrit_processor.ProcessLritDataBlock(inbuf, noutput_items, d_resynch_thresh, d_BERFileName);
  if(Synched){
      if (noutput_items==0){
          printf( "noutput_items = 0 \n");
      }
    return noutput_items;
  }else{
    printf( "LRIT Image_sink PROCESSOR:: Not synched \n");
    return -1;
  }
#else
  lrit_processor.ProcessLritDataBlock(inbuf, noutput_items, d_resynch_thresh, d_BERFileName);
  if (noutput_items==0){
          printf( "noutput_items = 0 \n");
  }
 return noutput_items;
#endif

 
 
}

