#pragma once

#include "gnuradio_misc.h"

#include "circular_buffer.h" // Thread-safe ring buffer
#include <gr_single_threaded_scheduler.h>

//! BPSK modem configuration.  This structure contains all the variables
//! needed to define a BPSK modem.
typedef struct bpsk_config {
    float Fs;       //!< Sampling frequency at the input
    float Fc;       //!< Subcarrier frequency
    float bit_rate; //!< Symbol rate
    float shaping_rolloff; //!< Rolloff factor for the SRRCF.  No filter if
                           //!< rolloff > 0.99
    bool double_sided; //!< Double sided -- real output only
    // Esteban
    float   samples_per_symbol;
    float   TBW;
    float   PBW;
    float   AGC;
    int     Vit_Resynch_Period;

    //EV; Need this in BPSK for Single-Thread Mode 
    float   vthresh;
    int     resynch_thresh;
    int     resynch_period;     
    char * BERFileName;
    bool    sink_data;
    
    // Sockets for single thread mode
    int     socket_flag; 
    int     socket_lrit_flag;
    int     only_socket;

    bpsk_config() : Fs(0),Fc(0),   bit_rate(0), shaping_rolloff(0), 
        double_sided(false),samples_per_symbol(0)
    { }
} bpsk_config;

// This is used for re-acq versions
typedef struct lrit_config {
    float   Fs;       //!< Sampling frequency at the input
    float   Fc;       //!< Subcarrier frequency
    float   bit_rate; //!< Symbol rate
    float   shaping_rolloff; //!< Rolloff factor for the SRRCF.  No filter if
                           //!< rolloff > 0.99
    bool    double_sided; //!< Double sided -- real output only
    // Esteban
    float   samples_per_symbol;
	int		file_sink_on;
    float   icyFreqMult;
    bool    lrit_filter_on;
    bool    using_rec_data;
    bool    file_sync_on;
    char    in_rec_data_file[200];
    bool    RunACQEngine;
    int     decim_rate;
    float   foff;
    int     buff_size;
    float   TBW;
    float   PBW;
    float   AGC;
    int     using_icy;
    // NEW : Do we ever use lrit_config ?
    float   vit_thresh;
    int     resynch_period;     
    int     resynch_thresh;
    int     num_fft_avg; 
    float   resolution;
    float   freq_range;
    bool    square_law;
    bool    fourth_law;
    bool    repeat;
    char * BERFileName;
    int     socket_flag; 
    int     socket_lrit_flag;
    int     only_socket;

    lrit_config() : Fs(0), Fc(0), bit_rate(0), shaping_rolloff(0), 
        double_sided(false),samples_per_symbol(0),icyFreqMult(1.0),
        lrit_filter_on(TRUE),file_sync_on(FALSE),num_fft_avg(30)
    { }
} lrit_config;



typedef struct  viterbi_config {
    float AGC;       //!< Sampling frequency at the input
    int resynch_thresh;
    float vthresh;
    int resynch_period; 
    //FILE * BERFILE;
    char * BERFileName;
    bool verbose;
    int     socket_flag; 
    int     socket_lrit_flag;
    int     only_socket;


    viterbi_config() : AGC(10.0), resynch_thresh(0), resynch_period(0), vthresh(0.0)
    { }
} viterbi_config;

using namespace std;

/*! BPSK transmitter class.  This class encapsulates a number of
 *  GNU radio blocks to implement a BPSK modulator
 */
class bpsk_tx
{
private:
    vector<gr_block_sptr> d_blocks;

public:
    //! Create the transmitter.  See documentation on bpsk_config
    bpsk_tx(vector<gr_block_sptr> &blocks, const bpsk_config &config);
    //! Return a vector of all the blocks created in this module
    vector<gr_block_sptr> &get_blocks() {return d_blocks;}
    //! Get the first block -- bit to symbol mapper
    gr_block_sptr get_input_block() 
        {return find_block(d_blocks, "bpsk_tx::mapper");}
    //! Get the last block -- adder for the 70 MHz tone
    gr_block_sptr get_output_block() 
        {return find_block(d_blocks, "bpsk_tx::add_tone");}
    //! Set the amplitude of the 70 MHz tone
    void set_70MHz_amplitude(float amplitude);
    //! Set the amplitude of the BPSK constellation
    void set_amplitude(float amplitude);
};

/*! BPSK receiver class.  This class encapsulates a number of
 *  GNU radio blocks to implement a BPSK demodulator
 */
class bpsk_rx
{
private:
    vector<gr_block_sptr> d_blocks;
     bpsk_config d_config;

    // Keep a local copy of frequently used blocks to avoid wasting time
    // in the find_block routine
    gr_block_sptr d_phase_track;

public:
    //! Create the receiver.  See documentation on bpsk_config
    bpsk_rx(vector<gr_block_sptr> &blocks, const bpsk_config &config);
    
    //! Return a vector of all the blocks created in this module
    vector<gr_block_sptr> &get_blocks() {return d_blocks;}

    //! Get the first block -- subcarrier mixer
    gr_block_sptr get_input_block()
        {return find_block(d_blocks, "bpsk_rx::agc");} 

    //! Get the last block -- the slicer
    gr_block_sptr get_bits() 
        {return find_block(d_blocks, "bpsk_rx::slicer");}
    
   gr_block_sptr get_softout() 
        {return find_block(d_blocks, "bpsk_rx::slicer_real");}

    //! Check if the phase tracking locked
    bool get_phase_lock();
    //! Set the bandwidth of the phase tracking loop (costas) [Hz]
    void set_phase_bandwidth(float bw);
    //! Set the bandwidth of the symbol/timing tracking loop (M&M) [Hz]
    void set_timing_bandwidth(float bw);
 
    bool get_packet_lock();

    //! See comment in ::transponder
    circular_buffer<gr_complex> *get_bent_pipe();

};


//! Viterbi Decoder configuration.  This structure contains all the variables
//! needed to define a Viterbi Decoder modem.
/*
typedef struct viterbi_config {   
    bool diff_decoder; //!< Differential Output
 
    viterbi_config() : diff_decoder(false)
    { }
} viterbi_config;
*/
class viterbi_rx
{
private:
    vector<gr_block_sptr> d_blocks;
    viterbi_config d_config;
      // Keep a local copy of frequently used blocks to avoid wasting time
    // in the find_block routine
    gr_block_sptr           d_lrit, d_emwin;

public:
    //! Create the decoder.  See documentation on bpsk_config
    //viterbi_rx(vector<gr_block_sptr> &blocks, const viterbi_config &config);
    viterbi_rx(vector<gr_block_sptr> &blocks, const viterbi_config &config);
    viterbi_rx(vector<gr_block_sptr> &blocks);
    viterbi_rx(vector<gr_block_sptr> &blocks, int Resynch_Period);

    //! Return a vector of all the blocks created in this module
    vector<gr_block_sptr> &get_blocks() {return d_blocks;}

     gr_block_sptr get_input_block()
        {return find_block(d_blocks, "viterbi_rx::input");}
    
    //circular_buffer<gr_complex> *put_bent_pipe();

    bool get_packet_lock();

};
