/*
GOES Radio Comments
This is the bpsk demodulation class.  This class configures a BPSK
demodulator based on a configuration struct that is created inside  main_emwin_hrit.cpp. 
BPSK demodulator creates and configures the following blocks: An AGC, A carrier recovery and tracking 
loop based on the costas loop structure.  The bandwidth of this loop is set using the function 
set_phase_bandwidth.  The loop coefficients rely on the S-curve of the loop, which needs to know the AGC
level.  

After phase tracking, matched filtering is performed, and then clock recovery and tracking. The
timing tracking algorithm is implemented using the mueller & muller structure.  The timing bandwidth is 
set using a function set_timing_bandwidth ().  After timing tracking, the output is sent to a data converter that
turns the complex data into a pair of floats (slicer real).  This is basically a data type conversion that takes
a variable of type gr_complex, x = a + b*j, into a series of paired floats <a,b>.  Its called slicer_real, but it
is not a hard decision slicer, since all of the precision is retained. 

Slicer real is then either fed into a null_sink, where it is eventually sent to the Viterbi Decoder via a ring_buffer for 
further processing (multithreaded solution), or it goes to a local instance of the Viterbi decoder 
(single threaded solution).  There are optional data sinks that print out various test points 
throughout the BPSK demodulation chain, but these sinks are completely commented out.  There is an 
option to write the output of the timing tracked data to a file, based on a sink flag. 

For Single Core Proecssing, the ouput of the viterbi decoder goes to the data link layer (ar_lrit_image_sink)
or (ar_emwin_image_sink).  If USE_EMWIN is defined, only embedded emwin will be processed, if USE_LRIT is defined
both EMWIN and LRIT data will be processed.  
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

//AERO RADIO
#include <ar_fractional_interpolator_cc.h>
#include <ar_clock_recovery_mm_cc.h>
#include <ar_mixer_cc.h>
#include <ar_short_to_uchar.h>
#include "ar_lrit_image_sink.h"
#include "ar_emwin_image_sink.h"
#include <ar_snr_estimator_cf.h>
#include "ar_viterbi.h"
#include "bpsk_modem.h"
//GNU RADIO
#include <ar_sig_source_c.h>
#include <gr_file_sink.h>
#include <gr_firdes.h>
#include <ar_costas_loop_cc.h>
#include <ar_fir_filter_ccf.h>
#include <ar_throughput.h>
#include <ar_complex_to_xxx.h>
#include <gr_binary_slicer_fb.h>
#include <gr_agc_cc.h>
#include "gnuradio.h"
#include "gnuradio_misc.h"
#include <ar_file_sink.h>
#include <gr_null_sink.h>
// MultiThread
// RING BUFFER
#include <ar_ring_buffer_source2.h>
#include <ar_ring_buffer_sink2.h>
#include <gr_unpacked_to_packed_bb.h>
using namespace std;

//*****************************
// BLOCKS TO USE
#define MULTI_THREAD
//#define USE_SNR_ESTIMATOR
//#define USE_LRIT
//#undef USR_LRIT
//#define USE_EMWIN
//#undef MULTI_THREAD
#define THROUGHPUT
//#undef THROUGHPUT
//#define THROUGHPUT_VITERBI
//#undef THROUGHPUT_VITERBI
//#define SINK_ALL
#undef SINK_ALL

//*****************************


// MULTITHREAD VERSION
bpsk_rx::bpsk_rx(vector<gr_block_sptr> &blocks, const bpsk_config &config)
{
    d_config = config;

    // MEM LEAK TEST
    gr_block_sptr null_sink_f =  gr_make_null_sink (sizeof(float));
    gr_block_sptr null_sink_c =  gr_make_null_sink (sizeof(char));

    // **********************************************************************
    // ************** Create all the signal processing blocks ***************
    // **********************************************************************

    // Current signal chain:
    //  [source (mixed to baseband during freq.acq] ->
    // (1) -> agc -> costas_loop -> matched_filter -> clock_recovery -> slicer
    // --------------------- AGC --------------------------------------------
    // The AGC is used to present a constant magnitude into the costas loop
    // to keep the loop gain constant.
    float agc_rate			= 1e-5;
    gr_agc_cc_sptr agc		= gr_make_agc_cc (agc_rate, config.AGC);
    agc->set_name("bpsk_rx::agc");
    //---------------------- BPSK: Costas Loop phase tracking ------------
    gr_costas_loop_cc_sptr phase_track = // Set bandwidth later
        gr_make_costas_loop_cc (0, 0, 0.1, -0.1, 2); // configure for bpsk only
    // Costas Input Parameters: alpha, beta, max_freq, min_freq, order
    // Right now the loop is turned off
    phase_track->set_name("bpsk_rx::phase_track");
    phase_track->set_lock_threshold(50);
    // --> Bandwidth must be set after blocks are connected <<--
    //
    //---------------------- Matched Filter -----------------------
    // Shaping filter: Root raised cosine or Rectangle
    std::vector<float> shaping_taps;
    if (config.shaping_rolloff < 0.99){
        shaping_taps =					// gain, sampling freq, symbol rate, alpha, ntaps
            gr_firdes::root_raised_cosine (1.0, config.Fs, config.bit_rate, config.shaping_rolloff, 65);
    }else{// All ones (normalized to gain of 1)
        shaping_taps.resize(config.samples_per_symbol, 1/config.samples_per_symbol);
    }
    gr_fir_filter_ccf_sptr match_filt = gr_make_fir_filter_ccf (1, shaping_taps);
    //
    //---------------------- BPSK: M&M timing tracking ------------------
    //float timing_omega		= config.samples_per_symbol; // Nominal is 4 samples/symbol
    float timing_omega			= config.Fs/config.bit_rate; // Nominal is 4 samples/symbol
    float timing_omega_limit	     = 0.01;
    ar_clock_recovery_mm_cc_sptr timing_track = // Set bandwidth later
        ar_make_clock_recovery_mm_cc(timing_omega, 0, 0, 0, timing_omega_limit);
    timing_track->set_name("bpsk_rx::timing_track");
    //---------------------- Slicer ------------------------------------
    gr_complex_to_real_sptr slicer_real     = gr_make_complex_to_real ();
    slicer_real->set_name("bpsk_rx::slicer_real");
    gr_binary_slicer_fb_sptr slicer         = gr_make_binary_slicer_fb ();
    slicer->set_name("bpsk_rx::slicer");
    

    //---------------------- BPSK: observation points ------------------

//IF in Single Threaded Mode
#ifndef MULTI_THREAD   
#ifdef USE_LRIT  //If the LRIT data is a product of interest  
    //Create an instance of the data link layer processor locally
    ar_lrit_image_sink_sptr lrit_sink =
        ar_make_lrit_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/lrit_dump.dat", config.resynch_thresh, config.BERFileName);
    
#endif
#ifdef USE_EMWIN  //If in single threaded mode, but interested in only embedded EMWIN products
    //Create an instance of the embedded EMWIN data link layer processor locally
    ar_emwin_image_sink_sptr   emwin_sink =
        ar_make_emwin_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/emwin_dump.dat");
#else
//IF in multithreaded mode, create a null sink, since data link layer processing will happen inside viterbi_Rx 
#ifndef USE_LRIT  
    gr_block_sptr null_sink_c =  gr_make_null_sink (sizeof(unsigned char));
#endif
#endif
#endif
    
    // **********************************************************************
    // ************** Connect the blocks together ***************************
    // **********************************************************************
    // The syntax is self explanatory: from_block.from_port -> to_block.to_port
    //--------------------------------------------------------------
    //--------------------------------------------------------------
    // FIXED BLOCKS
    connect_blocks(agc,             0, phase_track,     0, d_blocks);
    connect_blocks(phase_track,	    0, match_filt,	     0, d_blocks);
    connect_blocks(match_filt,	    0, timing_track,    0, d_blocks);
    connect_blocks(timing_track,    0, slicer_real,	0, d_blocks); 

#ifdef THROUGHPUT
    gr_throughput_sptr throughput =
        gr_make_throughput(sizeof(gr_complex), 3, "cout", "BPSK_RX");
    connect_blocks(slicer_real,  0, throughput, 0, d_blocks);
#endif
    // -----------------------    RING BUFFER (Multithreaded Output to Viterbi Decoder   --------------------------
#ifdef MULTI_THREAD
    gr_block_sptr  null_sink1 =
        gr_make_null_sink (sizeof(float));
    connect_blocks(slicer_real,  0, null_sink1,	0, d_blocks);  //End of Multithreaded BPSK demodulator!!!! 

#else  
     //--------------------------------------------------------------
    //In Single Threaded Mode, Do Viterbi Decoding and Data Link layer processing as defined below
    //--------------------------------------------------------------
    
    
    //------------------------- Bits to Bytes     ----------------------------------
    gr_unpacked_to_packed_bb_sptr   viterbi_bytes
        = gr_make_unpacked_to_packed_bb(1,GR_MSB_FIRST);
    viterbi_bytes->set_name("bpsk_rx::viterbi_bytes");
    //------------------------------------------------------------------------------
    //gr_viterbi_sptr     viterbi_decoder = gr_make_viterbi();
    //BPSK
    gr_viterbi_sptr     viterbi_decoder = gr_make_viterbi(2,config.resynch_period, config.AGC, config.vthresh);
        viterbi_decoder->set_name("viterbi_rx::input");
    
    // VITERBI DECODER
    connect_blocks(slicer_real,	     0, viterbi_decoder,	     0, d_blocks);
    connect_blocks(viterbi_decoder,	0, viterbi_bytes,	     0, d_blocks);

    //connect_blocks(slicer_real,	     0, null_sink_f,	     0, d_blocks); --> No leaks up to here


//--------------------------------------------------------------
// data link layer processing
//--------------------------------------------------------------
   
#ifdef USE_LRIT
    connect_blocks(viterbi_bytes,    0, lrit_sink, 0, d_blocks);
#endif
#ifdef USE_EMWIN
    connect_blocks(viterbi_bytes,    0, emwin_sink, 0, d_blocks);
#else
#ifndef USE_LRIT
    connect_blocks(viterbi_bytes,    0, null_sink_c,   0, d_blocks);
#endif

#endif

#ifdef THROUGHPUT_VITERBI
    gr_throughput_sptr throughputVit =
        gr_make_throughput(sizeof(gr_complex), 3, "cout", "viterbi_rx");
    // THROUGHPUT
    connect_blocks(viterbi_decoder,  0, throughputVit, 0, d_blocks);
#endif

#endif
    //--------------------------------------------------------------
    // SNR ESTIMATOR
    //-----------------------SNR -------------------------------------
#ifdef USE_SNR_ESTIMATOR
    vector<gr_complex> symbol_table(2);
    symbol_table[0] = gr_complex(-config.AGC, 0.0);
    symbol_table[1] = gr_complex( config.AGC, 0.0);
    gr_snr_estimator_cf_sptr snr_estimator = gr_make_snr_estimator_cf(symbol_table);
    snr_estimator->set_name("bpsk_rx::snr_estimator_cf");
    snr_estimator->set_averaging(0.001);

    // SNR SINK
    gr_file_sink_sptr snr_sink =
         gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/bpsk_rx_snr_est.dat");

    connect_blocks(timing_track,  0, snr_estimator, 0, d_blocks);
    connect_blocks(snr_estimator, 0, snr_sink, 0, d_blocks);
#endif
    //--------------------------------------------------------------
    // FILE SINKS
    //---------------------

   gr_file_sink_sptr timing_sink;
   if(config.sink_data){
        timing_sink = gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_tracked.dat");
        connect_blocks(timing_track,        0, timing_sink,    0, d_blocks);
    }


#ifdef SINK_ALL
    /*
    gr_file_sink_sptr agc_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/bpsk_rx_agc.dat");
    connect_blocks(agc,           0, agc_sink,     0, d_blocks);

    gr_file_sink_sptr phase_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/phase_tracked.dat");
    connect_blocks(phase_track,	0, phase_sink,	     0, d_blocks);

    gr_file_sink_sptr phase_err_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/phase_error.dat");
    gr_file_sink_sptr timing_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_tracked.dat");
    //gr_file_sink_sptr timing_err_sink =
    //	gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_error.dat");
    gr_file_sink_sptr mixer_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/bpsk_rx_mixer.dat");
    gr_file_sink_sptr match_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/match_filter.dat");
        connect_blocks(match_filt,	0, match_sink,    0, d_blocks);
    //gr_file_sink_sptr slicer_sink =
    //		gr_make_file_sink (sizeof(char), "c:/temp/usrp_log/slicer.dat");
    gr_file_sink_sptr slice_real_sink =
        gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/slicer_real.dat");
    */

#endif

    //------------------------------------------------------------------------
    // COSTAS LOOP (cont.)
    // Bandwidth must be set after blocks are connected.  Note that we rely
    // on the front end PLL for the actual frequency offset/doppler tracking.
    // The costas loop is for phase correction only.
    set_phase_bandwidth(config.bit_rate / config.PBW); //JR Uncomment
    //set_phase_bandwidth(0);  //JR Debug S-curve
    set_timing_bandwidth(config.bit_rate / config.TBW);
    //------------------------------------------------------------------------

    // Append a copy of pointers to all the d_blocks created here to the
    // input list of d_blocks.  Note the funky 'back_insert_iterator'
    // EV Problems Here
    copy(d_blocks.begin(), d_blocks.end(),
         back_insert_iterator<vector<gr_block_sptr> >(blocks));
}

/*
bool bpsk_rx::get_phase_lock(){
    gr_costas_loop_cc *phase_track =
        find_block_ptr<gr_costas_loop_cc>(d_blocks, "bpsk_rx::phase_track");
    return phase_track->lock_detector();
}
*/
// EV According to Eugene this version is faster
bool bpsk_rx::get_phase_lock()
{
    gr_costas_loop_cc_sptr phase_track =
        cast_block<gr_costas_loop_cc>(d_phase_track);

    return phase_track->lock_detector();
}

void bpsk_rx::set_phase_bandwidth(float bw){
    gr_costas_loop_cc *phase_track =
        find_block_ptr<gr_costas_loop_cc>(d_blocks, "bpsk_rx::phase_track");

    // Compute the loop coefficients.
    // The phase detector for BPSK is simply e=Re[x] * Im[x].  For
    // angle, z, and a signal magnitude A, e = A Cos[z] * A Sin[z]
    // For small z, e = A * A * z = A^2*z.  The AGC sets the signal
    // magnitude such that Mean[Sqrt[Abs[x]]] = ref level.  The loop
    // gain is then G(z) = e(z) / z = (ref level)^2
    gr_agc_cc *agc = find_block_ptr<gr_agc_cc>(d_blocks, "bpsk_rx::agc");
    float K_ped = agc->reference() * agc->reference()*.75*.75;  //S-curve is .75^2 at 1dB, but 1^2 at 15 dB! No proof that Kped should be set to S-curve at Es/No = 1 dB 
    // See a document titled "2nd order tracking loop analysis" for details
    // Also see carrier_loop_coefficients.m
    float S      = 1.0 / sqrt(2.0); // damping factor
    float SS     = (S+1/(4*S));

    float     K_loop = K_ped ;
    float     Fs = d_config.bit_rate * d_config.samples_per_symbol; // Running at 4 samples/symbol

    float     alpha = 4 * bw * S  / (K_loop * Fs * SS);
    float     beta  = 4 * bw * bw / (K_loop * Fs * Fs * SS * SS);

    phase_track->set_loop_coefs(alpha, beta);

}

void bpsk_rx::set_timing_bandwidth(float bw){
    ar_clock_recovery_mm_cc *timing_track =
        find_block_ptr<ar_clock_recovery_mm_cc>(d_blocks, "bpsk_rx::timing_track");

    // Compute the loop coefficients.
    // The phase detector for M&M is
    // Re[ (Sign[x(n-2)] - Sign[x(n)]) * Conj[x(n-1)] ]
    // The loop gain is therefore proportional to the amplitude of the signal.
    // The actual loop gain is nontrivial to derive (see e.g. Mengali and
    // d'Andrea p. 386).  The AGC sets the signal magnitude such that
    // Mean[Sqrt[Abs[x]]] = ref level.  The loop gain is then = ref level
    gr_agc_cc *agc = find_block_ptr<gr_agc_cc>(d_blocks, "bpsk_rx::agc");
    float K_ped = agc->reference() * 0.35; // 0.35 is a fudge factor
    // See a document titled "2nd order tracking loop analysis" for details
    // Also see carrier_loop_coefficients.m
    float S      = 1.0 / sqrt(2.0); // damping factor
    float SS     = (S+1/(4*S));

    float K_loop = K_ped ;
    float  Fs = d_config.bit_rate ; // Update once per symbol

    float alpha = 4 * bw * S  / (K_loop * Fs * SS);
    float beta  = 4 * bw * bw / (K_loop * Fs * Fs * SS * SS);

    timing_track->set_gain_mu (alpha); // 1st order: alpha
    timing_track->set_gain_omega (beta); //2nd order: beta

}




