/*
GOES Radio Comments
This is the viterib decoder class.  This class is only valid in multithreading mode! 
This class receives the output of BPSK (slicer real).  Recall that slicer real is 
the output of the timing tracked data, that is a reformatted version of complex data
, and does not represent hard decision.  

An instance of a Viterbi Encoder is created.  The data is decoded.  The output of the 
viterbi decoder is in the form of shorts.  Each short presents one bit, and 
8 bits are packeged into one bytes via the (unpacked_to_packed) conversion block.  

The decoded data bytes are fed into either an EMWIN data link layer processor or an 
LRIT data link layer processor.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_sig_source_c.h>
#include <gr_file_sink.h>
#include <gr_firdes.h>
#include <ar_throughput.h>
#include <ar_complex_to_xxx.h>
#include <gr_binary_slicer_fb.h>
#include "gnuradio.h"
#include "gnuradio_misc.h"
#include "bpsk_modem.h"
#include "ar_viterbi.h"
#include <ar_short_to_uchar.h>
#include "ar_lrit_image_sink.h"
#include "ar_emwin_image_sink.h"
#include <gr_null_sink.h>
#include <gr_diff_decoder_bb.h>
#include <gr_unpacked_to_packed_bb.h>

// RING BUFFER
#include <ar_ring_buffer_source2.h>
#include <ar_ring_buffer_sink2.h>
using namespace std;

//*****************************
// BLOCKS TO USE
#define USE_LRIT
//#define USE_EMWIN
//#undef USE_LRIT

//#define THROUGHPUT_VITERBI
//#undef THROUGHPUT_VITERBI

//*****************************

viterbi_rx::viterbi_rx(vector<gr_block_sptr> &blocks, const viterbi_config &config)
{
     d_config = config;
    //------------------------- Create VITERBI DECODER --------------------------------
    gr_viterbi_sptr     viterbi_decoder = gr_make_viterbi(2,config.resynch_period, config.AGC, config.vthresh);
    viterbi_decoder->set_name("viterbi_rx::input");


    //------------------------- Bits to Bytes     ------------------------------------
    gr_unpacked_to_packed_bb_sptr   viterbi_bytes     = gr_make_unpacked_to_packed_bb(1,GR_MSB_FIRST);
    viterbi_bytes->set_name("viterbi_rx::viterbi_bytes");
    connect_blocks(viterbi_decoder,    0, viterbi_bytes, 0, d_blocks);

///--------------------------------------------------------------------------------------
// Define Data Link Layer Processing (either LRIT + EMWIN, or just EMWIN without LRIT products
//---------------------------------------------------------------------------------------
 #ifdef USE_LRIT
    ar_lrit_image_sink_sptr lrit_sink =
         ar_make_lrit_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/dump.dat", config.resynch_thresh, config.BERFileName, config.socket_flag, config.socket_lrit_flag);
    lrit_sink->set_name("viterbi_rx::lrit_sink");
    d_lrit=lrit_sink;
#endif
#ifdef USE_EMWIN
   ar_emwin_image_sink_sptr emwin_sink =
        ar_make_emwin_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/dump.dat", config.resynch_thresh, config.BERFileName);
    emwin_sink->set_name("viterbi_rx::emwin_sink");
    d_emwin =emwin_sink;   
#else
    #ifndef USE_LRIT
        // NULL SINK
        gr_block_sptr null_sink_c = 		gr_make_null_sink (sizeof(unsigned char));        
    #endif
#endif

///--------------------------------------------------------------------------------------
// Connect Viterbi decoder to data link layer processing 
//---------------------------------------------------------------------------------------
    #ifdef USE_LRIT
        connect_blocks(viterbi_decoder,    0, viterbi_bytes, 0, d_blocks);
        connect_blocks(viterbi_bytes,      0, lrit_sink, 0, d_blocks);
    #endif
    #ifdef USE_EMWIN
        connect_blocks(viterbi_decoder,    0, viterbi_bytes, 0, d_blocks);
        connect_blocks(viterbi_bytes,    0, emwin_sink, 0, d_blocks);
    #else
        #ifndef USE_LRIT
            connect_blocks(viterbi_decoder,    0, viterbi_bytes, 0, d_blocks);            
            connect_blocks(viterbi_bytes,      0, null_sink_c,   0, d_blocks);
        #endif
    #endif

    //-------------------------   ENABLE THROUGHPUT ?      -------------------------------
#ifdef THROUGHPUT_VITERBI
    gr_throughput_sptr throughput =
        gr_make_throughput(sizeof(gr_complex), 3, "cout", "Viterbi_rx");
    // THROUGHPUT
    connect_blocks(viterbi_decoder,  0, throughput, 0, d_blocks);
#endif

    //--------------------------  ASK EUGENE ABOUT THIS    -----------------------------
    // Append a copy of pointers to all the d_blocks created here to the
    // input list of d_blocks.  Note the funky 'back_insert_iterator'
    copy(d_blocks.begin(), d_blocks.end(),
         back_insert_iterator<vector<gr_block_sptr> >(blocks));
}

bool viterbi_rx::get_packet_lock(){
    // Slow Version
    /*
    #ifdef USE_LRIT
        ar_lrit_image_sink *LRIT_process  =
            find_block_ptr<ar_lrit_image_sink>(d_blocks, "viterbi_rx::lrit_sink");
        return LRIT_process->Synched;
    #endif
    */
#ifdef USE_LRIT
       ar_lrit_image_sink_sptr LRIT_process  =
        cast_block<ar_lrit_image_sink>(d_lrit);
    bool temp= LRIT_process->Synched;
    return LRIT_process->Synched;
    
    // return 0;
#else
    // Need to write this part
    return 1;
#endif


}


