/*
GOES Radio Comments
This file is a wrapper for the data link layer.  It calls ProcessFSKDataBlock(), which is defined in FSK_ImageProcessor.cpp
It passes two additional configuration variables: the socket port number, and the log file name. 
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_fsk_image_sink.h>
#include <gr_io_signature.h>
#include <cstdio>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdexcept>

// win32 (mingw/msvc) specific
#ifdef HAVE_IO_H
#include <io.h>
#endif

#ifdef O_BINARY
#define	OUR_O_BINARY O_BINARY
#else
#define	OUR_O_BINARY 0
#endif

// should be handled via configure
#ifdef O_LARGEFILE
#define	OUR_O_LARGEFILE	O_LARGEFILE
#else
#define	OUR_O_LARGEFILE 0
#endif

//extern bool fsk_image_initialize();
//extern int ProcessFSKDataBlock(int *ReadBuffer, int next_block_size, int flip_sign, SOCKET sockets);

//ar_fsk_image_sink::ar_fsk_image_sink(size_t itemsize, const char *filename, int use_socket_flag, FILE * BERFILE) //JR int flip sign

ar_fsk_image_sink::ar_fsk_image_sink(size_t itemsize, const char *filename, int use_socket_flag, char * BERFileName) //JR int flip sign
        : gr_sync_block ("file_sink",
                         gr_make_io_signature(1, 1, itemsize),
                         gr_make_io_signature(0, 0, 0)),
        d_itemsize(itemsize), d_fp(0), d_new_fp(0), d_updated(false), d_BERFileName(BERFileName)
{
    Initialize_FSKprocessor(use_socket_flag);
    if (!open(filename))
        throw std::runtime_error ("can't open file");

    // if(!fsk_image_initialize());  //Initialize:  //JR I don't need to initialize the processor anymore.  It crashes anyway!
    //  throw std::runtime_error ("can't initialize goes processor");
}

ar_fsk_image_sink_sptr
ar_make_fsk_image_sink (size_t itemsize, const char *filename) //JR int flip sign
{
    return ar_fsk_image_sink_sptr (new ar_fsk_image_sink (itemsize, filename, 0, NULL)); //JR flip sign
}

ar_fsk_image_sink_sptr ar_make_fsk_image_sink (size_t itemsize, const char *filename, int use_socket_flag, char * BERFileName) //JR int flip sign
{
    return ar_fsk_image_sink_sptr (new ar_fsk_image_sink (itemsize, filename, use_socket_flag, BERFileName)); //JR flip sign
}
ar_fsk_image_sink::~ar_fsk_image_sink ()
{
    close();
    Uninitialize_FSKprocessor();
    if (d_fp){
        fclose((FILE *) d_fp);
        d_fp = 0;
    }
}

bool
ar_fsk_image_sink::open(const char *filename)
{
    omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this function

    // we use the open system call to get access to the O_LARGEFILE flag.
    int fd;
    if ((fd = ::open (filename,
                      O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY, 0664)) < 0){
        perror (filename);
        return false;
    }

    if (d_new_fp){		// if we've already got a new one open, close it
        fclose((FILE *) d_new_fp);
        d_new_fp = 0;
    }

    if ((d_new_fp = fdopen (fd, "wb")) == NULL){
        perror (filename);
        ::close(fd);		// don't leak file descriptor if fdopen fails.
    }

    d_updated = true;
    return d_new_fp != 0;
}

void
ar_fsk_image_sink::close()
{
    omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this function
    if (d_new_fp){
        fclose((FILE *) d_new_fp);
        d_new_fp = 0;
    }
    d_updated = true;
}

int
ar_fsk_image_sink::work (int noutput_items,
                         gr_vector_const_void_star &input_items,
                         gr_vector_void_star &output_items)
{
    //During the first packet recieved, I am not synched, but after each subsquent packet, I am synched.
    static int     Synched          =0;
    int processed_data = 0; //This is how much data you process during the last call

    
    //EV:  what the ???   not using external input
    //int flip_sign = 1;  //Can't propogate variable down
    
    static int my_flip_sign = 1;  //Can't propogate variable down

    int *inbuf = (int *) input_items[0];
    int  nwritten = 0;

    if (d_updated){
        omni_mutex_lock	l(d_mutex);	// hold mutex for duration of this block
        if (d_fp)
            fclose((FILE *)d_fp);
        d_fp = d_new_fp;			// install new file pointer
        d_new_fp = 0;
        d_updated = false;
    }

    if (!d_fp)
        return noutput_items;		// drop output on the floor

    //printf("noutout_items = %d\n", noutput_items);
    processed_data = ProcessFSKDataBlock(inbuf, noutput_items, my_flip_sign,d_BERFileName);

    if (processed_data == 0 ){
        printf("Inverting Data 180deg. Current Phase=%i \n",180*my_flip_sign);
        my_flip_sign = (my_flip_sign==0);
    }
    

    //printf("processed_data = %d\n", processed_data);

    //printf("consumed %d data\n\n\n\n\n", processed_data);

    //ProcessFSKDataBlock(inbuf, noutput_items, 1);
    return processed_data;
}
