#pragma once
#include "gnuradio_misc.h"
#include "circular_buffer.h" // Thread-safe ring buffer

//! FSK modem configuration.  This structure contains all the variables
//! needed to define a FSK modem.
typedef struct fsk_config {
    float           Fs;         //!< Input sampling frequency [Hz]
    float           Tb;         //!< Symbol period [s]
    vector<float>   tones;      //!< Tone frequencies [Hz]
    bool            rz;         //!< Return to zero
    bool            double_sided; //!< Double sided -- real output only
    bool            aid_box; //!< Double sided -- real output only
    int             flip_sign;
    struct {
        int         length;     //!< Burst length in samples
        int         delay;      //!< Burst delay in samples
    }               burst;      //!< Define the burst behavior

    int socket_flag;
    // Frequency Acq
    float alpha,beta;
    //FILE * BERFILE;
    char * BERFileName;

    // set the defaults here
    fsk_config() : Fs(0), Tb(0), rz(false), double_sided(false)
    {
        burst.length = -1; burst.delay = 0;
    }
} fsk_config;


/*! FSK demodulator class.  This class encapsulates a number of GNU radio 
 *  blocks to implement a FSK demodulator.
 */

class fsk_rx
{
private:
    vector<gr_block_sptr> d_blocks;
    // Keep a local copy of frequently used blocks to avoid wasting time
    // in the find_block routine
    gr_block_sptr d_fsk_demod;
    gr_block_sptr d_tone3_integr;
    
public:
    //! Create the demodulator.  See documentation on config
    fsk_rx(vector<gr_block_sptr> &blocks, const fsk_config &config);
    //! Return a vector of all the blocks created in this module
    vector<gr_block_sptr> &get_blocks() {return d_blocks;}
    //! Get the first block -- the mixer
    gr_block_sptr get_input_block() 
        {return find_block(d_blocks, "fsk_rx::input_fsk");}
    //! Get the last block -- the slicer
    gr_block_sptr get_bits() 
        {return find_block(d_blocks, "fsk_rx::slicer");}
    //! See documentation for gr_fsk_demod_cf::get_signal_power
    float get_signal_power();
    //! See documentation for gr_fsk_demod_cf::get_noise_power
    float get_noise_power();
    //! Returns the average power of the 3rd tone
    float get_tone3_power();
    //! See documentation for gr_fsk_demod_cf::set_locking
    void set_locking(float threshold, int lock_after, int unlock_after);
    //! See documentation for gr_fsk_demod_cf::set_snr_averaging
    void set_snr_averaging(float snr_alpha);
    //! Get the estimated SNR in dB
    float get_snr_estimate();
};



/*! FSK modulator class.  This class encapsulates a number of 
 *  GNU radio blocks to implement a FSK modulator.  See 
 *  Documentation for gr_fsk_mod_bc for details
 */