/*
GOES Radio Comments
The fsk_rx class gets data from main_emwin_i() and goes through a low pass filter.  The filtered
data is passed through two correlators, one for tone1 and one for tone 2.  The output of these
correlates gets averaged and then the magnitude is taken. These processed correlation outputs go 
into ar_fsk_demod_cf() for extensive processing where the FSK engine finds the bit boundaries
and determines which tone has the larger value, and then decides either a 1 or a 0. The output
of the ar_fsk_demod_cf() engine converts the data into integers, then feeds this data into 
the data link layer processor ar_fsk_image_sink()

*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_throughput.h>
#include <ar_fir_filter_ccf.h>
#include <gr_file_source.h>
#include <gr_file_sink.h>
#include <gr_firdes.h>
#include <ar_snr_estimator_ff.h>
#include <ar_multistage_decimator_cc.h>
#include <gr_delay.h>
#include <string>
#include <iostream>
#include <fstream>
#include <ar_mixer_cc.h>
#include <gr_add_cc.h>
#include <gr_conjugate_cc.h>
#include <ar_integrate_dump_cc.h>
#include <gr_binary_slicer_fb.h>
#include <ar_fsk_demod_cf.h>
#include <gr_null_sink.h>
#include <gr_complex_to_xxx.h>
#include <gr_unpacked_to_packed_bb.h>
#include <ar_integrate_dump_cc.h>B
#include "ar_running_average.h"
#include "ar_fsk_image_sink.h"
#include "ar_integer_slicer.h"
#include "gnuradio.h"
#include "gnuradio_misc.h"
#include "fsk_modem.h"

//*****************************
// BLOCKS TO USE

#define THROUGHPUT
//#undef THROUGHPUT

#define SINK_ALL
//#undef SINK_ALL


//*****************************

fsk_rx::fsk_rx(vector<gr_block_sptr> &blocks, const fsk_config &config)
{
    // **********************************************************************
    // ************** Create all the signal processing blocks ***************
    // **********************************************************************
    // Current signal chain:
    // mixer -> decimator .-> mixer for tone 1 ->.
    //                    |                      | -> symbol timing -> slicer
    //                    .-> mixer for tone 2 ->.
    //                    |
    //
    //             ---------    FILTER: FRONT END       --------------
    vector<float> lowpass_taps;
    gr_fir_filter_ccf_sptr lowpass_filter;
    grab(lowpass_taps, "c:/temp/usrp_log/input_data/lowpass_FSK_Front_40taps.coef");


    //grab(lowpass_taps, "c:/temp/usrp_log/input_data/lowpass_FSK_50taps_02wide.coef");
    // Load coefs from a file
    lowpass_filter =  gr_make_fir_filter_ccf (1, lowpass_taps);
    //for (int ii=0;ii<41;ii++){
    //     printf("Filter[%i]=%4.3e \n",ii,lowpass_40taps[ii]);
    //}
    //INPUT TO THE DEMODULATOR
    lowpass_filter->set_name("fsk_rx::input_fsk");

    //--------------------------- NO DECIMATOR-------------------------
    float Fs_decim = config.Fs ;
    int input_decim_ratio = 1;
    //--------------------------- Mixers ----------------------------
    
    // Mixers for the two tones.  Note that the negative of the tone freq
    // is used.  Then, exp(j f_k t) * exp(-j f_k t) ~ constant
    gr_mixer_cc_sptr         mixer_tone1 =
        gr_make_mixer_cc (Fs_decim, -config.tones[0]);
    gr_mixer_cc_sptr         mixer_tone2 =
        gr_make_mixer_cc (Fs_decim, -config.tones[1]);


    float samples_per_symbol = Fs_decim * config.Tb;
    //
    // ------------------ Running average (integration) for tones 1 and 2
    // In RZ mode, treat each symbol as two adjacent symbols.  The first
    // symbol actually has power, the second is blank.  Also see
    // gr_fsk_demod
    int avg_length = (int)(samples_per_symbol + 0.5);  //(ROUNDING)

    gr_running_average_cc_sptr         avg_tone1 =
        gr_make_running_average_cc (avg_length);
    gr_running_average_cc_sptr         avg_tone2 =
        gr_make_running_average_cc (avg_length);
    // ------------------ Magnitude
    gr_complex_to_mag_sptr             mag_tone1 =
        gr_make_complex_to_mag (input_decim_ratio);
    gr_complex_to_mag_sptr             mag_tone2 =
        gr_make_complex_to_mag (input_decim_ratio);

    //----------------------      FSK demodulator ---------------------
    // USES symbol timing acquisition. Supports fractional
    // samples_per_symbol
    gr_fsk_demod_cf_sptr               fsk_demod =
        gr_make_fsk_demod_cf(samples_per_symbol, config.rz, config.alpha, config.beta);
    fsk_demod->set_name("fsk_rx::demod");
    d_fsk_demod = fsk_demod; // Save for later ??????????????????????

    //---------------------   Slicer to generate 1/0 output     ------------------------
    gr_binary_slicer_fb_sptr        slicer =        gr_make_binary_slicer_fb ();
    slicer->set_name("fsk_rx::slicer");
    
    //------------------------- Bits to Bytes     ------------------------------------
    gr_unpacked_to_packed_bb_sptr   slicer_bytes
    = gr_make_unpacked_to_packed_bb(1,GR_MSB_FIRST);
    slicer_bytes->set_name("fsk_rx::slicer_bytes");

    //------------------------- Integer Slicer     ------------------------------------ 
    gr_integer_slicer_sptr        slicer_integer =       gr_make_integer_slicer (); //JR
    
    //-------------------------  FSK IMAGE EXTRACTOR     ------------------------------------ 
    /*
    ar_fsk_image_sink_sptr fsk_sink 
    = ar_make_fsk_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/fsk_dump.dat", config.socket_flag, config.BERFILE);
    */
    //JR config.flip_sign
    ar_fsk_image_sink_sptr fsk_sink 
        = ar_make_fsk_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/fsk_dump.dat", config.socket_flag, config.BERFileName);

    
    // **********************************************************************
    // ************** Connect the blocks together ***************************
    // **********************************************************************
    // The syntax is self explanatory: from_block.from_port -> to_block.to_port
    //gr_block_sptr mixer_input; // cleaner code
 
    //connect_blocks(input_decim,       0, lowpass_filter,  0, d_blocks);
    connect_blocks(lowpass_filter,    0, mixer_tone1,     0, d_blocks);
    connect_blocks(lowpass_filter,    0, mixer_tone2,     0, d_blocks);

    // NO DECIMATOR
    connect_blocks(mixer_tone1,        0, avg_tone1,     0, d_blocks);  // Running Average
    connect_blocks(mixer_tone2,        0, avg_tone2,     0, d_blocks);  // Running Average

    // DETECT MAGNITUDE
    connect_blocks(avg_tone1,          0, mag_tone1,       0, d_blocks); // Magnitude
    connect_blocks(avg_tone2,          0, mag_tone2,       0, d_blocks); // Magnitude
    //DEMODULATE
    connect_blocks(mag_tone1,          0, fsk_demod,       0, d_blocks);
    connect_blocks(mag_tone2,          0, fsk_demod,       1, d_blocks);
    //Slice & Demodulate 
    connect_blocks(fsk_demod,          0, slicer_integer,   0, d_blocks);
    connect_blocks(slicer_integer,     0, fsk_sink,         0, d_blocks);   



    
#ifdef THROUGHPUT
    gr_throughput_sptr throughput =
        gr_make_throughput(sizeof(gr_complex), 3, "cout", "fsk_rx");
    //connect_blocks(lowpass_filter,  0, throughput, 0, d_blocks);    // 9.6ksps*SamplesPerSec
    //connect_blocks(avg_tone1,  0, throughput, 0, d_blocks);     // 9.6ksps*SamplesPerSec
    //connect_blocks(mag_tone1,  0, throughput, 0, d_blocks);    // 9.6ksps*SamplesPerSec
    connect_blocks(fsk_demod,  0, throughput, 0, d_blocks);
    //connect_blocks(slicer,  0, throughput, 0, d_blocks);     // 9.6 ksps / 8 bits/s = 1.2 ksps
    //connect_blocks(slicer_bytes,  0, throughput, 0, d_blocks);     // 9.6 ksps / 8 bits/s = 1.2 ksps
#endif

    //--------------------------------------------------------------
    //                       SINK IT !
    //------------------------ debug and monitor points


#ifdef SINK_ALL
     // Low pass Filter
  /*  
    gr_file_sink_sptr        filter_sink =
         gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/fsk_filter_out.dat");
     connect_blocks(lowpass_filter,     0, filter_sink,     0, d_blocks);
     

     gr_file_sink_sptr        mixer_tone1_sink =
         gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/fsk_mixer_tone1.dat");
     connect_blocks(mixer_tone1,        0, mixer_tone1_sink,     0, d_blocks);
     
     gr_file_sink_sptr        mixer_tone2_sink =
         gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/fsk_mixer_tone2.dat");
     connect_blocks(mixer_tone2,        0, mixer_tone2_sink,     0, d_blocks);
    
     gr_file_sink_sptr        avg1_sink =
         gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/fsk_avg1.dat");
     connect_blocks(avg_tone1,          0, avg1_sink,            0, d_blocks);
      
     gr_file_sink_sptr        avg2_sink =
         gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/fsk_avg2.dat");
     connect_blocks(avg_tone2,          0, avg2_sink,            0, d_blocks);   
     // MAG SINK
    gr_file_sink_sptr        mag1_sink =
        gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/fsk_mag1.dat");
    connect_blocks(mag_tone1,          0, mag1_sink,            0, d_blocks);

    gr_file_sink_sptr        mag2_sink =
      gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/fsk_mag2.dat");
    connect_blocks(mag_tone2,          0, mag2_sink,            0, d_blocks);

    // SOFT DECISIONS 
    gr_file_sink_sptr        demod_sink = //JR
        gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/fsk_demod.dat"); //JR
    connect_blocks(fsk_demod,          0, demod_sink,           0, d_blocks); //JR

    // DATA SINKS
    gr_file_sink_sptr        slicer_sink =
        gr_make_file_sink (sizeof(unsigned char), "c:/temp/usrp_log/fsk_slicer.dat");
    connect_blocks(slicer,             0, slicer_sink,          0, d_blocks);

    
    gr_file_sink_sptr        slicer_bytes_sink =
        gr_make_file_sink (sizeof(unsigned char), "c:/temp/usrp_log/fsk_slicer_bytes.dat");
    connect_blocks(slicer_bytes,       0, slicer_bytes_sink,    0, d_blocks);
    */

    // DECIMATOR SINK
    //connect_blocks(decim_tone1,        0, decim1_sink,     0, d_blocks);
    // connect_blocks(decim_tone2,        0, decim2_sink,     0, d_blocks);
#endif

   // Append a copy of pointers to all the d_blocks created here to the
    // input list of d_blocks.  Note the funky 'back_insert_iterator'
    copy(d_blocks.begin(), d_blocks.end(),
         back_insert_iterator<vector<gr_block_sptr> >(blocks));

}

float fsk_rx::get_signal_power()
{
    gr_fsk_demod_cf_sptr fsk_demod =
        cast_block<gr_fsk_demod_cf>(d_fsk_demod);
    return fsk_demod->get_signal_power();
}
float fsk_rx::get_noise_power()
{
    gr_fsk_demod_cf_sptr fsk_demod =
        cast_block<gr_fsk_demod_cf>(d_fsk_demod);
    return fsk_demod->get_noise_power();
}

float fsk_rx::get_tone3_power()
{
    gr_integrate_dump_cc_sptr tone3_integr =
        cast_block<gr_integrate_dump_cc>(d_tone3_integr);
    return tone3_integr->get_mean_pwr();
}

void fsk_rx::set_locking(float threshold, int lock_after, int unlock_after)
{
    gr_fsk_demod_cf *fsk_demod =
        find_block_ptr<gr_fsk_demod_cf>(d_blocks, "fsk_rx::demod");
    fsk_demod->set_locking(threshold, lock_after, unlock_after);
}

void fsk_rx::set_snr_averaging(float snr_alpha)
{
    gr_fsk_demod_cf *fsk_demod =
        find_block_ptr<gr_fsk_demod_cf>(d_blocks, "fsk_rx::demod");
    fsk_demod->set_snr_averaging(snr_alpha);
}

/*
float fsk_rx::get_snr_estimate()
{
    gr_snr_estimator_ff *snr_estimator =
        find_block_ptr<gr_snr_estimator_ff>(d_blocks, "fsk_rx::snr_estimator");
    return snr_estimator->get_snr_estimate();
}
*/