/*
GOES Radio Comments
This is the oqpsk_rx demodulator class.  This class first does the physical layer demodulation
and then does viterbi decoding, and finally it does data link layer processing. 

The first block from main_emwin_n() goes into an agc block. After going through AGC, it is 
decimated from its current samples/symbol rate to a desired_sample_per_symbols rate, which is 
typically 2 samples/symbol.  AFter that, it goes through a carrier recovery and tracking loop (costas)
. After that, the I channel is delayed by T/2 to convert OQPSK to QPSK.  From there it is Match 
Filtered, and then sent through a Muller & Mueller clock recovery and tracking loop.  From there
it goes through a form of data conversion so it is compatible with the viterbi decoder, and then 
it is Viterbi decoded.  From there it goes through a binary differential decoder. 

The output of the differential decoder is in the form of shorts. Each short represents one bit, and 
8 bits are packeged into one bytes via the (unpacked_to_packed) conversion block.  These bytes are 
fed into the EMWIN data link layer for final processing. 

The carrier and clock recovery and tracking loops are set using the functions set_phase_bandwidth()
and set_timing_bandwidth respectively. 
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
//AERO RADIO
#include <ar_fractional_interpolator_cc.h>
#include <ar_clock_recovery_mm_cc.h>
#include <ar_mixer_cc.h>
#include <ar_short_to_uchar.h>
#include <ar_snr_estimator_cf.h>
#include <tcp_debug_sink.h>
#include "ar_viterbi.h"
#include "oqpsk_modem.h"
//GNU RADIO
#include <ar_sig_source_c.h>
#include <gr_file_sink.h>
#include <gr_firdes.h>
#include <ar_costas_loop_cc.h>
#include <ar_fir_filter_ccf.h>
#include <ar_throughput.h>
#include <ar_complex_to_xxx.h>
#include <ar_float_to_complex.h>
//#include <gr_binary_slicer_fb.h>
#include <gr_agc_cc.h>
#include "gnuradio.h"
#include "gnuradio_misc.h"
#include <gr_file_sink.h>
#include <gr_null_sink.h>
#include <gr_delay.h>
#include <gr_diff_decoder_bb.h>
#include <gr_unpacked_to_packed_bb.h>
#include <ar_multiply_const_cc.h>
// MultiThread
// RING BUFFER
#include <ar_ring_buffer_source2.h>
#include <ar_ring_buffer_sink2.h>
#include "ar_emwin_image_sink.h"


using namespace std;

//*****************************
// BLOCKS TO USE
// NOT USED   #define USE_EMWIN
#define THROUGHPUT
//#define USE_SNR_ESTIMATOR
//#define SINK_ALL
#define ALLOW_ROTATION
// MULTITHREAD VERSION
oqpsk_rx::oqpsk_rx(vector<gr_block_sptr> &blocks, const oqpsk_config &config)
{
    d_config = config;
    // **********************************************************************
    // ************** Create all the signal processing blocks ***************
    // **********************************************************************
    // Current signal chain:
    // [source I&Q ] -> mixer to baseband  -> agc -> Carrier ACQ  --> (1)
    // (1)-matched_filter -> -> [I T/2 Delay] --> clock_recovery -> slicer
    //
    //
    //// **********************************************************************

    // This creates a software decimator to push the rate down to "desired_samples_per_symbol"
    // which is usually 2 samples/symbol.  This decimation process happens in 3 stages.  
    // Decimation example:
    // Let the input signal be 63 samples/symbol, and the desired_samples_per_symbol be 10
    // The integer decimation is split into two phases. First, 63 is decimated by 6 in two
    // stages.  decim1 is set to 3, decim2 is set to 2, thus 63/decim1/decim2 = 10.5
    // A resampler is used as the last stage of decimation, where 10.5 is pulled down to 10 even
    // The resampler rate would be 1.05.  

    //samples per sybmoby 6.3 samples/symbol 
    float decim_ratio	= config.Fs / (config.bit_rate * config.samples_per_symbol_decim); 
    int decim1_ratio	= sqrt(decim_ratio);
    int decim2_ratio	= decim_ratio / decim1_ratio;
    int decim_ratio_int = decim1_ratio * decim2_ratio;

    std::vector<float> decim1_taps = // First decimation stage 
        gr_firdes::low_pass(1.0, 1.0, 0.5 / decim1_ratio, 0.1);
	dump(decim1_taps, "c:/temp/usrp_log/oqpsk_decim1_taps.coef");

   	gr_fir_filter_ccf_sptr decim1 = 
        gr_make_fir_filter_ccf (decim1_ratio, decim1_taps);

    std::vector<float> decim2_taps = // Second decimation stage
        gr_firdes::low_pass(1.0, 1.0, 0.5 / decim2_ratio, 0.1);
    dump(decim2_taps, "c:/temp/usrp_log/oqpsk_decim2_taps.coef");

   	gr_fir_filter_ccf_sptr decim2 = 
        gr_make_fir_filter_ccf (decim2_ratio, decim2_taps);

    // The integer decimators above are followed by a fractional decimator
    // to get the sample rate very close to 4x symbol rate.
	float decim_ratio_frac = decim_ratio / decim_ratio_int;

   	ar_fractional_interpolator_cc_sptr resampler = 
		ar_make_fractional_interpolator_cc(0.0, decim_ratio_frac);

    // printf("\n\n Decimation Ratio = %g\n Sampled Samples Per Symbol = %g\n Decimated Samples Per Symbol = %d\n Fractional Decimation Ratio = %g\n Decimation 1 = %d\n Decimation 2 = %d\n\n", decim_ratio, config.samples_per_symbol, config.samples_per_symbol_decim, decim_ratio_frac, decim1_ratio, decim2_ratio);

    // --------------------- AGC --------------------------------------------
    // The AGC is used to present a constant magnitude into the costas loop
    // to keep the loop gain constant.
    float agc_rate			= 1e-5;
    gr_agc_cc_sptr agc		= gr_make_agc_cc (agc_rate, config.AGC);
    agc->set_name("oqpsk_rx::agc");

    //---------------------- OQPSK: Costas Loop phase tracking ------------
    //QPSK
    gr_costas_loop_cc_sptr phase_track;
	phase_track = gr_make_costas_loop_cc (0, 0, 0.1, -0.1, 5); 
    phase_track->set_name("oqpsk_rx::phase_track");

    // This is in case we use a PLL ???? Disable for Costas Loop
    //phase_track->set_lock_threshold(50);

    // --> Bandwidth must be set after blocks are connected <<--
    //
    //---------------------- OQPSK: DELAY REAL COMPONENT to Create QPSK ------------
    // Complex to Real
    gr_complex_to_float_sptr      split_IQ = gr_make_complex_to_float (1);
    split_IQ->set_name("oqpsk_rx::split_IQ");

    gr_delay_sptr           delay_real,delay_real_1;
    delay_real = gr_make_delay( sizeof(float), int(config.samples_per_symbol_decim/2) );

    gr_float_to_complex_sptr      combined_IQ = gr_make_float_to_complex();
    //---------------------- Matched Filter -----------------------
    // Shaping filter: Root raised cosine or Rectangle
    std::vector<float> shaping_taps;
    if (config.shaping_rolloff < 0.99){
        shaping_taps =					// gain, sampling freq, symbol rate, alpha, ntaps
            gr_firdes::root_raised_cosine (1.0, config.samples_per_symbol_decim, 1.0, config.shaping_rolloff, 65);
          
    }else{// All ones (normalized to gain of 1)
        shaping_taps.resize(config.samples_per_symbol_decim, 1/config.samples_per_symbol_decim);
    }
    gr_fir_filter_ccf_sptr match_filt = gr_make_fir_filter_ccf (1, shaping_taps);

    //
    //---------------------- OQPSK: M&M timing tracking ------------------
    float timing_omega			= config.samples_per_symbol_decim; // Nominal is 4 samples/symbol
    float timing_omega_limit	= 0.01;

    ar_clock_recovery_mm_cc_sptr       timing_track = // Set bandwidth later
        ar_make_clock_recovery_mm_cc(timing_omega, 0, 0, 0, timing_omega_limit);
    timing_track->set_name("oqpsk_rx::timing_track");
    //---------------------- OQPSK: Consider all possible phase rotations ------------------ 
    #ifdef ALLOW_ROTATION
    gr_complex  QPSK_shift;      
    if (config.qpsk_shift==1){       
            printf( " \t 90 PHASE SHIFT \n\n");
            QPSK_shift=gr_complex(0.0,1.0);         // + 90            
    }else if (config.qpsk_shift==2){          
            printf( " \t 180 PHASE SHIFT \n\n");
            QPSK_shift=gr_complex(-1.0,0.0);         // + 180            
    }else if (config.qpsk_shift==3){          
            printf( " \t 270 PHASE SHIFT \n\n");
            QPSK_shift=gr_complex(0.0,-1.0);         // + 270
    }else{  
            printf( " \t 0 PHASE SHIFT \n\n"); 
            QPSK_shift=gr_complex(1.0,0.0);         // 0                
    }
    ar_multiply_const_cc_sptr    IQ_shift = ar_make_multiply_const_cc(QPSK_shift);  
    #endif
    //----------------------   INTERLEAVER      ------------------------------------    
    gr_complex_to_interleave_float_sptr      Interleave_QPSK;
	Interleave_QPSK    = gr_make_complex_to_interleave_float ();
    //---------------------------   VITERBI DECODER         --------------------------
    gr_viterbi_sptr     viterbi_decoder = gr_make_viterbi(4,config.resynch_period, config.AGC, config.vthresh);
    viterbi_decoder->set_name("viterbi_rx::input");
    //---------------------------   DIFFERENTIAL DECODER         --------------------------
    gr_diff_decoder_bb_sptr  diff_decoder   =gr_make_diff_decoder_bb(2);
       diff_decoder->set_name("oqpsk_rx::diff_decoder");
    //------------------------- Bits to Bytes     ------------------------------------
    gr_unpacked_to_packed_bb_sptr   viterbi_bytes     
        = gr_make_unpacked_to_packed_bb(1,GR_MSB_FIRST);
    viterbi_bytes->set_name("viterbi_rx::viterbi_bytes");    

    //--------------------   EMWIN data link layer processing      ------------------------------------------
    ar_emwin_image_sink_sptr   emwin_sink =
       ar_make_emwin_image_sink(sizeof(unsigned char), "c:/temp/usrp_log/emwin_dump.dat",config.resynch_thresh, config.BERFileName, 
       config.socket_flag,-1,0);

    gr_block_sptr            null_sink = 
        gr_make_null_sink (sizeof(unsigned char));   

    // **********************************************************************
    // ************** Connect the blocks together ***************************
    // **********************************************************************
    // The syntax is self explanatory: from_block.from_port -> to_block.to_port
    //--------------------------------------------------------------

    //----------------------        CONNECT BLOCKS          ----------------------------
   
     connect_blocks(agc,		0, decim1,		0, d_blocks);      
     connect_blocks(decim1,		0, decim2,		0, d_blocks); 
     connect_blocks(decim2,        0, resampler,	     0, d_blocks); 
     connect_blocks(resampler,           0, phase_track,     0, d_blocks);   
    connect_blocks(phase_track,        0, split_IQ,             0, d_blocks);     //JR uncomment
    connect_blocks(split_IQ,           0, delay_real,           0, d_blocks);   
    connect_blocks(delay_real,         0, combined_IQ,          0, d_blocks);
    connect_blocks(split_IQ,           1, combined_IQ,          1, d_blocks);
    connect_blocks(combined_IQ,        0, match_filt,           0, d_blocks);    
    connect_blocks(match_filt,         0, timing_track,         0, d_blocks);  

    #ifdef ALLOW_ROTATION
        connect_blocks(timing_track,        0, IQ_shift,            0, d_blocks);
        connect_blocks(IQ_shift,            0, Interleave_QPSK,	    0, d_blocks);  
    #else
        connect_blocks(timing_track,        0, Interleave_QPSK,            0, d_blocks);        
    #endif
    connect_blocks(Interleave_QPSK,     0, viterbi_decoder,     0, d_blocks);
	connect_blocks(viterbi_decoder,     0, diff_decoder,        0, d_blocks);
    connect_blocks(diff_decoder,        0, viterbi_bytes,       0, d_blocks);
    connect_blocks(viterbi_bytes,       0, emwin_sink,          0, d_blocks); 

  // -----------------------  DIAGNOSTICS -----------------------------
#ifdef THROUGHPUT
     gr_throughput_sptr throughput =
        gr_make_throughput(sizeof(gr_complex), 3, "cout", "OQPSK_RX");
    connect_blocks(timing_track,       0, throughput, 0, d_blocks);
#endif


    //-------------------------         Write to FIle       ------------------------------------
   gr_file_sink_sptr timing_sink;
   if(config.sink_data){
        timing_sink = gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_tracked.dat");
        connect_blocks(timing_track,        0, timing_sink,    0, d_blocks);
    }

#ifdef SINK_ALL
     
    /*gr_file_sink_sptr agc_sink =
       gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/agc.dat");
      connect_blocks(agc,        0, agc_sink,    0, d_blocks);*/

    // gr_file_sink_sptr combined_IQ_sink =
    //    gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/combined_IQ.dat");
    // connect_blocks(combined_IQ,        0, combined_IQ_sink,    0, d_blocks);

    //gr_file_sink_sptr match_filt_sink =
    //   gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/match_filter.dat");
    //connect_blocks(match_filt,        0, match_filt_sink,    0, d_blocks);

    /* gr_file_sink_sptr timing_track_sink =
       gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_tracked.dat");
    connect_blocks(timing_track,        0, timing_track_sink,    0, d_blocks);

           gr_file_sink_sptr phase_err_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/phase_error.dat");

       gr_file_sink_sptr timing_err_sink =
        gr_make_file_sink (sizeof(gr_complex), "c:/temp/usrp_log/timing_error.dat");

    gr_file_sink_sptr Interleave_sink =
        gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/interleave_QPSK.dat");
    connect_blocks(Interleave_QPSK,     0,  Interleave_sink ,   0, d_blocks);
 
    gr_file_sink_sptr viterbi_SINK_bits =
            gr_make_file_sink (sizeof(unsigned char), "c:/temp/usrp_log/viterbi_out_bits.dat");      
   connect_blocks(viterbi_decoder,     0,  viterbi_SINK_bits ,   0, d_blocks);

   gr_file_sink_sptr Diff_sink =
        gr_make_file_sink (sizeof(unsigned char), "c:/temp/usrp_log/Diff_QPSK.dat");
    connect_blocks(diff_decoder,     0,  Diff_sink ,   0, d_blocks);

    gr_file_sink_sptr viterbi_sink_bytes =
            gr_make_file_sink (sizeof(unsigned char), "c:/temp/usrp_log/viterbi_out_bytes.dat");
    connect_blocks(viterbi_bytes,     0,  viterbi_sink_bytes ,   0, d_blocks);*/



#endif


//-----------------------SNR -------------------------------------
#ifdef USE_SNR_ESTIMATOR
    vector<gr_complex> symbol_table(4);
    symbol_table[0] = gr_complex(  10.0,  10.0);
    symbol_table[1] = gr_complex( -10.0,  10.0);
    symbol_table[2] = gr_complex(  10.0, -10.0);
    symbol_table[3] = gr_complex( -10.0, -10.0);
    gr_snr_estimator_cf_sptr snr_estimator = gr_make_snr_estimator_cf(symbol_table);
    snr_estimator->set_name("oqpsk_rx::snr_estimator_cf");
    snr_estimator->set_averaging(0.001);

    connect_blocks(timing_track, 0, snr_estimator, 0, d_blocks);
    gr_file_sink_sptr snr_sink =
        gr_make_file_sink (sizeof(float), "c:/temp/usrp_log/snr_est.dat");
    connect_blocks(snr_estimator, 0, snr_sink, 0, d_blocks);
#endif
 
    //------------------------------------------------------------------------
    // COSTAS LOOP (cont.)
    // Bandwidth must be set after blocks are connected.  Note that we rely
    // on the front end PLL for the actual frequency offset/doppler tracking.

    // The costas loop is for phase correction only.
    set_phase_bandwidth(config.bit_rate / config.PBW);  //JR uncomment
    //set_phase_bandwidth(0); //JR Debug S-curve
    set_timing_bandwidth(config.bit_rate / config.TBW);
    //------------------------------------------------------------------------

    // Append a copy of pointers to all the d_blocks created here to the
    // input list of d_blocks.  Note the funky 'back_insert_iterator'

    copy(d_blocks.begin(), d_blocks.end(),
         back_insert_iterator<vector<gr_block_sptr> >(blocks));
}


bool oqpsk_rx::get_phase_lock(){
    gr_costas_loop_cc *phase_track =
        find_block_ptr<gr_costas_loop_cc>(d_blocks, "oqpsk_rx::phase_track");
    return phase_track->lock_detector();
}

void oqpsk_rx::set_phase_bandwidth(float bw){
    gr_costas_loop_cc *phase_track =
        find_block_ptr<gr_costas_loop_cc>(d_blocks, "oqpsk_rx::phase_track");

    // Compute the loop coefficients.
    // The phase detector for BPSK is simply e=Re[x] * Im[x].  For
    // angle, z, and a signal magnitude A, e = A Cos[z] * A Sin[z]
    // For small z, e = A * A * z = A^2*z.  The AGC sets the signal
    // magnitude such that Mean[Sqrt[Abs[x]]] = ref level.  The loop
    // gain is then G(z) = e(z) / z = (ref level)^2
   // gr_agc_cc *agc = find_block_ptr<gr_agc_cc>(d_blocks, "oqpsk_rx::agc");
    //float K_ped = agc->reference() * agc->reference()*.5;
     float K_ped; 
    K_ped = 15;  //JR DEBUG, Scurve = 5 @ 4 dB, 15 @ 300 dB
    // See a document titled "2nd order tracking loop analysis" for details
    // Also see carrier_loop_coefficients.m
    float S      = 1.0 / sqrt(2.0); // damping factor
    float SS     = (S+1/(4*S));

    float     K_loop = K_ped ;
    float     Fs = d_config.bit_rate * d_config.samples_per_symbol; // Running at 4 samples/symbol

    float     alpha = 4 * bw * S  / (K_loop * Fs * SS);
    float     beta  = 4 * bw * bw / (K_loop * Fs * Fs * SS * SS);

    phase_track->set_loop_coefs(alpha, beta);
    //alpha =2.6655500e-005
    //beta  =3.5525787e-008
}

void oqpsk_rx::set_timing_bandwidth(float bw){
    ar_clock_recovery_mm_cc *timing_track =
        find_block_ptr<ar_clock_recovery_mm_cc>(d_blocks, "oqpsk_rx::timing_track");

    // Compute the loop coefficients.
    // The phase detector for M&M is
    // Re[ (Sign[x(n-2)] - Sign[x(n)]) * Conj[x(n-1)] ]
    // The loop gain is therefore proportional to the amplitude of the signal.
    // The actual loop gain is nontrivial to derive (see e.g. Mengali and
    // d'Andrea p. 386).  The AGC sets the signal magnitude such that
    // Mean[Sqrt[Abs[x]]] = ref level.  The loop gain is then = ref level
  
     gr_agc_cc *agc = find_block_ptr<gr_agc_cc>(d_blocks, "oqpsk_rx::agc");  //JR DEBUG
    float K_ped = agc->reference() * 0.35; // 0.35 is a fudge factor  

    //float K_ped = 3.5; //JR DEBUG

    // See a document titled "2nd order tracking loop analysis" for details
    // Also see carrier_loop_coefficients.m
    float S      = 1.0 / sqrt(2.0); // damping factor
    float SS     = (S+1/(4*S));

    float K_loop = K_ped ;
    float  Fs = d_config.bit_rate ; // Update once per symbol

    float alpha = 4 * bw * S  / (K_loop * Fs * SS);
    float beta  = 4 * bw * bw / (K_loop * Fs * Fs * SS * SS);

    timing_track->set_gain_mu (alpha); // 1st order: alpha
    timing_track->set_gain_omega (beta); //2nd order: beta

}


