#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_check_lfsr15_b.h>
#include <gr_io_signature.h>
#include <stdlib.h>
#include <stdio.h>

int gr_check_lfsr15_b::LFSR_BER_LENGTH = (1<<16);
int gr_check_lfsr15_b::LFSR_LENGTH     = 15;

gr_check_lfsr15_b_sptr
gr_make_check_lfsr15_b ()
{
  return gr_check_lfsr15_b_sptr (new gr_check_lfsr15_b ());
}

gr_check_lfsr15_b::gr_check_lfsr15_b ()
  : gr_sync_block ("gr_check_lfsr15_b",
		   gr_make_io_signature (1, 1, sizeof (char)),
		   gr_make_io_signature (0, 0, 0)),
    d_state(LOAD_LFSR), d_load_count(0)
{
}

int
gr_check_lfsr15_b::work (int noutput_items,
			   gr_vector_const_void_star &input_items,
			   gr_vector_void_star &output_items)
{
  char *in = (char *) input_items[0];

  for (int i = 0; i < noutput_items; i++){
    char x = 1 -(in[i] & 0x1);  // For some reason gr_glfsr_source is inverted
    char expected;
    
    switch (d_state){

    case LOAD_LFSR: // Prime the LFSR with received bits
		if (d_load_count < LFSR_LENGTH)
		{
			d_lfsr <<= 1;
			d_lfsr |= x;
			d_load_count++;
		} else {
			d_state = CHECK_BER;
			d_load_count = 0;
			d_total_bits = 0; d_total_errs = 0;
			d_bits = 0; d_errs = 0;
			// The next bit has a latency of 1 cycle.  Prepare by computing
			// one extra bit here.
			next_lfsr_output();
		}
      break;
      
	// Check the BER by computing the number of errors over the last
	// LFSR_BER_LENGTH bits.  If the BER > 0.25, we probably lost 
	// lock and will attempt to re-acquire
	case CHECK_BER: 
		expected = next_lfsr_output();
		d_bits++;
		if (x != expected) d_errs++;

		if (d_bits == LFSR_BER_LENGTH) 
		{
			double ber = (double)d_errs / (double)d_bits;

			d_total_bits += d_bits; d_total_errs += d_errs;
			double total_ber = d_total_errs / d_total_bits;
			printf("CheckLFSR15: Last BER = %3.3g.  Avg BER = %3.3g [%g bits]\n",
				ber, total_ber, d_total_bits);

			d_bits = 0; d_errs = 0;
			
			if (ber > 0.25)
			{
				printf("CheckLFSR15: Lost lock");
				d_state = LOAD_LFSR;
			}
		}
		break;
	}
  }

  return noutput_items;
}

  // Compute the next bit output based on the current state and update the
  // state
char
gr_check_lfsr15_b::next_lfsr_output()
{
	int tmp;
	tmp = (d_lfsr >> 14) ^ (d_lfsr >> 0);
	tmp &= 0x1;
	d_lfsr <<= 1; // shift left by one
	d_lfsr |= tmp;
	return tmp;
}