
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_fir_filter_ccf.h>
#include <gr_fir_ccf.h>
#include <gr_fir_util.h>
#include <gr_io_signature.h>

gr_fir_filter_ccf_sptr gr_make_fir_filter_ccf (int decimation, const std::vector<float> &taps)
{
  return gr_fir_filter_ccf_sptr (new gr_fir_filter_ccf (decimation, taps));
}

#define MAX_BLOCK_SIZE      (1<<15)
gr_fir_filter_ccf::gr_fir_filter_ccf (int decimation, const std::vector<float> &taps)
  : gr_sync_decimator ("fir_filter_ccf",
		       gr_make_io_signature (1, 1, sizeof (gr_complex)),
		       gr_make_io_signature (1, 1, sizeof (gr_complex)),
		       decimation),
    d_updated (false), d_filt_re(0), d_filt_im(0)
{
    d_buf_re = ippsMalloc_32f(MAX_BLOCK_SIZE);
    d_buf_im = ippsMalloc_32f(MAX_BLOCK_SIZE);

    create_filters(taps);
}

gr_fir_filter_ccf::~gr_fir_filter_ccf ()
{
    ippsFIRFree_32f(d_filt_re);
    ippsFIRFree_32f(d_filt_im);
    ippsFree(d_buf_re);
    ippsFree(d_buf_im);

}

void
gr_fir_filter_ccf::create_filters (const std::vector<float> &taps)
{
    // Delete filters if they already exist (changing taps)
    if (d_filt_re) ippsFIRFree_32f(d_filt_re);
    if (d_filt_im) ippsFIRFree_32f(d_filt_im);

    if (decimation() == 1) // single rate
    {
        ippsFIRInitAlloc_32f (&d_filt_re, &taps[0], taps.size(), 0);
        ippsFIRInitAlloc_32f (&d_filt_im, &taps[0], taps.size(), 0);
    } else {
        ippsFIRMRInitAlloc_32f(&d_filt_re, &taps[0], taps.size(), 
                                        1, 0, decimation(), 0, 0);
        ippsFIRMRInitAlloc_32f(&d_filt_im, &taps[0], taps.size(), 
                                        1, 0, decimation(), 0, 0);
    }
}

void
gr_fir_filter_ccf::set_taps (const std::vector<float> &taps)
{
  // Don't actually change the filters here.  This function may be
  // called from another thread, while the 'work' thread is already
  // processing data using the filters.  Make sure to only update
  // the filters in the 'work' function, called from the 'work' thread.
  d_new_taps = taps;
  d_updated = true;
}

int
gr_fir_filter_ccf::work (int noutput_items,
		   gr_vector_const_void_star &input_items,
		   gr_vector_void_star &output_items)
{
    gr_complex *in = (gr_complex *) input_items[0];
    gr_complex *out = (gr_complex *) output_items[0];

    if (d_updated) { // See comment in set_taps
        d_updated = false;
        create_filters(d_new_taps);
    }

    int ninput_items = noutput_items * decimation();
    
    // Convert the complex input into two real streams
    ippsCplxToReal_32fc((Ipp32fc *)in, d_buf_re, d_buf_im, ninput_items);
    // Compute the FIR
    ippsFIR_32f_I(d_buf_re, noutput_items, d_filt_re);
    ippsFIR_32f_I(d_buf_im, noutput_items, d_filt_im);
    // Copy the (decimated) input to the output buffer, converting back to 
    // complex
    ippsRealToCplx_32f(d_buf_re, d_buf_im, (Ipp32fc *)out, noutput_items);

  return noutput_items;
}
