#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_freq_acq.h>
#include <algorithm>
#include <gr_io_signature.h>
#include <ipp.h>
#include <iostream>
#include <fstream>

//This is the old method
//#define SAVE_STREAM
//This is the new method. Comment both to avoid data loggin

//#define NEW_SAVE_STREAM

gr_freq_acq::gr_freq_acq (size_t fft_size)
  : gr_sync_block ("freq_acq",
		   gr_make_io_signature (1, 1, sizeof(float)*fft_size),
		   gr_make_io_signature (0, 0, 0)),
    d_fft_size(fft_size), 
    d_fft_result(fft_size, 0.0), 
    d_result_ready(false)
{
}

gr_freq_acq_sptr
gr_make_freq_acq (size_t fft_size)
{
  return gr_freq_acq_sptr (new gr_freq_acq (fft_size));
}

int
gr_freq_acq::work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items)
{
  float *in = (float *) input_items[0];

  for (int i=0; i < noutput_items; i++)
  {
    // Don't modify the result vector if the get_freqs function is being
    // called.
    d_mutex.lock(); 
    d_fft_result = vector<float>(in, in + d_fft_size);
    d_result_ready = true;
    d_mutex.unlock(); 
  }

  return noutput_items;
}

// Fancy way to sort such that we get the index
template<class T> struct index_cmp {
    index_cmp(const T arr) : arr(arr) {}
    bool operator()(const size_t a, const size_t b) const
        { return arr[a] > arr[b]; } // Descending order
    const T arr;
};

bool gr_freq_acq::get_freqs(int num_peaks, vector<float> &freqs, vector<float> &powers)
{
    // Make sure we have at least the first result ready
    if (!d_result_ready) return false;

    freqs.resize(num_peaks);
    powers.resize(num_peaks);

    vector<int> results_idx(d_fft_size);
    for (int i = 0; i<d_fft_size; ++i) results_idx[i] = i;

    // Make sure the d_fft_results does not change while we are computing
    // the largest peaks.
    d_mutex.lock(); 

    d_result_ready = false; // wait until we get a new result after this one

#ifdef SAVE_STREAM
    ofstream debug("c:/temp/usrp_log/acq_debug.dat");
    copy (d_fft_result.begin(), d_fft_result.end(), ostream_iterator<float>(debug," "));
    debug << "\n";
#endif

    sort(results_idx.begin(), results_idx.end(), index_cmp<vector<float>&>(d_fft_result));
#ifdef SAVE_STREAM
    copy (results_idx.begin(), results_idx.end(), ostream_iterator<float>(debug," "));
#endif

#ifdef NEW_SAVE_STREAM
    // Plain C: EV !!!!
    char *DataFileName ;    // Log File for  results
	DataFileName =(char *) calloc(200,sizeof(char));
	strcpy(DataFileName,"c:/temp/usrp_log/acq_debug.dat" );
	FILE			* pFile;
    pFile = fopen (DataFileName, "wt" );
	if (pFile==NULL) {
		cout << "Unable to open file" << endl;
		cout << " File Name = \t "<< DataFileName << endl;
		exit (1);
	}

	for(  int i =0; i < d_fft_size ; i++){			
		fprintf(pFile,"%6.3e \n", d_fft_result[i]);
	}

	fclose (pFile);
#endif

    // Compute the average point the the non-peak bins
    float noise_est = 0.0;
    for (int i = num_peaks; i<d_fft_size; i++)
        noise_est += d_fft_result[results_idx[i]];

    noise_est /= (d_fft_size - num_peaks); // average power

    // Set the results vector to the ratio between the top peaks and
    // the noise estimate;
    for (int i = 0; i < num_peaks; i++)
    {
        powers[i] = d_fft_result[results_idx[i]] / noise_est;
        freqs[i] = results_idx[i] ;
    }

    d_mutex.unlock();
    
    return true;
}

#define PEAK_WIDTH 8
bool gr_freq_acq::get_peak_freq(float *freq, float *power)
{
    // Make sure we have at least the first result ready
    if (!d_result_ready) return false;

    // Make sure the d_fft_results does not change while we are computing
    // the largest peaks.
    d_mutex.lock(); 

#ifdef SAVE_STREAM
    ofstream debug("c:/temp/usrp_log/acq_debug.dat");
    copy (d_fft_result.begin(), d_fft_result.end(), ostream_iterator<float>(debug," "));
    debug << "\n";
#endif
#ifdef NEW_SAVE_STREAM
    // Plain C: EV !!!!
    char *DataFileName ;    // Log File for  results
	DataFileName =(char *) calloc(200,sizeof(char));
	strcpy(DataFileName,"c:/temp/usrp_log/acq_debug.dat" );
	FILE			* pFile;
    pFile = fopen (DataFileName, "wt" );
	if (pFile==NULL) {
		cout << "Unable to open file" << endl;
		cout << " File Name = \t "<< DataFileName << endl;
		exit (1);
	}

	for(  int i =0; i < d_fft_size ; i++){			
		fprintf(pFile,"%6.3e \n", d_fft_result[i]);
	}

	fclose (pFile);

#endif
    // Find the index of the largest peak
    int max_idx = max_element(d_fft_result.begin(), d_fft_result.end()) 
                  - d_fft_result.begin();

    // Compute center of mass for the peak defined by the PEAK_WIDTH points
    int left_idx = max_idx - PEAK_WIDTH/2; 
    if (left_idx < 0) left_idx =0; // no wraparound.  Fix later?
    int right_idx = max_idx + PEAK_WIDTH/2;
    if (right_idx > d_fft_size - 1) right_idx = d_fft_size-1; // no wraparound.  Fix later?

    float total_power = 0.0;
    for (int i=left_idx; i<right_idx; i++) total_power += d_fft_result[i];
    
    long double mean_idx = 0.0;
    
    for (int i=left_idx; i<right_idx; i++) mean_idx += d_fft_result[i] * i;
    mean_idx /= total_power;
      
    // Compute the power in the non-peak bins.  Assume that at least
	// 1/4 of the captured bandwidth is noise.  Take the mean of those.
    float noise_est = 0.0;
	sort(d_fft_result.begin(), d_fft_result.end()); 
    for (int i = 0; i<d_fft_size/4; i++) noise_est += d_fft_result[i];

    *freq = mean_idx;

	// Watch out for the case where there's no input at all.  In that case,
	// the DC offset may appear as a valid peak.  Check total power for
	// a sanity check.
	if (noise_est < 10 || total_power < 10)
		*power = 0;
    else{
		//EV *power = total_power / noise_est;
        *power = (total_power/(PEAK_WIDTH+1)) / (noise_est/(d_fft_size/4));
    }

    d_mutex.unlock(); 


    printf(" ----------------------\n");
    printf(" FILE: get_peak_freq \n");
    printf(" \t max_idx=%i \n",max_idx);
    printf(" \t total_power=%4.3f \n",total_power);
    printf(" \t mean_idx=%4.3f \n",(float) mean_idx);
    printf(" \t noise_est=%g \n",noise_est);
    printf(" ----------------------\n");

    return true;
}


bool gr_freq_acq::get_crosscorr_freq(const vector<float> &mask, 
                   int *max_idx, float *max_power,float *mean_power,
                   vector<float> &fft)
{
    // Make sure we have at least the first result ready
    if (!d_result_ready) return false;

    vector<float> result(2*d_fft_size); 

    // Make sure the d_fft_results does not change while we are computing
    // the largest peaks.
    d_mutex.lock(); 
    fft = d_fft_result;

    // We probably don't need the entire crosscorrelation length of 2x FFT size
    ippsCrossCorr_32f((Ipp32f*) &d_fft_result[0], d_fft_size, 
                      (Ipp32f*) &mask[0], d_fft_size, 
                      (Ipp32f*) &result[0], 2*d_fft_size, -d_fft_size);

#ifdef SAVE_STREAM
    ofstream debug("c:/temp/acq_debug.dat");
    copy (d_fft_result.begin(), d_fft_result.end(), ostream_iterator<float>(debug," "));
    debug.close(); 
    debug.open("c:/temp/acq_debug1.dat");
    copy (result.begin(), result.end(), ostream_iterator<float>(debug," "));
    debug.flush();
#endif
    
    // Find the index of the largest peak
    *max_idx = max_element(result.begin(), result.end()) 
               - result.begin();

    *max_power = result[*max_idx];

    ippsMean_32f((float *)&result[0], 2*d_fft_size, mean_power, ippAlgHintFast);

    d_mutex.unlock(); 
    return true;
}

