/*
GOES Radio Comments
This file takes in two correlated outputs from tones fo and f1, that went through an averager and magnitude operation.
The bit boundaries are acquired and tracked, and the demodulator produces {0,1} based on whether f0 or f1 is larger
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gr_io_signature.h>
#include <gr_prefs.h>
#include <ar_fsk_demod_cf.h>
#include <stdexcept>
#include <fstream>

#define TIMING_ACQ_HISTORY 6
// Public constructor

// 0: none, 2: symbol rate vars, 3: timing acq
#define DEBUG_FSK_DEMOD 0

#define DEBUG_EV 1

#if DEBUG_FSK_DEMOD
    ofstream debug("c:/temp/usrp_log/fsk_debug.dat");
    ofstream debug1("c:/temp/usrp_log/fsk_debug1.dat");
    ofstream debug2("c:/temp/usrp_log/fsk_debug2.dat");
#endif

template <class T>
static inline float clip(const T in, const T min_val, const T max_val)
{
    if      (in < min_val) return min_val;
    else if (in > max_val) return max_val;
    else                   return in;
}

gr_fsk_demod_cf_sptr 
gr_make_fsk_demod_cf(float samples_per_symbol, bool rz, 
                     float alpha, float beta)
{
  return gr_fsk_demod_cf_sptr (new gr_fsk_demod_cf (samples_per_symbol, rz, 
                                                    alpha, beta));
}

gr_fsk_demod_cf::gr_fsk_demod_cf (float samples_per_symbol, bool rz, 
                                  float alpha, float beta)
  : gr_block ("fsk_demod_cf",
	      gr_make_io_signature (2, 2, sizeof (float)),
	      gr_make_io_signature (1, 1, sizeof (float))),
    d_samples_per_symbol(samples_per_symbol),
    d_alpha(alpha), d_beta(beta),
    d_snr_alpha(0.1),
    d_lock(true), 
    d_lock_after(-1), d_unlock_after(-1), 
    d_lock_count(0), d_unlock_count(0),
    d_lock_threshold(0),
    d_rz(rz),
    d_rz_blank(false)
{
  for (int k=0; k<N_FSK; k++) {
      d_ontime[k]   = 0.0; 
      d_early[k]    = 0.0; 
      d_late[k]     = 0.0; 
      d_prev[k]     = 0.0;
      d_timing_pwr[k].pwr = 0; 
      d_timing_pwr[k].idx = 0; 
      d_timing_pwr[k].other_tone = 0;
  }
  d_timing_acquired = false;
  d_timing_count = 0;
  d_samples_per_symbol_int = (int)(d_samples_per_symbol + 0.5);
  if (d_rz) // return to zero.  Half each symbol is zero
  {
      d_samples_per_symbol_int /= 2 ;
	  d_samples_per_symbol /= 2;
  }

  d_rz_blank = false;

  d_idx = 0.0;
  d_sample_count = 0.0;
  d_signal_pwr = 0.0;
  d_noise_pwr = 0.0;
  set_relative_rate (1.0 / samples_per_symbol);

  set_history(TIMING_ACQ_HISTORY*(d_samples_per_symbol_int+1));
}

gr_fsk_demod_cf::~gr_fsk_demod_cf ()
{
}

void
gr_fsk_demod_cf::forecast(int noutput_items, gr_vector_int &ninput_items_required)
{
  unsigned ninputs = ninput_items_required.size();
  for (unsigned i=0; i < ninputs; i++)
    ninput_items_required[i] =
      (int) noutput_items * d_samples_per_symbol_int + d_samples_per_symbol_int;
  // See comment [C1] in general_work 
}


// The FSK demodulation consists of two steps:
// 1. Acquisition
// 2. Demodulation
// Both functions require a history buffer of a few symbols.  This buffer
// allows acquisition to run on a sample-by-sample basis.  Once the peaks
// are detected, the acquisition searches back in time over the buffer to
// the first real peak.  
// The demodulation starts at the first real peak, and no bits are lost.
int
gr_fsk_demod_cf::general_work (int noutput_items,
				       gr_vector_int &ninput_items,
				       gr_vector_const_void_star &input_items,
				       gr_vector_void_star &output_items)
{
    static int Demod_cnt =0;
     if (!d_timing_acquired)
    {
        //printf( "ar_fsk_demod_cf.cc --> ACQUIRE \n");
        acquire_timing(noutput_items, ninput_items, input_items, output_items);
        d_sample_count = d_samples_per_symbol_int;
        return 0; // Acquisition does not generate any output
    }
    else
    {
        int got_bits = demod(noutput_items, ninput_items, input_items, output_items);

        if(got_bits>0){
          Demod_cnt ++;
          //printf( "ar_fsk_demod_cf.cc --> DEMOD %i bits=%i \n",Demod_cnt, got_bits);
        }
        

        return got_bits;
    }
}

int
gr_fsk_demod_cf::demod (int noutput_items,
				       gr_vector_int &ninput_items,
				       gr_vector_const_void_star &input_items,
				       gr_vector_void_star &output_items)
{
    // The input is a running average of the past 'd_samples_per_symbol' samples
    const float *in[2];
    in[0] = (const float *) input_items[0] + history() + d_first_peak_idx;
    in[1] = (const float *) input_items[1] + history() + d_first_peak_idx;
    float *out = (float *) output_items[0];

    float acc_pwr[N_FSK];
    int  ni = ninput_items[0] - history(); 

    int ii = 0;
    int oo = 0;

    while(oo < noutput_items && ii < ni) {
        // Assign current input to a temp variable for cleaner code
         for (int k=0; k<N_FSK; k++) {
              d_acc[k] = in[k][ii];
       
               // Save the accumulation value at T/2 to be used for timing tracking
               if ((int)(d_sample_count) == d_samples_per_symbol_int/2-1)
                    d_early[k] = d_acc[k];
        }
        ii++;  
        //EV
        //if ( (int)(d_sample_count) ==  int( d_samples_per_symbol_int/2-1) ){
        //   d_early[0] = d_acc[0];
        //   d_early[1] = d_acc[1];
        // }

        d_sample_count += 1;

        // Process a new symbol
        if (d_sample_count >= d_samples_per_symbol_int)
        {
            // Implement an early-late timing tracking detector and a
            // first order loop.  We need a history of the last two
            // 'ontime' values.  The current 'ontime' value cannot be
            // used since we don't have the 'late' value yet.  Note that
            // we run the detector independently on the two channels.
            // It may be possible to subtract the channels and then
            // use the difference for tracking, but that approach assumes
            // that the magnitudes of the two channels are the same.
            // The accumulation values increase linearly for the duration of a bit
            // period when that bit is on.  The values then decrease linearly when
            // a different bit is sent, or stay at the maximum level if the same
            // bit is sent.  The values then stay at 0 if a different bit is sent.
            // Unlike tranditional BPSK modulation, the values are between 0 and N
            // rather than -N/2 to +N/2.  Thus, the classical TED given by
            // e = (early - late) * ontime 
            // must be modified to shift the 'ontime' by N/2, giving 
            // e = (early - late) * (ontime - mean), where mean = (early + late) /2
            //
            // Only one of the two tones will be active (have a peak) on any 
            // symbol period.  Use only that tone for tracking.
            //
            // For RZ version, neither tone is active (has a peak) every other symbol
            // Don't update the loop during these idle periods.

            // Subtract the currently processed symbol
            d_sample_count -= d_samples_per_symbol_int;
            // Account for non-integer sample period
            d_sample_count += (d_samples_per_symbol_int - d_samples_per_symbol);

            if (!d_rz || (d_rz && d_rz_blank))
            {
                int decision_int = (d_ontime[0] > d_ontime[1]) ? 0 : 1;
                float delta = d_early[decision_int] - d_late[decision_int];
                float mean  = (d_early[decision_int] + d_late[decision_int]) / 2;

                float err = delta * (d_ontime[decision_int] - mean);

                d_idx = err * d_alpha; // + d_idx  Implement a first order loop
                // Clip the error to a reasonable value
                d_idx = clip(d_idx, -2.0f, 2.0f);

                // Add the output of the tracking loop
                d_sample_count -= d_idx;

                // Make sure the sample count does not take on unreasonable values.
                // This may happen if the input data is invalid or just noise.
                d_sample_count = clip(d_sample_count, -2.0f, 2.0f);

    #if DEBUG_FSK_DEMOD==2
            debug << "PWR[0,1]=" << acc_pwr[0] << " " << acc_pwr[1] << 
                " \tPrev=" << d_prev[decision_int] << " \tOn=" << d_ontime[decision_int] << 
                " \t\tEarly=" << d_early[decision_int] << " \tLate=" << d_late[decision_int] << 
                " \tSample=" << d_sample_count << " \tErr=" << err << " \tIdx=" << d_idx << endl;
    #endif
            }

            // Save the current values to be used in the next symbol period
            for (int k=0; k<N_FSK; k++) 
            {
                d_prev[k] = d_ontime[k];
                d_ontime[k] = d_acc[k];
                d_late[k] = d_early[k];
            }

            // Compute the average power in the signal and noise.  Signal
            // power is assumed to be in the larger of the two tones, while
            // noise power is in the smaller.  This should work since only
            // one of the tones may be on at any given time.  These values
            // can be used to estimate the SNR and determine if data is
            // valid.
            float acc_pwr[N_FSK];
            for (int k=0; k<N_FSK; k++) {
                 acc_pwr[k] = d_acc[k] * d_acc[k];
            }

            float signal = max(acc_pwr[0], acc_pwr[1]);
            float noise  = min(acc_pwr[0], acc_pwr[1]);
            d_signal_pwr = (1-d_snr_alpha) * d_signal_pwr + signal * d_snr_alpha;
            d_noise_pwr  = (1-d_snr_alpha) * d_noise_pwr  + noise  * d_snr_alpha;

            if (!d_rz_blank) check_lock(signal);

            // The decision is made by comparing the magnitudes of the 
            // accumulations for each channel. Generate an output here
            // and increment the output counter.  The convention is:
            // x < 0 --> '0', x > 0 --> '1'.  Only output symbols if
            // we are locked.  In RZ mode, every other symbol is blank,
            // so don't output it either.
		    // Note that we check the lock status just before outputing
		    // a new symbol.  This allows us to set the 'unlock_after' to
		    // 0 and never output an invalid symbol
            if (d_lock && !d_rz_blank){
                  out[oo++] = acc_pwr[1] - acc_pwr[0];
                 //out[oo++] = (int)(acc_pwr[1] - acc_pwr[0]);
            }

            if (d_rz) d_rz_blank = !d_rz_blank; // Next symbol is different
        }

    }

    consume_each (ii);

    return oo;
}

void gr_fsk_demod_cf::check_lock(float signal)
{
    float threshold = d_noise_pwr * d_lock_threshold;
    if (d_lock && d_unlock_after != -1) // Check if we should unlock
    {
        if (signal < threshold) d_unlock_count++; else d_unlock_count = 0;
        if (d_unlock_count > d_unlock_after) 
        {
            d_lock = false;
            // Force the timing re-acquisition after a loss of lock
            d_timing_acquired = false;
            d_unlock_count = 0;

          #ifdef DEBUG_EV
            printf("Locked --> Unlocked \n");
          #endif
        }
    } else if (!d_lock && d_lock_after != -1) { // Check if we should lock
        if (signal > threshold) d_lock_count++; else d_lock_count = 0;
        if (d_lock_count > d_lock_after)
        {
            d_lock = true;
            d_lock_count = 0;
            #ifdef DEBUG_EV
               printf("UnLocked --> Locked \n");
            #endif
        }
    }
}

// This function only works for N_FSK = 2.
// Acquisition works by looking for a peak on each tone over a sliding 
// window of 2T (nrz) or 3T (rz).  The window slides one symbol (T) at a 
// time
int gr_fsk_demod_cf::acquire_timing(int noutput_items,
				       gr_vector_int &ninput_items,
				       gr_vector_const_void_star &input_items,
				       gr_vector_void_star &output_items)
{
    // The input is a running average of the past 'd_samples_per_symbol' samples
    const float *in[2];
    in[0] = (const float *) input_items[0] + history();
    in[1] = (const float *) input_items[1] + history();

    float acc_pwr[N_FSK];

    int ii = 0;
    int ni = ninput_items[0] - history(); 

    while (ii < ni && !d_timing_acquired)
    {
        for (int k=0; k<N_FSK; k++) acc_pwr[k] = in[k][ii] * in[k][ii];
        ii++;

        // Save the peaks and the troughs for two consecutive symbols
        for (int k=0; k<N_FSK; k++)
        {
            if (acc_pwr[k] > d_timing_pwr[k].pwr)
            {
                d_timing_pwr[k].pwr = acc_pwr[k];
                d_timing_pwr[k].idx = d_timing_count;
                d_timing_pwr[k].other_tone = acc_pwr[N_FSK-1-k]; 
            }
    #if DEBUG_FSK_DEMOD==4
        debug1 << acc_pwr[k] << " " << d_timing_pwr[k].pwr << " " 
                << d_timing_pwr[k].idx << " " << d_timing_pwr[k].other_tone << " ";
    #endif
        }

        d_timing_count++;
    #if DEBUG_FSK_DEMOD==4
        debug1 << " " << d_timing_count << endl;
        debug1.flush();
    #endif

        int offset = -1;
        int period = (d_rz) ? d_samples_per_symbol_int * 3 : d_samples_per_symbol_int * 2;
        if (d_timing_count > period)
        {
            // Check six conditions to determine if a valid FSK signal with
            // a data transition has been received:
            // 1. Each peak magnitude is at least d_samples_per_symbol^2/4. This
            //    implies that the input power is at least 1/4
            // 2. Each peak greater than d_lock_threshold times the trough
            // 3. Each peak is within 15% of the other
            d_timing_acquired = true;
            float min_pwr = d_samples_per_symbol_int*d_samples_per_symbol_int/4;
            if (d_timing_pwr[0].pwr < min_pwr || d_timing_pwr[1].pwr < min_pwr)
                d_timing_acquired = false; 
            if (d_timing_pwr[0].pwr < d_lock_threshold * d_timing_pwr[0].other_tone)
                d_timing_acquired = false; 
            if (d_timing_pwr[1].pwr < d_lock_threshold * d_timing_pwr[1].other_tone)
                d_timing_acquired = false; 
            if (d_timing_pwr[0].pwr > 1.15 * d_timing_pwr[1].pwr)
                d_timing_acquired = false;
            if (d_timing_pwr[0].pwr < 0.87 * d_timing_pwr[1].pwr)
                d_timing_acquired = false;

            if (d_timing_acquired)
            {
                // The offset is computed by taking the average of the indexes for
                // the two peaks. 
                int offset0 = min(d_timing_pwr[0].idx, d_timing_pwr[1].idx);
                int offset1 = max(d_timing_pwr[0].idx, d_timing_pwr[1].idx);
                offset1 -= (d_rz) ? d_samples_per_symbol_int*2 : d_samples_per_symbol_int;
                offset = (offset0 + offset1)/2;
                // Initialize the noise and signal level estimates based on
                // these two symbols
                d_signal_pwr = (d_timing_pwr[0].pwr + d_timing_pwr[1].pwr) / 2;
                d_noise_pwr = (d_timing_pwr[0].other_tone + d_timing_pwr[1].other_tone) / 2;
                // We don't want to loose any bits, including those that were
                // used for acquisition.  Starting with this peak, search back
                // for up to TIMING_ACQ_HISTORY symbols, looking for peaks. A
                // peak must be within 20% of the d_signal_pwr.  Search back
                // until peaks are smaller than that.
                int p;
                for (p=1; p < TIMING_ACQ_HISTORY; p++)
                {
                    // The current sample number is 'ii', which also corresponds
                    // to sample number 'd_timing_count'  The previous peak (if
                    // any) must have occurred at
                    int idx = ii + offset - p * d_samples_per_symbol_int;
                    for (int k=0; k<N_FSK; k++) 
                        acc_pwr[k] = in[k][idx] * in[k][idx];

                    float peak_pwr = max(acc_pwr[0], acc_pwr[1]);
                    if (peak_pwr < d_signal_pwr * 0.8) // small, done searching
                        break;
                }
                // Process the samples starting with the oldest peak.
                d_first_peak_idx = offset - (p - 1) * d_samples_per_symbol_int;

                // Prepare to lock on the next few symbols
                d_lock_count = 0; d_unlock_count = 0;
                d_rz_blank = false;
                // Clean up for the next acquisition
                for (int k=0; k<N_FSK; k++) 
                {
                    d_timing_pwr[k].pwr = 0;
                    d_timing_pwr[k].idx = 0;
                    d_timing_pwr[k].other_tone = 0;
                }
                d_timing_count = 0;
            } else { // Slide window by one symbol
                for (int k=0; k<N_FSK; k++) 
                {
                    if (d_timing_pwr[k].idx < d_samples_per_symbol_int) // old
                    {
                        d_timing_pwr[k].pwr = 0; 
                        d_timing_pwr[k].idx = 0; 
                        d_timing_pwr[k].other_tone = 0;
                    } else { // Slide back one symbol
                        d_timing_pwr[k].idx -= d_samples_per_symbol_int;
                    }
                }
                d_timing_count -= d_samples_per_symbol_int;
            }

        }
    }
    consume_each(ii);

    //printf("Consumed %i bits . d_first_peak_idx=%i \t Acquired=%i \n",  ii,  d_first_peak_idx, d_timing_acquired*1);
    return d_first_peak_idx;
}
