
#ifndef INCLUDED_GR_fsk_demod_cf_H
#define	INCLUDED_GR_fsk_demod_cf_H

#include <gr_block.h>
#include <gr_complex.h>

using namespace std;
class ari_mmse_fir_interpolator_cc;

class gr_fsk_demod_cf;
typedef boost::shared_ptr<gr_fsk_demod_cf> gr_fsk_demod_cf_sptr;

#define N_FSK 2 // Only supporting 2-FSK for now

/*! Create an FSK demodulator
*  \param samples_per_symbol   Number of samples per symbol
*  \param alpha                Timing tracking loop coefficient
*  \param beta                 Timing tracking loop coefficient
*/
gr_fsk_demod_cf_sptr 
gr_make_fsk_demod_cf (float samples_per_symbol, bool rz, float alpha, float beta);

/*!
 * Binary FSK demodulator with timing tracking.  This block takes two
 * inputs generated by multiplying the input samples by two complex
 * exponentials corresponding to the _negative_ of the two FSK tone
 * frequencies, computing a magnitude of the running sum over the last
 * samples_per_symbol saples.  The input is therefore:
 *
 * y_k(n) = Abs[ Sum[ x(n-t) * Exp[j*f_k*(n-t) , samples_per_symbol<t<0 ] ] ]
 *
 * The block outputs the difference between the accumulations of the
 * two mixer outputs over the duration of a symbol.
 * In addition to providing demodulation, this block also keeps
 * track of the mean (signal+noise) and noise powers.
 * \ingroup block
 *
 */
class gr_fsk_demod_cf : public gr_block
{
 public:
  ~gr_fsk_demod_cf ();
  void forecast(int noutput_items, gr_vector_int &ninput_items_required);
  int general_work (int noutput_items,
		    gr_vector_int &ninput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items);

  //! Returns the average value of the larger of the two tones
  //! y[n+1]=y[n]*(1-snr_averaging)+snr_averaging*x[n]
  float get_signal_power() {return d_signal_pwr;}
  //! Returns the average value of the smaller of the two tones
  //! y[n+1]=y[n]*(1-snr_averaging)+snr_averaging*x[n]
  float get_noise_power()  {return d_noise_pwr;}
  //! Set the power averaging constant used in get_[signal/noise]_pwr
  void  set_snr_averaging(float snr_alpha) { d_snr_alpha = snr_alpha;}
  //! Check if the receiver is locked
  bool get_locked() {return d_lock;}
  //! Set the locking criteria.  The demodulator compares the larger of the
  //! two inputs to the average of the smaller input (noise_power) on every
  //! symbol.  Lock is declared after 'lock_after' consecutive symbols that
  //! exceed noise_pwr*lock_threshold.  Loss of lock is declared after 
  //! 'unlock_after' consecutive symbols that are below noise_pwr*lock_threshold
  //! Calling this function forces a loss of lock.  
  //! \param threshold Set the threshold for 'good' symbols
  //! \param lock_after Declare lock after this many 'good' symbols.  Set to -1
  //!                   to always declare lock.
  //! \param unlock_after Loose lock after this many 'bad' symbols.  Set to -1
  //!                     to never loose lock after the first lock
  void set_locking(float threshold, int lock_after, int unlock_after)
  {
      d_lock_threshold = threshold; 
      d_lock_after = lock_after; 
      d_unlock_after = unlock_after;
      if (d_lock_after != -1) d_lock = false; else d_lock = true;
  }
  

protected:
  gr_fsk_demod_cf (float samples_per_symbol, bool rz, float alpha, float beta);

private:
  float d_early[N_FSK], d_late[N_FSK], d_ontime[N_FSK], d_prev[N_FSK];
  float d_acc[N_FSK];
  float d_idx;
  float d_sample_count;
  float d_samples_per_symbol;
  int   d_samples_per_symbol_int;

  float d_alpha, d_beta;
  float d_signal_pwr;
  float d_noise_pwr;
  float d_snr_alpha;

  float d_lock_threshold;
  int   d_lock_after, d_unlock_after;
  int   d_lock_count, d_unlock_count;
  bool  d_lock;

  struct {
      float pwr;
      int idx;
      float other_tone;
  } d_timing_pwr[N_FSK];
  int d_timing_count;
  bool d_timing_acquired;
  int d_first_peak_idx;

  bool d_rz;
  bool d_rz_blank;

  void check_lock(float signal);

  int acquire_timing (int noutput_items,
	         gr_vector_int &ninput_items,
             gr_vector_const_void_star &input_items,
             gr_vector_void_star &output_items);

  int demod (int noutput_items,
	         gr_vector_int &ninput_items,
             gr_vector_const_void_star &input_items,
             gr_vector_void_star &output_items);

  friend gr_fsk_demod_cf_sptr
  gr_make_fsk_demod_cf (float samples_per_symbol, bool rz, float alpha, float beta);
};

#endif
