
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_fsk_mod_bc.h>
#include <gr_io_signature.h>
#include <math.h>
#include <iostream>
#include <fstream>

#define M_TWOPI (2*M_PI)

gr_fsk_mod_bc_sptr
  gr_make_fsk_mod_bc(float sampling_freq, float samples_per_symbol, 
							   const std::vector<float> &freqs)
{
  return gr_fsk_mod_bc_sptr (new gr_fsk_mod_bc (sampling_freq, samples_per_symbol, freqs));
}

gr_fsk_mod_bc::gr_fsk_mod_bc (float sampling_freq, float samples_per_symbol, 
							   const std::vector<float> &freqs)
  : gr_block ("fsk_mod_bc",
		   gr_make_io_signature (1, 1, sizeof (char)),
		   gr_make_io_signature (1, 1, sizeof (gr_complex))),
    d_sampling_freq(sampling_freq), d_samples_per_symbol(samples_per_symbol),
    d_symbol(0), d_mode(TX_DATA), d_burst_length(-1), 
    d_burst_delay(0), d_rz(false), d_amplitude(1.0),
    d_valid_data(false)
{
  d_samples_per_symbol_int = (int)(d_samples_per_symbol + 0.5);
  // Initialize the sample counter to an end of symbol to make sure a new
  // bit is fetched the first time 'work' is called
  d_sample = d_samples_per_symbol_int;
  set_freqs(freqs);
}

int gr_fsk_mod_bc::general_work (int noutput_items,
		    gr_vector_int &ninput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items)
{
  const char *in = (const char *) input_items[0];
  gr_complex *out = (gr_complex *) output_items[0];

  // The modulator can be in one of two states: TX_DATA or BURST_DELAY
  // BURST_DELAY is only applicable if burst mode is configured
  int oo = 0, ii = 0; // count the number of output samples and input bits
  while (oo < noutput_items)
  {
    if (d_mode == TX_DATA)
    {
      // Compute values one at a time.  This can be sped up quite a bit
      // by computing a block of values at once.
      if (d_sample < d_samples_per_symbol_int)
      {
        // Some standards use on-off-keying (OOK) on top of FSK.  This is 
        // equivalent to RZ (return to zero).  If RZ mode is configured,
        // one half of the symbol is just set to zero.  Also note that if
        // the input data is not valid, the output amplitude is set to 0
        if ((!d_rz || d_sample < d_samples_per_symbol_int / 2) && d_valid_data)
          d_nco.sincos(out + oo, 1, d_amplitude);
        else
          out[oo] = 0.0;

        d_sample += 1.0;
        oo++;
      } else {
        // We just completed a symbol.  Either get the next symbol, or if this
        // is the end of a burst, go to BURST_DELAY
        if (d_symbol < d_burst_length || d_burst_length == -1)
        {
          char data = data = in[ii++];
          if (data != -1) // negative one means don't send anything
          {
            d_nco.set_freq(d_freqs[ data ]);
            d_valid_data = true;
          } else
            d_valid_data = false;

          d_symbol++;
          d_sample -= d_samples_per_symbol; // subtract processed samples
        } else { // start of burst delay
          d_symbol = 0;
          d_sample = 0.0; // start burst delay
          d_mode = BURST_DELAY;
        }
      }
    } else { // BURST_DELAY
      if (d_sample < d_burst_delay)
      {
        out[oo] = 0.0;
        d_sample += 1.0;
        oo++;
      } else {
        d_symbol = 0;
        d_sample = 0; // start a new symbol
        d_mode = TX_DATA;
      }
    }
  }

  consume_each(ii);

  return noutput_items;
}

void
gr_fsk_mod_bc::forecast(int noutput_items, gr_vector_int &ninput_items_required)
{
  unsigned ninputs = ninput_items_required.size();

  for (unsigned i=0; i < ninputs; i++)
    ninput_items_required[i] = (int) ceil((float)noutput_items / 
                                          (float)d_samples_per_symbol_int);
}

void gr_fsk_mod_bc::set_freqs (const std::vector<float> &freqs)
{	
  d_freqs.resize(freqs.size());
  float min_freq = d_sampling_freq; // could not be any higher!

  for(int i=0; i< d_freqs.size(); i++) // Convert from Hz to radians / sample
  {
    float freq = freqs[i];
    // Do some sanity checking.  Ideally, we want to sample at 4xNyquist, but
    // at least 2xNyquist
    if (freq > d_sampling_freq / 2)
      std::cerr << "fsk_tx: tone " << i << " freq > Fs/2 : " << freq << std::endl;
    if (freq > d_sampling_freq / 4)
      std::cerr << "fsk_tx: tone " << i << " freq > Fs/4 : " << freq << std::endl;

    min_freq = std::min(freq, min_freq);

    d_freqs[i] = freq / d_sampling_freq * M_TWOPI;
  }

  // Check that we get at least one full period at the lowest frequency
  if (min_freq * d_samples_per_symbol < d_sampling_freq)
    std::cerr << "fsk_tx: Symbol duration is shorter than one period\n";
}
