#ifndef INCLUDED_GR_fsk_mod_bc_H
#define INCLUDED_GR_fsk_mod_bc_H

#include <gr_block.h>
#include <gr_fxpt_nco.h>

class gr_fsk_mod_bc;
typedef boost::shared_ptr<gr_fsk_mod_bc> gr_fsk_mod_bc_sptr;

/*! Create an FSK modulator
*  \param sampling_freq       Output sampling frequency in Hz
*  \param samples_per_symbol   Number of samples per symbol
*  \param freqs                A vector of FSK tone frequencies in Hz
*/
gr_fsk_mod_bc_sptr gr_make_fsk_mod_bc (float sampling_freq, 
    float samples_per_symbol, const std::vector<float> &freqs);

/*!
 * Generate FSK tones based on the input data.  Supports multiple FSK 
 * tones (i.e. > 2).  Supports continous or bursty transmission.  Supports
 * RZ (OOK).
 * \ingroup block
 */
class gr_fsk_mod_bc : public gr_block
{
  friend gr_fsk_mod_bc_sptr gr_make_fsk_mod_bc (float sampling_freq, 
    float samples_per_symbol, const std::vector<float> &freqs);

  std::vector<float> d_freqs; //!< Frequencies to transmit for different bits
  float d_samples_per_symbol;
  int   d_samples_per_symbol_int;
  float d_sampling_freq;
  int d_burst_length, d_burst_delay; //!< Length in symbols, delay in samples
  bool d_rz;                         //!< Return to zero / OOK
  float d_sample;
  int d_symbol;           // keep a running count of processed samples and symbols
  float d_amplitude;
  enum {TX_DATA, BURST_DELAY} d_mode;
  bool d_valid_data;

  gr_fxpt_nco d_nco;

  gr_fsk_mod_bc (float sampling_freq, float samples_per_symbol, const std::vector<float> &freqs);

 public:
  std::vector<float> &freqs() { return d_freqs; }

  /*! Specify a vector of frequencies (2 elements for 2-FSK).  Values in [Hz] */
  void set_freqs (const std::vector<float> &freqs);

  /*! Transmission may either be contiguous, or bursty.  A burst consists of 
   *  an integer number of symbols, followed by a delay during which nothing
   *  is sent (zeros).  Continous is the default mode.
   * \param burst_length   Number of symbols in a burst
   * \param burst_delay    Delay between bursts, in seconds.  Will be 
   *                       converted to samples using sampling_freq
   */
  void set_burst (int burst_length, float burst_delay) 
    { d_burst_length = burst_length; d_burst_delay = burst_delay / d_sampling_freq;}

  /*! Some standards use on-off-keying (OOK) on top of FSK.  This is 
   *  equivalent to RZ (return to zero).  If RZ mode is configured,
   *  one half of the symbol is just set to zero.
   * \param rz Set to true for RZ mode, or false for NRZ (default)
   */
  void set_rz(bool rz) { d_rz = rz;}

  /*! Set the amplitude of the tone.  Relative units */
  void set_amplitude(float amplitude) { d_amplitude = amplitude;}

  //! Return the currently set amplitude (set in set_amplitude) 
  float get_amplitude() const { return d_amplitude;}

  void forecast(int noutput_items, gr_vector_int &ninput_items_required);
  int general_work (int noutput_items,
		    gr_vector_int &ninput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items);
};

#endif
