
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_integrate_dump_ff.h>
#include <gr_io_signature.h>
#include <ipp.h>

gr_integrate_dump_ff_sptr 
    gr_make_integrate_dump_ff (int samples_per_symbol, int vlen)
{
  return gr_integrate_dump_ff_sptr (new gr_integrate_dump_ff (samples_per_symbol, vlen));
}

gr_integrate_dump_ff::gr_integrate_dump_ff (int samples_per_symbol, int vlen)
  : gr_block ("integrate_dump_ff",
		      gr_make_io_signature (1, 1, sizeof (float)*vlen),
		      gr_make_io_signature (1, 1, sizeof (float)*vlen)),
    d_samples_per_symbol(samples_per_symbol), d_vlen(vlen),
    d_sample_count(0), d_mean_pwr(0.0), d_pwr_alpha(0.1)
{
    d_acc = new float[vlen];
}

gr_integrate_dump_ff::~gr_integrate_dump_ff()
{ delete [] d_acc; }

// The forcast is intentionally incorrect.  We know that it takes
// ninput_items_required = samples_per_symbol*noutput_items.  However,
// the scheduler will not give us any data until it has enough for at
// least one output.  If samples_per_symbol is large, and vlen is
// also large, the total number of required samples may exceed the size
// of the inter-block buffers.  Thus, we only ask for one input for
// every output, but don't generate outputs for every input.
void gr_integrate_dump_ff::forecast(int noutput_items, 
                                    gr_vector_int &ninput_items_required)
{
  unsigned ninputs = ninput_items_required.size ();
  for (unsigned i = 0; i < ninputs; i++)
    ninput_items_required[i] = noutput_items;
}

int
gr_integrate_dump_ff::general_work (int noutput_items,
		    gr_vector_int &ninput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items)

{
  float *in = (float *) input_items[0];
  float *out = (float *) output_items[0];

  int input_size = d_vlen * sizeof(float);

  int ni = ninput_items[0];
  float sum;

  int oo = 0, ii = 0; // Count number of input and output items

  while (ii < ni && oo < noutput_items)
  {
    // Figure out how many inputs should be processed to complete the next
    // output.
    int samples_left = d_samples_per_symbol - d_sample_count;
    // Check how many input samples we actually have left.
    int samples = std::min(samples_left, ni - ii);
    ii += samples; // consumed this many input samples
    // Check if we are about to complete a symbol.  I.e. we got samples_left
    bool symbol_done = (samples == samples_left);
    if (d_vlen == 1) // Sum consequtive elements
    {
        if (d_sample_count == 0) d_acc[0] = 0.0; // Reset accumulator to start
        float sum;
        ippsSum_32f(in, samples, &sum, ippAlgHintFast);
        d_acc[0] += sum;
        in += samples; // next block
    } else { // Sum across vectors
        // Copy the first input to the output to start the accumulation process
        int i = 0;
        if (d_sample_count == 0)
        {
            memcpy(d_acc, in, input_size);
            i = 1; // Consume one sample
            in += d_vlen;
        }
        
        for (; i < samples; i++) // Use up the rest of the samples
        {
            ippsAdd_32f_I(in, d_acc, d_vlen);
            in += d_vlen;
        }

    }
    
    if (symbol_done)
    {
        memcpy(out, d_acc, input_size);
        out += d_vlen; // Move to the next output
        oo++; // Increment the number of output symbols
        // Compute the sum over the entire vector
        float sum;
        ippsSum_32f(d_acc, d_vlen, &sum, ippAlgHintFast);
        d_mean_pwr = (1-d_pwr_alpha) * d_mean_pwr + d_pwr_alpha * sum;
        d_sample_count = 0;
    } else {
        d_sample_count += samples;
    }
  }

  consume_each (ii);

  return oo;
}
