#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_interp_fir_filter_ccf.h>
#include <gr_fir_util.h>
#include <gr_io_signature.h>
#include <stdexcept>
#include <iostream>

#define MAX_BLOCK_SIZE (1<<15)

gr_interp_fir_filter_ccf_sptr gr_make_interp_fir_filter_ccf (unsigned interpolation, const std::vector<float> &taps)
{
  return gr_interp_fir_filter_ccf_sptr (new gr_interp_fir_filter_ccf (interpolation, taps));
}

gr_interp_fir_filter_ccf::gr_interp_fir_filter_ccf (unsigned interpolation, const std::vector<float> &taps)
  : gr_sync_interpolator ("interp_fir_filter_ccf",
			  gr_make_io_signature (1, 1, sizeof (gr_complex)),
			  gr_make_io_signature (1, 1, sizeof (gr_complex)),
			  interpolation),
    d_updated (false), d_taps(taps), d_interpolation(interpolation)
{
  if (interpolation == 0)
  throw std::out_of_range ("interpolation must be > 0");

  if (interpolation == 1) // single rate
  {
    ippsFIRInitAlloc_32f(&d_filt_re, (Ipp32f *)&d_taps[0], d_taps.size(), 0);
    ippsFIRInitAlloc_32f(&d_filt_im, (Ipp32f *)&d_taps[0], d_taps.size(), 0);
  }
  else
  {
    ippsFIRMRInitAlloc_32f(&d_filt_re, (Ipp32f *)&d_taps[0], d_taps.size(), 
                           interpolation, 0, 1, 0, 0);
    ippsFIRMRInitAlloc_32f(&d_filt_im, (Ipp32f *)&d_taps[0], d_taps.size(), 
                           interpolation, 0, 1, 0, 0);
  }

  d_buf_re = ippsMalloc_32f(MAX_BLOCK_SIZE);
  d_buf_im = ippsMalloc_32f(MAX_BLOCK_SIZE);

}

gr_interp_fir_filter_ccf::~gr_interp_fir_filter_ccf ()
{
    ippsFIRFree_32f(d_filt_re);
    ippsFIRFree_32f(d_filt_im);
    ippsFree(d_buf_re);
    ippsFree(d_buf_im);
}

void
gr_interp_fir_filter_ccf::set_taps (const std::vector<float> &taps)
{
}


void
gr_interp_fir_filter_ccf::install_taps (const std::vector<float> &taps)
{
  d_taps = taps;
  ippsFIRSetTaps_32f((Ipp32f *)&d_taps[0], d_filt_re);
  ippsFIRSetTaps_32f((Ipp32f *)&d_taps[0], d_filt_im);
}

int
gr_interp_fir_filter_ccf::work (int noutput_items,
		   gr_vector_const_void_star &input_items,
		   gr_vector_void_star &output_items)
{
    const gr_complex *in = (const gr_complex *) input_items[0];
    gr_complex *out = (gr_complex *) output_items[0];

    if (d_updated) {
        install_taps (d_taps);
        return 0;		     // history requirements may have changed.
    }

    // The filter generates more output than input samples.  
    int ninput_items = noutput_items / interpolation();
    noutput_items = ninput_items * interpolation(); // In case of non-integer

    // Convert the complex input into two real streams
    ippsCplxToReal_32fc((Ipp32fc *)in, d_buf_re, d_buf_im, ninput_items);
    // Filter real and imaginary parts separately
    ippsFIR_32f_I(d_buf_re, ninput_items, d_filt_re);
    ippsFIR_32f_I(d_buf_im, ninput_items, d_filt_im);
    // Copy the (decimated) input to the output buffer, converting back to 
    // complex
    ippsRealToCplx_32f(d_buf_re, d_buf_im, (Ipp32fc *)out, noutput_items);

    return noutput_items;
}
