#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_multistage_decimator_cc.h>
#include <gr_io_signature.h>
#include <cmath>
#include <gr_firdes.h>
#include "ari_rate_change_factor.h"

static const int MAX_DECIMATION_PER_STAGE = 6;
#define MAX_BLOCK_SIZE (1<<15)

gr_multistage_decimator_cc_sptr gr_make_multistage_decimator_cc (int decimation)
{
  return gr_multistage_decimator_cc_sptr (new gr_multistage_decimator_cc (decimation));
}


gr_multistage_decimator_cc::gr_multistage_decimator_cc (int decimation)
  : gr_sync_decimator ("multistage_decimator_cc",
		       gr_make_io_signature (1, 1, sizeof (gr_complex)),
		       gr_make_io_signature (1, 1, sizeof (gr_complex)),
		       decimation)
    
{
    if (decimation == 0)
    throw std::out_of_range ("decimation must be > 0");

    // Create the decimation stages
    d_ratios = ari_rate_change_factor(decimation, MAX_DECIMATION_PER_STAGE);
    int num_stages = d_ratios.size();

    d_taps.resize(num_stages);
    d_filt_re.resize(num_stages); d_filt_im.resize(num_stages);

    for (int i=0; i<num_stages; i++)
    {
        // Generate the taps
        d_taps[i] = gr_firdes::low_pass(1.0, 1.0, 0.5 / d_ratios[i], 0.1);
        // Now, create the IPP filters
        ippsFIRMRInitAlloc_32f(&(d_filt_re[i]), &(d_taps[i][0]),
                                d_taps[i].size(), 1, 0, d_ratios[i], 0, 0);
        ippsFIRMRInitAlloc_32f(&(d_filt_im[i]), &(d_taps[i][0]),
                                d_taps[i].size(), 1, 0, d_ratios[i], 0, 0);
    }
    d_buf_re = ippsMalloc_32f(MAX_BLOCK_SIZE);
    d_buf_im = ippsMalloc_32f(MAX_BLOCK_SIZE);

    // Update the decimation ratio in the underlying gr_sync_decimator block
    decimation = 1;
    for (int i=0; i<num_stages; i++) decimation *= d_ratios[i];
    set_decimation(decimation);
}

gr_multistage_decimator_cc::~gr_multistage_decimator_cc ()
{
    for (int i=0; i<d_taps.size(); i++) 
    {
        ippsFIRFree_32f(d_filt_re[i]);
        ippsFIRFree_32f(d_filt_im[i]);
    }
    ippsFree(d_buf_re);
    ippsFree(d_buf_im);
}

int
gr_multistage_decimator_cc::work (int noutput_items,
		   gr_vector_const_void_star &input_items,
		   gr_vector_void_star &output_items)
{
    gr_complex *in  = (gr_complex *) input_items[0];
    gr_complex *out = (gr_complex *) output_items[0];

    int ninput_items = noutput_items * decimation();
    // Convert the complex input into two real streams
    ippsCplxToReal_32fc((Ipp32fc *)in, d_buf_re, d_buf_im, ninput_items);
    for (int i=0; i<d_taps.size(); i++)
    {
        // Each stage uses 'ninput_items' * d_radios[i] input samples and
        // generates 'ninput_items' output samples
        ninput_items /= d_ratios[i];

        ippsFIR_32f_I(d_buf_re, ninput_items, d_filt_re[i]);
        ippsFIR_32f_I(d_buf_im, ninput_items, d_filt_im[i]);
    }
    // Copy the (decimated) input to the output buffer, converting back to 
    // complex
    ippsRealToCplx_32f(d_buf_re, d_buf_im, (Ipp32fc *)out, noutput_items);

    return noutput_items;
}
