
#ifndef INCLUDED_gr_ring_buffer_sink_H
#define INCLUDED_gr_ring_buffer_sink_H

#include <gr_sync_block.h>
#include "circular_buffer.h" // Thread-safe ring buffer

template <class T> class gr_ring_buffer_sink;
typedef gr_ring_buffer_sink<char> gr_ring_buffer_sink_b;
typedef boost::shared_ptr<gr_ring_buffer_sink_b> gr_ring_buffer_sink_b_sptr;
typedef gr_ring_buffer_sink<gr_complex> gr_ring_buffer_sink_c;
typedef boost::shared_ptr<gr_ring_buffer_sink_c> gr_ring_buffer_sink_c_sptr;

//! Create a gr_ring_buffer_sink.  Define a unique function for each data type
//! allow_overflow If true, the buffer always accept all the input
//! and never blocks.  If data overflows, it is discarded.
//! \param size Set the buffer size.  If zero, defaults to 16384 (2^14) entries
gr_ring_buffer_sink_b_sptr gr_make_ring_buffer_sink_b(bool allow_overflow,
                                                          int size = 0);

gr_ring_buffer_sink_c_sptr gr_make_ring_buffer_sink_c(bool allow_overflow,
                                                          int size = 0);
/*!
 * A general-purpose buffer to take data from the GNU radio environment  
 * to 'C++' world
 */
template <class T>
class gr_ring_buffer_sink : public gr_sync_block
{
  friend gr_ring_buffer_sink_b_sptr 
      gr_make_ring_buffer_sink_b(bool allow_overflow, int size);

  friend gr_ring_buffer_sink_c_sptr
      gr_make_ring_buffer_sink_c(bool allow_overflow, int size);

 private:
  circular_buffer<T> *buffer;
  bool d_allow_overflow;
  bool d_overflow;

 protected:
  gr_ring_buffer_sink(bool allow_overflow, int size);

 public:
	
  ~gr_ring_buffer_sink();

  int work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items);

  //! Return the buffer to the C++ world
  circular_buffer<T> *get_buffer() {return buffer;}
  //! Check if the buffer has overflowed.  Only meaningful if 
  //! 'allow_overflow' is set to true.
  bool overflow() {return d_overflow;}
  //! Clear the overflow flag.  Only meaningful if 'allow_overflow'
  //! is set to true.
  void clear_overflow() {d_overflow = false;}
};

#define GR_RING_BUFFER_SINK_SIZE  (1<<14) // Default buffer size

#endif /* INCLUDED_gr_ring_buffer_sink_H */
