
#ifndef INCLUDED_gr_ring_buffer_source_H
#define INCLUDED_gr_ring_buffer_source_H

#include <gr_sync_block.h>
#include "circular_buffer.h" // Thread-safe ring buffer

template <class T> class gr_ring_buffer_source;
typedef gr_ring_buffer_source<char> gr_ring_buffer_source_b;
typedef boost::shared_ptr<gr_ring_buffer_source_b> gr_ring_buffer_source_b_sptr;
typedef gr_ring_buffer_source<gr_complex> gr_ring_buffer_source_c;
typedef boost::shared_ptr<gr_ring_buffer_source_c> gr_ring_buffer_source_c_sptr;

//! Create a gr_ring_buffer_source.  Define a unique function for each data type
//! \param allow_underflow If true, the buffer always generates as much input
//! as requested by the scheduler.  If more data is requested than available,
//! the output data is padded with 'default_value'
//! \param size Set the buffer size.  If zero, defaults to 16384 (2^14) entries
gr_ring_buffer_source_b_sptr gr_make_ring_buffer_source_b(bool allow_underflow,
                                                          int size = 0);

gr_ring_buffer_source_c_sptr gr_make_ring_buffer_source_c(bool allow_underflow,
                                                          int size = 0);
/*!
 * A general-purpose buffer to take data from the 'C++' world to GNU radio 
 * environment
 */
template <class T>
class gr_ring_buffer_source : public gr_sync_block
{
  friend gr_ring_buffer_source_b_sptr 
      gr_make_ring_buffer_source_b(bool allow_underflow, int size);

  friend gr_ring_buffer_source_c_sptr
      gr_make_ring_buffer_source_c(bool allow_underflow, int size);

 private:
  circular_buffer<T> *buffer;
  bool d_allow_underflow;
  bool d_underflow;
  T d_default_data;

 protected:
  gr_ring_buffer_source(bool allow_underflow, int size);

 public:
	
  ~gr_ring_buffer_source();

  int work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items);

  //! Return the buffer to the C++ world
  circular_buffer<T> *get_buffer() {return buffer;}
  //! Check if the buffer has underflowed.  Only meaningful if 
  //! 'allow_underflow' is set to true.
  bool underflow() {return d_underflow;}
  //! Clear the underflow flag.  Only meaningful if 'allow_underflow'
  //! is set to true.
  void clear_underflow() {d_underflow = false;}
  //! Set the default value to output if the buffer underflows.
  //! Only meaningful if 'allow_underflow' is set to true.
  void set_default_data(T default_data) {d_default_data = default_data;}
};

#define GR_RING_BUFFER_SOURCE_SIZE  (1<<14) // Default buffer size

#endif /* INCLUDED_gr_ring_buffer_source_H */
