/* -*- c++ -*- */
/*
 * Copyright 2005 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_throughput.h>
#include <gr_io_signature.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

gr_throughput_sptr
gr_make_throughput(size_t itemsize, double update_rate, 
				   const char *filename, const char *label)
{
  return gr_throughput_sptr(new gr_throughput(itemsize, update_rate,
	                                          filename, label));
}

gr_throughput::gr_throughput(size_t itemsize, double update_rate,
							 const char *filename, const char *label)
  : gr_sync_block("throughput",
		  gr_make_io_signature(1, 1, itemsize),
		  gr_make_io_signature(0, 0, 0)),
    d_update_rate(update_rate*CLOCKS_PER_SEC),
    d_total_samples(0), d_prev_time(0), d_label(label)
{
	if (strcmp(filename, "cout") == 0)
		d_outfile = stdout;
	else
		d_outfile = fopen(filename, "w");
}

gr_throughput::~gr_throughput()
{
}

int
gr_throughput::work (int noutput_items,
		   gr_vector_const_void_star &input_items,
		   gr_vector_void_star &output_items)
{
  double delta_time = clock() - d_prev_time;
  if (d_prev_time != 0 && delta_time > d_update_rate && d_update_rate > 0)
  {
	double throughput = d_total_samples / delta_time * CLOCKS_PER_SEC / 1000.0;
    fprintf(d_outfile, "Throughput::%s = %3.5g Ksps\n", d_label.c_str(), throughput);
	d_total_samples = 0;
  }

  if (d_total_samples == 0) // First time we got called or just reset
  	  d_prev_time = clock();

  d_total_samples += noutput_items;

  return noutput_items;
}
