/* 
GOES Radio Comments
There are many overloaded versions of the class gr_viterbi, but the main one is gr_viterbi(int modulation, resynch_period, float AGC, float vthresh)
This class is a wrapper for a Viterbi core decoder.  THe core decoder is defined by the function sdvd_intel() and is defined in ar_viterbi_lib.cc
The modulation can be set to either 2 (BPSK) or 4 (QPSK).  QPSK and BPSK share the same viterbi class.  The AGC level is used to make sure that the constellation
data has the proper dynamic range (namely {-1,1} instead of {-agc, agc}).  THe reason the range of the signal is important is because a look
up table utilized for soft-decision Viterbi decoder.  The viterbi decoder performs synchronizatoin to find out the pairing of the signals.  It does
this by creating the proper number of hypotheses, 2 for BPSK, and 9 for OQPSK.  To check each hypothesis the data is decoded, then re-encoded.  
The BER between the re-encoded sequence and the original Viterbi input is checked.  IF the BER is less than vthresh for a given hypothesis, then 
Viterbi synchronization lock is declared and that hypothesis is chosen.  Once the Viterbi decoder is locked it continually decodes data, and only
checks to verify that it is still synched every resynch_period.  

*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ar_viterbi_lib.h>
#include "ar_viterbi.h"
#include <gr_io_signature.h>
#include <ipp.h>
#include <stdio.h>


#define TRIAL_OFFSET (10000)
#define TRIALBITS (800)

#define BEGINN (1)
#define FORCE_SYNC
//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
#define RESYNC
//#define DISPLAY_RESYNC
//-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-



gr_viterbi_sptr     gr_make_viterbi(int modulation, int resynch_period, float AGC, float vthresh)
{
    return gr_viterbi_sptr (new gr_viterbi (modulation, resynch_period, AGC, vthresh) );
}

gr_viterbi_sptr     gr_make_viterbi(int modulation,int resynch_period)
{
    return gr_viterbi_sptr (new gr_viterbi (modulation, resynch_period, 10, 0.125) );
}

gr_viterbi_sptr     gr_make_viterbi(int modulation)
{
    return gr_viterbi_sptr (new gr_viterbi (modulation, 0, 10, 0.125) );
}

gr_viterbi_sptr     gr_make_viterbi()
{
    return gr_viterbi_sptr (new gr_viterbi (2, 0, 10, .125) );
}

gr_viterbi::gr_viterbi (int modulation, int resynch_period, float AGC, float vthresh)
        : gr_block ("viterbi",
                    gr_make_io_signature (1, 1, sizeof (float)),
                    gr_make_io_signature (1, 1, sizeof (unsigned char)))  ,
        d_modulation(modulation) ,
        d_AGC(AGC),
        d_resynch_period(resynch_period),
        d_vthresh(vthresh)
{ }

gr_viterbi::~gr_viterbi()
{ }

void gr_viterbi::forecast(int noutput_items,
                          gr_vector_int &ninput_items_required)
{
    ninput_items_required[0] = ( 5000+ noutput_items  + DEPTH_TRELLIS) * 2;
}




int
gr_viterbi::general_work (int                       noutput_items,
                          gr_vector_int             &ninput_items,
                          gr_vector_const_void_star &input_items,
                          gr_vector_void_star       &output_items){

    // FYI:  DEPTH_TRELLIS=35
    float     *in_vec             =  (float *)   input_items[0];
    int       begin_state         =0;
    int       begin_state_tmp     =0;
    int       ni                  = ninput_items[0];
    int       channel_length;
    unsigned char  *vit_bytes     = (unsigned char *)   output_items[0];
    //Use Multiples of Bytes. 16 = 8 * 1/CodeRate
    channel_length = (ni/16)*16;
    // Warning: Viterbi Core uses Intel IPP functions that returns 16 bit shorts !,
    //          not 8 bit unsignedchars
    //short          *vit_out		     = new short [channel_length];
    short          *vit_out		     = new short [channel_length/2+1];
    //static int     VitSynched          =0;
    int            generated_output_bits ;

    for (int nn=0; nn<ninput_items[0]; nn++){
        //Make the symbols nominal {-1,1} for consitency of table lookup
        in_vec[nn] = in_vec[nn]/(d_AGC*.8); //JR, The AGC doesn't put out 10, it usually puts out 8 due to noise!
    }
    static int         VitSynched          =0;
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ifdef RESYNC
    static int OnceInSynch=0;
    static int PreviousState=-1;
    static int CurrentState=-1;
    static int ConsecutiveSynchs   =0;

    //PreviousState = CurrentState;
    //EV If we don't make this static, then it tries to re-synch for every frame
    //static int         ConsecutiveSynchs   =0;
    if ( (d_resynch_period >0) && ((ConsecutiveSynchs%d_resynch_period)==0) ){
        if ( d_resynch_period !=1){
            VitSynched       = ConsecutiveSynchs %  d_resynch_period;
        }else{// ==1, then resynch every time
            VitSynched =0;
        }
        // VitSynched       = ConsecutiveSynchs %  5;
    }
    ConsecutiveSynchs++;
    //printf(" ===========> Synched ? = %i  <============\r", VitSynched);
#endif
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    int StartTest               = TRIAL_OFFSET;

    //----------
    // DECODE
    //----------
    //  Uncomment to prevent Phase testing
    if (VitSynched==0){
        //We are not synched, so there's no point in decoding all bits, since we have
        // a 50% chance of being wrong
        // So we decode a small number of packets:  1000
        int  TestPkts=TRIALBITS;
        if ( channel_length < TestPkts){
            TestPkts=channel_length;
        }
        // Discard the Initial Bits for testing

        generated_output_bits       = TestPkts/2 - DEPTH_TRELLIS;
        if (StartTest + TestPkts > channel_length){
            StartTest = channel_length - TestPkts-2;
        }
#ifdef DISPLAY_RESYNC
        printf("StartTest=%i \t TestBits= %i OUTBits=%i \t\t ChannLength=%i \n",
               StartTest,TestPkts,generated_output_bits, channel_length);
#endif
        // Input has length  = channel_length
        Ipp32f          *myShiftedInput;
        myShiftedInput  = ippsMalloc_32f(channel_length);
        // Generated Output Bits has half the length of the channel
        Ipp16s          *myEncodedBits;
        // This is a temporary vector, used to encode the previously decoded test bits
        myEncodedBits  = ippsMalloc_16s(2* generated_output_bits);

        //----------------------------------------------------------------------------
        // TEST SYNCH
        //------------
        if (d_modulation==2){ // The other possibility is the correct one

            int TestVitSynched =0;
            for (int ii=0; ii<TestPkts ; ii++){
                myShiftedInput[ii] = in_vec[ii+StartTest];
            }
            begin_state_tmp     =0;
            //sdvd_intel( TestPkts, myShiftedInput, vit_out	,&VitSynched,&begin_state_tmp);
            sdvd_test_intel( TestPkts, myShiftedInput, vit_out	,&TestVitSynched,&begin_state_tmp);
            // vit_out has length= TestPkts/2
            //----------------------
            // PHASE 0: No Rotation
            //----------------------
            int SumErrors[2]={0,0};
            // Recall modulation here is   y=1-2*x ==> 0->1, 1->-1
            //  generated_output_bits = Bits to encode
            begin_state_tmp     =0;
            // Input for Encoder    = generated_output_bits
            // Output from Encoder  =  VitOut of size 2*generated_output_bits
            generated_output_bits       = TestPkts/2 - DEPTH_TRELLIS;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            int SynchThreshold    =int (floor( (generated_output_bits*2)*d_vthresh)); //JR Used to be 10.0
            int SumTmp=0;
            for (int ii=0; ii<  2*generated_output_bits; ii++){ // Compare Encoded PKTS
                SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
#ifdef DEBUG0
                if ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]){
                    printf("AB<-(A,B) ERROR  %i : in=%f->%i  \t Encoded=%i \n",
                           ii,myShiftedInput[ii],myShiftedInput[ii]<0, myEncodedBits[ii]);
                }
#endif
            }

            SumErrors[0]=SumTmp;
#ifdef DISPLAY_RESYNC
            printf("Case 0: AB<-(A,B) Sum Errors = %i \t  Threshold=%i \t BER=%4.3e \tBegining State=%i  \n",
                   SumTmp,SynchThreshold,float(SumTmp)/float(2*(generated_output_bits-DEPTH_TRELLIS)),begin_state_tmp);
#endif
            //----------------------------------------------------------------------------------------
            // PHASE 1:     Q,I:   No Rotation + Time Shift
            //----------------------------------------------
            for (int ii=0; ii<TestPkts ; ii++){
                myShiftedInput[ii]   = in_vec[ii+1 +StartTest];
            }
            //Decode Shifted Input: In=myShiftedInput, Out= vit_out
            //sdvd_intel( TestPkts, myShiftedInput,vit_out,&VitSynched,&begin_state_tmp);
            // We use the Test version, so that we don't loose the state information from "real" decoding stream
            sdvd_test_intel( TestPkts, myShiftedInput,vit_out,&TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            SumTmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("Case 1: 0deg+ AB<-(B,A)\t Errors = %i \t  Threshold=%i Bits%i \tBegining State=%i  \n",
                   SumTmp,SynchThreshold,2* (generated_output_bits-DEPTH_TRELLIS),begin_state);
#endif
            SumErrors[1]=SumTmp;
            //if (SumErrors[0]  < SumErrors[1]){
            if ((SumErrors[0]  < SynchThreshold) && (SumErrors[0]  < SumErrors[1])){
#ifdef DISPLAY_RESYNC
                printf("-----------------------------------------\n");
                printf("Viterbi Decoder: NO PHASE SHIFT REQUIRED \n");
                printf("-----------------------------------------\n");
#endif
                // We are synched !
                // Decode All PKTS
                begin_state=0;
                generated_output_bits    = channel_length/2; // This is already an even number
#ifdef RESYNC
                // EV TEST
                VitSynched     = (OnceInSynch>0);
                sdvd_intel( channel_length, in_vec, vit_out	,&VitSynched,&begin_state);
                VitSynched     =1;
                OnceInSynch ++;
                CurrentState =0;
#else
                // Use Traditional Viterbi Decoding Function
                VitSynched     =0;
                sdvd_intel( channel_length, in_vec, vit_out	,&VitSynched,&begin_state);
                VitSynched     =1;
#endif


            }else if ((SumErrors[1]  < SynchThreshold) && (SumErrors[1]  < SumErrors[0])){
                // We are NOT synched !!
#ifdef DISPLAY_RESYNC
                printf("---------------------------------------\n");
                printf("Viterbi Decoder: SHIFT 180(deg) \n");
#endif
                //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                for (int ii=0; ii<channel_length-1 ; ii++){
                    myShiftedInput[ii] = in_vec[ii+1];
                }
                // DECODE WHOLE VECTOR
                begin_state=0;
                generated_output_bits    = channel_length/2-1;
                channel_length           = channel_length -1;
#ifdef RESYNC
                // EV TEST
                //VitSynched     =1;
                VitSynched     = (OnceInSynch>0);
                sdvd_intel( channel_length, myShiftedInput,vit_out,&VitSynched,&begin_state);
                VitSynched     =1;
                OnceInSynch ++;
                CurrentState =1;
#else
                // Use Traditional Viterbi Decoding Function
                VitSynched     =0;
                sdvd_intel( channel_length, myShiftedInput,vit_out,&VitSynched,&begin_state);
                VitSynched     =1;
               
#endif
            }


#ifdef RESYNC
            if ((SumErrors[1] > SynchThreshold)&&(SumErrors[0] > SynchThreshold)){
                // We are not Synched !
                ConsecutiveSynchs = 0;
                OnceInSynch=0;
            }
            /*
            if (OnceInSynch==1){
                // GENERATE OUTPUT BITS: The 1st time, the output vector is smaller
                for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS; nn ++){
                    vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
                }
            }else{
                // GENERATE OUTPUT BITS
                // If the orientatin switched, due to a cycle slip or something
                // there may be some errors due to the resync
                for (int nn = BEGINN; nn <= generated_output_bits; nn ++){
                    vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
                }
            }
            */

            if (CurrentState!=PreviousState){
                printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
                printf("Change Lock state: %i --> %i \n",PreviousState, CurrentState);
                printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
            }else{
                //printf("Viterbi Rotation Case= %i \r", CurrentState);
            }

            // GENERATE OUTPUT BITS: The 1st time, the output vector is smaller
            //for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS*(OnceInSynch==1); nn ++){
            for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS*(CurrentState!=PreviousState); nn ++){
                vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
            }
            ippsFree(myEncodedBits);
            ippsFree(myShiftedInput);
            consume_each (channel_length);
            delete[] vit_out;
            if (OnceInSynch==1){// Goes in Here, only the very 1st time
                return generated_output_bits-DEPTH_TRELLIS;
            }else{
                OnceInSynch = 10;  // Avoid Saturation of the variable by doing ++ every run
                return generated_output_bits;
            }

            
#else
            // GENERATE OUTPUT BITS
            for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS; nn ++){
                vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
            }
            ippsFree(myEncodedBits);
            ippsFree(myShiftedInput);
            consume_each (channel_length);
            delete[] vit_out;
            return generated_output_bits-DEPTH_TRELLIS;
#endif

        }else if (d_modulation==4){ // 0QPSK case  --> Many more to analyze
            // ----------
            // OQPSK CASE
            // ----------
            int SumErrors[9], SumTmp;
            int TestVitSynched =0;

            generated_output_bits   = TestPkts/2 - DEPTH_TRELLIS;
            int SynchThreshold    =int (floor( (generated_output_bits*2)*d_vthresh)); //JR Used to be 10.0
            //------------------------------
            // PHASE 1:     I,Q: NO ROTATION
            //------------------------------
            begin_state_tmp=0;
            // We use the Test version, so that we don't loose the state information from "real" decoding stream
            //   sdvd_intel( TestPkts, &in_vec[StartTest], vit_out &VitSynched,     &begin_state_tmp);
            sdvd_test_intel( TestPkts, &in_vec[StartTest], vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                //SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
                SumTmp += ( (in_vec[ii+StartTest] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#1# 0deg   \t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i  \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state);
#endif
            SumErrors[0]=SumTmp;
            //----------------------------------------------------------------------------------------
            // PHASE 2:     Q,I:
            //-----------------------------------
            begin_state_tmp=0;
            //   sdvd_intel( TestPkts-1, &in_vec[1+StartTest] ,vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel( TestPkts-1, &in_vec[1+StartTest], vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                // SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
                SumTmp += ( ( in_vec[ii+1+StartTest] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#2# 0deg+Shft\t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i\n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state);
#endif
            SumErrors[1]=SumTmp;

            //------------------------------------------------
            // PHASE 3:     90 Rotation  (-Q0,I0);(-Q1,I1);
            //-------------------------------------------------
            //Bits Processed: mSI[0], ... , mSI[2*(N/2-1)+1]=mSI[N-1]  --> total N=channel_length
            // ii   =    0      1       CL/2-2      CL/2-1
            // mSI  =   0,1     2,3     CL-4,CL-3   CL-2,CL-1
            // IN   =   1,0     3,2     CL-3,CL-4   CL-1,CL-2
            //
            for (int ii=0; ii<=TestPkts/2-1 ; ii++){
                myShiftedInput[2*ii]   =  -in_vec[2*ii +1 +StartTest];
                myShiftedInput[2*ii+1] =   in_vec[2*ii    +StartTest];
            }
            begin_state_tmp=0;
            //   sdvd_intel( TestPkts, myShiftedInput, vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel( TestPkts, myShiftedInput, vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#3# 90deg    \t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[2]=SumTmp;
            //-------------------------------------------------------
            // PHASE 4:   90 Rotation + Time.SHFT (I0,-Q1);(I1,-Q2);
            //-------------------------------------------------------
            // Bit offset will be considered only when the whole sequence is decoded
            // Right now we can always have an even number of samples
            //STEP 1: Case 3
            begin_state_tmp=0;
            //   sdvd_intel(TestPkts-1, &myShiftedInput[1], vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel(TestPkts-1, &myShiftedInput[1], vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii+1] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#4# 90deg+Shft\t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[3]=SumTmp;

            //----------------------------------
            // PHASE 5:    180 Rotation -I,-Q
            //----------------------------------
            for (int ii=0; ii<TestPkts ; ii++){
                myShiftedInput[ii]   =  -in_vec[ii+1+StartTest];
            }
            begin_state_tmp=0;
            //   sdvd_intel(TestPkts, myShiftedInput, vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel(TestPkts, myShiftedInput, vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#5# 180deg    \t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i \n"
                   ,SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[4]=SumTmp;
            //----------------------------------
            // PHASE 6:    180 + Shft
            //----------------------------------
            begin_state_tmp=0;
            //   sdvd_intel( TestPkts-1, &myShiftedInput[1], vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel( TestPkts-1, &myShiftedInput[1], vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii+1] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#6# 180deg+Shft\t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i  \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[5]=SumTmp;
            //-------------------------------------------------------
            // PHASE 7:   270 Rotation: (Q0,-I0);(Q1,-I1);...
            //---------------------------------------------------------
            /*
            for (int ii=0; ii<=TestPkts/2-1 ; ii++){
                myShiftedInput[2*ii]   =   in_vec[2*ii+1+StartTest];
                myShiftedInput[2*ii+1] =  -in_vec[2*ii+StartTest];
            }
            SumTmp=0;
            begin_state_tmp=0;
            //   sdvd_intel(TestPkts, myShiftedInput, vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel(TestPkts, myShiftedInput, vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii ] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#7# 270deg     \t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[6]=SumTmp;
            */
            SumErrors[6]=MAXINT;
            //---------------------
            // PHASE 8:   270 Rotation+Time SHFT: (-I0,Q1);(-I1,Q2);...
            //---------------------
            begin_state_tmp=0;
            //   sdvd_intel( TestPkts-1, &myShiftedInput[1], vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel( TestPkts-1, &myShiftedInput[1], vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii+1] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#8# 270deg+Shft\t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i\n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[7]=SumTmp;
            //------------------------------------------------------------------
            // Some Extra cases  --> Not physical ??
            // Jeremy's Case 5: (I1,Q0)(I2,Q1)  --> Flip I,Q, then discard one sample
            //------------------------------------------------------------------------
            for (int ii=0; ii<=TestPkts/2-1 ; ii++){
                myShiftedInput[2*ii]   =     in_vec[2*ii +2 +StartTest];
                myShiftedInput[2*ii+1] =     in_vec[2*ii +1 +StartTest];
            }
            begin_state_tmp=0;
            //   sdvd_intel( TestPkts, myShiftedInput, vit_out, &VitSynched,    &begin_state_tmp);
            sdvd_test_intel( TestPkts, myShiftedInput, vit_out, &TestVitSynched,&begin_state_tmp);
            begin_state_tmp=0;
            cnv_encd_intel(generated_output_bits, vit_out, myEncodedBits,&begin_state_tmp) ;
            SumTmp=0;
            for (int ii=0; ii< 2*generated_output_bits; ii++){
                SumTmp += ( (myShiftedInput[ii] < 0)!=myEncodedBits[ii]);
            }
#ifdef DISPLAY_RESYNC
            printf("#A# 90deg    \t Errors = %i \t  Threshold=%i \tBits=%i \t Begining State=%i \n",
                   SumTmp,SynchThreshold,2*(generated_output_bits),begin_state_tmp);
#endif
            SumErrors[8]=SumTmp;
            //------------------------------------------------
            //------------------------------------------------
            //Choose Best Case
            int BestCase =-1;
            SumTmp=10000000;
            for (int ii=0;ii<9;ii++){
                if (SumErrors[ii] < SumTmp){
                    BestCase=ii;
                    SumTmp=SumErrors[ii];
                }
            }
#ifdef FORCE_SYNC
            if (SumTmp > SynchThreshold ) {
                printf("Viterbi UNLOCKED:  Errors=%i, Max.Allowed=%i \n", SumTmp,SynchThreshold);
                VitSynched = 0;
                #ifdef RESYNC
                    ConsecutiveSynchs = 0;
                    OnceInSynch=0;
                #endif
            }else{
                //printf("Viterbi Rotation Case= %i \n", BestCase+1);
                if (CurrentState!=PreviousState){
                    printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
                    printf("Change Lock state: %i --> %i \n",PreviousState, CurrentState);
                    printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
                }
                VitSynched=BestCase+1;
            }
#else
            printf("Viterbi Rotation Case= %i \n", BestCase+1);
            VitSynched=BestCase+1;
            CurrentState =BestCase+1;

            if (CurrentState!=PreviousState){
                printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
                printf("Change Lock state: %i --> %i \n",PreviousState, CurrentState);
                printf("-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n");
            }
            
#endif
            //------------------------------------------------------------------
            //----------------------------------------------------------------
            // DECODE THE WHOLE VECTOR, not only the test packets
            generated_output_bits    = channel_length/2;
            begin_state =0;

            if (VitSynched==1){    //        	AB<-(A,B)
                //------------------------------
                // PHASE 1:     I,Q: NO ROTATION
                //------------------------------
                generated_output_bits    = channel_length/2;
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length, in_vec, vit_out	,&VitSynched,&begin_state);
                VitSynched =1;

            }else if (VitSynched==2){
                //----------------------------------------------
                // PHASE 2:     Q,I:   No Rotation + Time Shift
                //----------------------------------------------

#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif

                sdvd_intel( channel_length-1, &in_vec[1], vit_out	,&VitSynched,&begin_state);
                VitSynched =1;

            }else if (VitSynched==3){
                //------------------------------------------------
                // PHASE 3:     90 Rotation  (-Q0,I0);(-Q1,I1);
                //-------------------------------------------------
                //Bits Processed: mSI[0], ... , mSI[2*(N/2-1)+1]=mSI[N-1]  --> total N=channel_length
                // ii   =    0      1       CL/2-2      CL/2-1
                // mSI  =   0,1     2,3     CL-4,CL-3   CL-2,CL-1
                // IN   =   1,0     3,2     CL-3,CL-4   CL-1,CL-2
                //
                for (int ii=0; ii<=channel_length/2-1 ; ii++){
                    myShiftedInput[2*ii]   =  -in_vec[2*ii +1 ];
                    myShiftedInput[2*ii+1] =   in_vec[2*ii    ];
                }
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
                VitSynched =3;

            }else if (VitSynched==4){
                //-------------------------------------------------------
                // PHASE 4:   90 Rotation + Time.SHFT (I0,-Q1);(I1,-Q2);
                //-------------------------------------------------------
                /*
                for (int ii=0; ii<=channel_length/2-1 ; ii++){
                    myShiftedInput[2*ii]   =  -in_vec[2*ii ];
                    myShiftedInput[2*ii+1] =   in_vec[2*ii+3 ];
                }
                // ii   =    0      1       2       CL/2-2      CL/2-1
                // mSI  =   0,1     2,3     4,5     CL-4,CL-3   CL-2,CL-1
                // IN   =   0,3     2,5     4,7     CL-4,CL-1   CL-2,CL+1
                //
                // Too Messy !!!! --> Need to deal with CL+1 issue
                */

                // STEP 1: Do case 3
                // ii   =    0      1       CL/2-2      CL/2-1
                // mSI  =   0,1     2,3     CL-4,CL-3   CL-2,CL-1
                // IN   =   1,0     3,2     CL-3,CL-4   CL-1,CL-2
                for (int ii=0; ii<=channel_length/2-1 ; ii++){
                    myShiftedInput[2*ii]   =  -in_vec[2*ii +1 ];
                    myShiftedInput[2*ii+1] =   in_vec[2*ii    ];
                }
                //STEP 2: Throw Away the 1st bit
                //channel_length           = channel_length -1;
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length-1, &myShiftedInput[1], vit_out	,&VitSynched,&begin_state);
                VitSynched=3;

            }else if (VitSynched==5){
                //----------------------------------
                // PHASE 5:    180 Rotation -I,-Q
                //----------------------------------
                for (int ii=0; ii<channel_length ; ii++){
                    myShiftedInput[ii]   =  -in_vec[ii];
                }
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
                VitSynched =5;
            }else if (VitSynched==6){
                //----------------------------------------------------------------
                // PHASE 6:   Time Shft + 180 Rotation: (-Q0,-I1);(-Q1,-I2);...
                //---------------------------------------------------------------
                //STEP 1: Case 5
                for (int ii=0; ii<channel_length-1 ; ii++){
                    myShiftedInput[ii]   =  -in_vec[ii+1];
                }
                //Step 2
                //channel_length           = channel_length -1;
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel(channel_length-1, myShiftedInput, vit_out	,&VitSynched,&begin_state);
                VitSynched =5;

            }else if (VitSynched==7){
                //-------------------------------------------------------
                // PHASE 7:   270 Rotation: (Q0,-I0);(Q1,-I1);...
                //---------------------------------------------------------
                for (int ii=0; ii<=channel_length/2-1 ; ii++){
                    myShiftedInput[2*ii]   =   in_vec[2*ii+1];
                    myShiftedInput[2*ii+1] =  -in_vec[2*ii];
                }
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length, in_vec, vit_out	,&VitSynched,&begin_state);
                VitSynched =7;
            }else if (VitSynched==8){
                //---------------------
                // PHASE 8:   270 Rotation+Time SHFT: (-I0,Q1);(-I1,Q2);...
                //---------------------
                for (int ii=0; ii<=channel_length/2-1; ii++){
                    myShiftedInput[2*ii]   =   in_vec[2*ii+1 ];
                    myShiftedInput[2*ii+1] =  -in_vec[2*ii  ];
                }
                //channel_length           = channel_length -1;
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length-1, &myShiftedInput[1], vit_out	,&VitSynched,&begin_state);
                VitSynched =7;
            }else if (VitSynched==9){
                //---------------------
                // I,Q Swap + Shift
                //---------------------
                for (int ii=0; ii<channel_length/2-1; ii++){
                    myShiftedInput[2*ii]   =   in_vec[2*ii+2 ];
                    myShiftedInput[2*ii+1] =   in_vec[2*ii+1  ];
                }
#ifdef RESYNC
                VitSynched     = (OnceInSynch>0);
#else
                VitSynched =0;
#endif
                sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
                VitSynched =9;
            }

            generated_output_bits    = channel_length/2;

#ifdef RESYNC
            if (VitSynched >0){
                OnceInSynch ++;
                // GENERATE OUTPUT BITS: The 1st time, the output vector is smaller (CurrentState!=PreviousState)
                //for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS*(OnceInSynch==1); nn ++){
                for (int nn = BEGINN; nn <= generated_output_bits-DEPTH_TRELLIS*(CurrentState!=PreviousState); nn ++){
                    vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
                }
                ippsFree(myEncodedBits);
                ippsFree(myShiftedInput);
                consume_each (channel_length);
                delete[] vit_out;
                /*
                if (OnceInSynch==1){// Goes in Here, only the very 1st time
                    return generated_output_bits-DEPTH_TRELLIS;
                }else{
                    OnceInSynch = 10;  // Avoid Saturation of the variable by doing ++ every run
                    return generated_output_bits;
                }
                */
                if ((CurrentState!=PreviousState)){// Goes in Here, only the very 1st time
                    return generated_output_bits-DEPTH_TRELLIS;
                }else{
                    OnceInSynch = 10;  // Avoid Saturation of the variable by doing ++ every run
                    return generated_output_bits;
                }
            }else{
                OnceInSynch =0;
                // Couldn't SYNCH
                ippsFree(myEncodedBits);
                ippsFree(myShiftedInput);
                delete[] vit_out;
                //  delete[] myShiftedInput;
                consume_each ( (channel_length/2)*2-1);
                //consume_each (1001);
                return 0;
            }

#else
            //FORCE SYNC
            if (VitSynched >0){
                for (int nn = BEGINN; nn <=generated_output_bits-DEPTH_TRELLIS; nn ++){
                    vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
                }
                ippsFree(myEncodedBits);
                ippsFree(myShiftedInput);
                delete[] vit_out;
                //  delete[] myShiftedInput;
                consume_each (channel_length);
                return generated_output_bits-DEPTH_TRELLIS;
            }else{
                // Couldn't SYNCH
                ippsFree(myEncodedBits);
                ippsFree(myShiftedInput);
                delete[] vit_out;
                //  delete[] myShiftedInput;
                consume_each ( (channel_length/2)*2-1);
                //consume_each (1001);
                return 0;
            }
#endif
            ///----------------------------------------------------------------

        } // END MOD=4=QPSK

    }else{
        // ALREADY SYNCHED
        // -----------------------------------------------------------------
        //   NOW We are synched
        // -----------------------------------------------------------------
        generated_output_bits    = channel_length/2;
        Ipp32f                   *myShiftedInput;
        myShiftedInput           = ippsMalloc_32f(channel_length);
        if (VitSynched==1){    //        	AB<-(A,B)
            //------------------------------
            // PHASE 1:     I,Q: NO ROTATION
            //------------------------------
            sdvd_intel( channel_length, in_vec, vit_out	,&VitSynched,&begin_state);
        }else if (VitSynched==3){
            //---------------------
            // PHASE 3:     90 Rotation  (-Q0,I0);(-Q1,I1);
            //---------------------
            for (int ii=0; ii<=channel_length/2-1 ; ii++){
                myShiftedInput[2*ii]   =  -in_vec[2*ii +1 ];
                myShiftedInput[2*ii+1] =   in_vec[2*ii    ];
            }
            sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);

        }else if (VitSynched==5){
            //----------------------------------
            // PHASE 5:    180 Rotation -I,-Q
            //----------------------------------
            for (int ii=0; ii<channel_length ; ii++){
                myShiftedInput[ii]   =  -in_vec[ii];
            }
            sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
        }else if (VitSynched==7){
            //-------------------------------------------------------
            // PHASE 7:   270 Rotation: (Q0,-I0);(Q1,-I1);...
            //---------------------------------------------------------
            for (int ii=0; ii<=channel_length/2-1 ; ii++){
                myShiftedInput[2*ii]   =   in_vec[2*ii+1];
                myShiftedInput[2*ii+1] =  -in_vec[2*ii];
            }
            sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
        }else if (VitSynched==9){
            //-------------------------------------------------------------
            // IQ Shift and delay :  (2,1)(4,3)(6,5), ... (N-2,N-3)
            //-------------------------------------------------------------
            for (int ii=0; ii< channel_length/2-1 ; ii++){
                myShiftedInput[2*ii]   =   in_vec[2*ii+2];
                myShiftedInput[2*ii+1] =   in_vec[2*ii+1];
            }
            sdvd_intel( channel_length, myShiftedInput, vit_out	,&VitSynched,&begin_state);
        }
        ippsFree(myShiftedInput);
    }
    for (int nn = BEGINN; nn <=generated_output_bits; nn ++){
        vit_bytes[nn]=(vit_out[nn] & 0x0001) ;
    }

    delete[] vit_out;
    consume_each (channel_length);
    return (generated_output_bits);
}
