#include <cmath>
#include "ari_rate_change_factor.h"

using namespace std;

/*! This recursive function tries all combinations of factors from 1 to
 max_factor at each stage.
*/
vector<vector<int> > get_all_factors(int n, int max_factor)
{
    int num_stages = ceil(log((float)n) / log((float)max_factor) + 
                          1.0e-20); // don't want to get a zero
    
    vector<vector<int> > factors;
    if (num_stages == 1) // last stage, just return the remainder
    {
        factors.resize(1);
        factors[0].push_back(min((int)n, max_factor));
    } else {
        for (int f=2; f<=max_factor; f++)
        {
            // Recursively process the remainder (n/f)
            vector<vector<int> > sub_factors = get_all_factors(n/f, max_factor);
            int num_subfactors = sub_factors.size();
            for (int i=0; i<num_subfactors; i++)
            {
                // Don't forget the factor we just divided by
                sub_factors[i].push_back(f);
                factors.push_back(sub_factors[i]);
            }
        }
    }

    return factors;
}

vector<int> ari_rate_change_factor(int n, int max_factor)
{
    // Get all the combinations of values 1:n such that the
    // product is less than n
    vector<vector<int> > all_factors = get_all_factors(n, max_factor);

    // Compute the total rate change factor from this factorization.
    vector<int> p(all_factors.size(), 1);
    for (int i=0; i<all_factors.size(); i++)
        for (int j=0; j<all_factors[i].size(); j++)
            p[i] = p[i] * all_factors[i][j];

    // Compute errors relative to the desired rate change factor
    vector<float> err(all_factors.size());
    for (int i=0; i<all_factors.size(); i++)
        err[i] = 1 - (float) p[i] / (float) n;

    // Find the indices with minimum error (there may be more than one)
    vector<int> good_idx;
    float min_error = 1.0;
    for (int i=0; i<all_factors.size(); i++)
    {
        if (err[i] < min_error)
        {
            if (good_idx.size()) // clear the previous best
                good_idx.erase(good_idx.begin(), good_idx.end());
            min_error = err[i];
        }
        if (err[i] == min_error) // This index is as good as the best
            good_idx.push_back(i);
    }

    // Find the factorizations with the fewest stages.  Trivial (1)
    // stages are ignored.
    int min_non_ones = all_factors[0].size();
    int best_idx = good_idx[0];
    for (int i=0; i<good_idx.size(); i++)
    {
        vector<int> factors = all_factors[good_idx[i]];
        // check how many non-ones are here, and pick the least
        int non_ones = factors.size();
        for (int j=0; j<factors.size(); j++)
            if (factors[j] == 1) non_ones--;

        if (non_ones < min_non_ones)
        {
            min_non_ones = non_ones; 
            best_idx = good_idx[i];
        }
    }

    // Now, copy the best factorization into the output value.  Remove
    // all trivial (ones) stages
    vector<int> factors;
    for (int i=0; i<all_factors[best_idx].size(); i++)
        if (all_factors[best_idx][i] != 1)
            factors.push_back(all_factors[best_idx][i]);

    return factors;
}

