// This file defines some functions to use GNU radio without Python

#include <gr_buffer.h>
#include <gr_io_signature.h>
#include <gr_block_detail.h>
#include <gr_single_threaded_scheduler.h>
#include "gnuradio.h"

#include <windows.h> // Needed for CreateThread

using namespace std;

const int DEFAULT_BUFFER_ITEMS = (1<<15); // 32k

// Gnuradio has functions to traverse the block tree and determine the 
// number of input and output ports used for each block.  I can't use these
// functions since they are written in Python. 
// The 'create_detail' function is called the first time a block is seen.
// This function creates a default 'block_detail' with the minimum 
// possible number of input and output ports.

// Every time a block is used as a source, the function 'check_output_ports'
// is called.  In that function we check if the currently addressed output
// port is less or equal to the number of output ports already defined in
// the block's detail.  If not, a new detail is created, data from the old
// detail is copied.
void create_detail(gr_block_sptr block)
{
	int ninputs  = block->input_signature()->min_streams();
	int noutputs = block->output_signature()->min_streams();

	gr_block_detail_sptr detail = gr_make_block_detail (ninputs, noutputs);
	block->set_detail(detail);
}


void check_detail(gr_block_sptr block, int in_port, int out_port)
{
  gr_block_detail_sptr old_detail = block->detail();

	int ninputs  = old_detail->ninputs();
	int noutputs = old_detail->noutputs();
  if (out_port < noutputs && in_port < ninputs) return;

  noutputs = out_port + 1; // Counting from 0
  ninputs  = in_port  + 1;

  // Create a new detail, with 'port+1' number of ports
	gr_block_detail_sptr new_detail = gr_make_block_detail (ninputs, noutputs);

  // Copy data from old_detail into the new_detail
  // Inputs 
  for (int i=0; i<old_detail->ninputs(); i++)
    new_detail->set_input(i, old_detail->input(i));
  // Outputs
  for (int i=0; i<old_detail->noutputs(); i++)
    new_detail->set_output(i, old_detail->output(i));

  // Set the new detail
	block->set_detail(new_detail);

}


// This function will presumably be called at least once for each block 
// and each port.  It will:
// 1. Initialize the detailed structure of a block, the first time that
//    block is encountered
// 2. Create output buffer for an output port, the first time that
//    output port is encountered
// 3. Connect the source and destination ports
// 4. Keep a list of unique blocks that can then be passed to the scheduler.
void connect_blocks(gr_block_sptr src,  int src_port, 
					gr_block_sptr dest, int dest_port,
					vector<gr_block_sptr> &blocks)
{
	// A block has a 'block_detail' associated with it.  The detail is
	// not created automatically since it depends on the actual number of
	// ports that are _used_.  Create it here the first time a block is
	// encountered.  Also add that block to the list of known blocks.
	if (!src->detail())  {create_detail(src);  blocks.push_back(src);}
	if (!dest->detail()) {create_detail(dest); blocks.push_back(dest);}

  check_detail(src, src->detail()->ninputs()-1, src_port);
  check_detail(dest, dest_port, dest->detail()->noutputs()-1);

	// Assign Buffers.  Determine the buffer sizes to use, allocate them and
	// attach to detail.  We should really compute an appropriate buffer size
	// based on the function in 'flow_graph.py'.  For now, just make the size
	// DEFAULT_BUFFER_ITEMS items (each item is probably complex float = 8 B)
	// Note that an output buffer may have multiple readers.  Thus, create
	// a buffer only if it is not already there. 
	if (!src->detail()->output(src_port))
	{
		// The 'special' blocks are those that interface to hardware.  Give 
		// them a larger buffer.  This takes care of both source and sink 
		// hardware blocks.
		int buffer_size_mul = (src->special() || dest->special()) ? 4 : 1;
        int item_size = src->output_signature()->sizeof_stream_item(src_port);
        // Don't allocate more than DEFAULT_BUFFER_ITEMS*8 bytes, but at least
        // one sizeof_stream_item
        int num_items = min(DEFAULT_BUFFER_ITEMS*8/item_size, DEFAULT_BUFFER_ITEMS);
        num_items = max(num_items, 1);
		gr_buffer_sptr buffer =
			gr_make_buffer (num_items * buffer_size_mul, item_size);
		src->detail()->set_output(src_port, buffer);
	}

	// Create a reader that will pull data from the newly created buffer into
	// the destination
	gr_buffer_reader_sptr reader =
		gr_buffer_add_reader (src->detail()->output(src_port), 
		                      dest->history()-1);
  dest->detail()->set_input(dest_port, reader);
}


DWORD WINAPI _start_scheduler_thread( LPVOID pParam )
{
	((gr_single_threaded_scheduler *)pParam)->run();
    return 0;
}


gr_single_threaded_scheduler_sptr 
	start_scheduler_thread(std::vector<gr_block_sptr> &blocks, bool threaded)
{
	// ------------------- Create a single-threaded scheduler 
	gr_single_threaded_scheduler_sptr scheduler = 
		gr_make_single_threaded_scheduler (blocks);
	
	if (threaded)
		CreateThread(0, 0, _start_scheduler_thread, scheduler.get(), 0, 0);
	else
		scheduler->run();

	return scheduler;
}

