#include <gr_block.h>
#include <vector>
#include <gr_single_threaded_scheduler.h>

//! Connect two GNUradio blocks.  The last parameter maintains a list of all
//! connected blocks to be passed to create scheduler.
void connect_blocks(gr_block_sptr src,  int src_port, 
					gr_block_sptr dest, int dest_port,
					std::vector<gr_block_sptr> &blocks);

//! Start a scheduler thread.  This function returns a 'smart pointer' to the
//! created scheduler.  This pointer MUST be saved until the scheduler is 
//! stopped.  Otherwise, the auto-cleanup functionality of smart pointers will
//! delete the scheduler (no references left).
//! \param blocks List of blocks created using the connect_blocks function
//! \param threaded True to start a new thread for this scheduler.  If true,
//!                 this function returns immediately.  False to run the
//!                 scheduler from within the funciton.  If false, blocks
gr_single_threaded_scheduler_sptr 
	start_scheduler_thread(std::vector<gr_block_sptr> &blocks, bool threaded);
/*! 
GNU radio is a very elegant but somewhat complex system.  The current
version is designed to use a combination of C++ and a scripting language
called 'Python.'  It is clearly a Unix/Linux animal.  There's no full
Windows port at this time.  The next few paragraphs attempt to augment
the very sparse documentation that comes with GNU radio.  In particular,
I found no documentation at all about the guts of the system.  You don't 
need to understand how it works if you plan to use Python, but I had to
re-create the Python runtime environment in C++.

The basic unit is called a 'gr_block.'  It is very general and can be 
used to implement anything.  It has an arbitrary number of inputs and 
outputs and may have an arbitrary rate relationship between input and output.

A gr_block contains a 'gr_block_detail.'  A gr_block_detail is where all
the real magic happens.  The gr_block_detail has one gr_buffer_reader for
each input and one gr_buffer for each output.  The gr_buffer is a very 
clever ring buffer that somehow allows linear writes without overflowing.
The details are very system-specific.  Each gr_buffer is allocated with 
some fixed (relatively large) size > 64kB.  Multiple gr_buffer_readers can
be attached to a gr_buffer.  Note that each reader gets an independent 
'virtual' copy of the buffer. I.e. if you have 2 readers attached to a buffer,
the buffer does not empty any faster than the fastest reader. 

The blocks are connected by associating gr_buffer_reader's with gr_buffers.

A scheduler 'gr_single_threaded_scheduler_sptr' is responsible for monitoring
the levels in all the buffers to make sure they don't overflow or underflow.
The flow control is conceptually 'pull', where the downstream readers ask the
upstream buffers for data.

The compile environment for GNU radio using Visual Studio .NET 2003 is:
1. Preprocessor Defines
_USE_MATH_DEFINES;HAVE_CONFIG_H;_AFXEXT
2. Header directories
./
\projects\SDR\gnuradio\gnuradio-core\src\lib\omnithread
/projects/SDR/gnuradio/gnuradio-core/src/lib/filter
/projects/SDR/gnuradio/gnuradio-core/src/lib/io
/projects/SDR/gnuradio/gnuradio-core/src/lib/general
/projects/SDR/gnuradio/gnuradio-core/src/lib/gengen
/projects/SDR/gnuradio/gnuradio-core/src/lib/runtime
/projects/SDR/boost_1_34_0

*/