#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

// TODO: Add a "list" command (on opening a new connection, send "LIST" to get a list of all names)

#include "tcp_debug_server.h"
#include "tcp_debug_constants.h"
#include <assert.h>
#include <iostream>
#include <map>
#include <string>
#include <omnithread.h>
#include <windows.h> // For Windows networking
#include <winsock.h>
//#include <sys/socket.h> // Cross-platform networking
//#include <arpa/inet.h>  // ???
#include "tcp_debug_sink.h"

// For some reason Winsock.h doesn't define MSG_WAITALL, despite the fact that the
// MSDN page alludes to it: http://msdn2.microsoft.com/en-us/library/ms740121.aspx
#ifndef MSG_WAITALL
#define MSG_WAITALL 0 //???
#endif

class tcp_debug_sink_info {
  public:
    tcp_debug_sink_info()
        : itemsize(0), typecode(TYPECODE_UNKNOWN), ptr(NULL) {}
    tcp_debug_sink_info(size_t size, int type, tcp_debug_sink* tds)
        : itemsize(size), typecode(type), ptr(tds) {}

    size_t          itemsize;
    int             typecode;
    tcp_debug_sink* ptr;
};

// Global list of all registered tcp_debug_sinks, indexed by name.
typedef std::map<std::string, tcp_debug_sink_info> g_sinks_type;
g_sinks_type g_sinks;
omni_mutex g_sinks_mutex;

// Control for the TCP/IP server socket thread.
// (The actual socket is a local variable in that thread.)
omni_mutex g_server_running_mutex;
bool g_server_running = false;

// Function prototypes
void tcp_debug_server_init();
void tcp_debug_server_close();
DWORD WINAPI tcp_debug_server_listen(LPVOID foo);


void tcp_debug_server_register(const char* name, size_t itemsize, int typecode, tcp_debug_sink* tds)
{
    assert(strlen(name) < MAX_SINK_NAME_LENGTH);

    if(!g_server_running)
        tcp_debug_server_init();

    g_sinks_mutex.lock();
    g_sinks[name] = tcp_debug_sink_info(itemsize, typecode, tds);
    g_sinks_mutex.unlock();
}


void tcp_debug_server_unregister(tcp_debug_sink* tds)
{
    g_sinks_mutex.lock();

    // Can't lookup by pointer, so manually go through the whole list:
    for(g_sinks_type::iterator a = g_sinks.begin() ; a != g_sinks.end() ; ++a) {
        if(a->second.ptr == tds) {
            g_sinks.erase(a);
            break;
        }
    }

    if(g_sinks.empty())
        tcp_debug_server_close();

    g_sinks_mutex.unlock();
}


void tcp_debug_server_init()
{
    // Initialize connection to Windows networking system
	WSADATA wsaData;            // Used to open connection to the Windows networking DLLs
    int errorCode = WSAStartup(0x0101, &wsaData);
    if(errorCode != 0) {
        std::cerr << "Unable to open connection to Windows networking system." << std::endl;
        return;
    }

    // Start the server thread:
    g_server_running = true;
    g_server_running_mutex.lock();
    //???AfxBeginThread((AFX_THREADPROC)tcp_debug_server_listen, NULL);
    CreateThread(NULL, 0, tcp_debug_server_listen, NULL, 0, NULL);
}


void tcp_debug_server_close()
{
    // Close down the server socket and shut down Windows networking
    g_server_running = false;
    g_server_running_mutex.lock(); // Wait for server thread to stop.
    WSACleanup();
}


// Argument "foo" is not used, but we need it to match the default
// thread signature.
DWORD WINAPI tcp_debug_server_listen(LPVOID foo)
{
    // Setup the server socket to listen on the bitrate control port
    SOCKET server_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    struct sockaddr_in server_addr;
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(TCP_DEBUG_SERVER_PORT);
    server_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
    if(bind(server_socket, (SOCKADDR*)&server_addr, sizeof(server_addr)) != 0) {
        std::cerr << "Unable to start/bind bit-rate request server (" << WSAGetLastError() << ").  Port may already be in use." << std::endl;
        closesocket(server_socket);
        g_server_running_mutex.unlock();
        return 1; // Error
    }
    if(listen(server_socket, SOMAXCONN) != 0) {
        std::cerr << "Unable to listen to bitrate-control port (" << WSAGetLastError() << ").  Unknown cause." << std::endl;
        closesocket(server_socket);
        g_server_running_mutex.unlock();
        return 1; // Error
    }

    // Accept incoming connections until told to stop:
    while(g_server_running) {
        SOCKET new_connection = accept(server_socket, NULL, NULL);

        if(new_connection == INVALID_SOCKET) {
            // If invalid, dump the socket and try again...
            closesocket(new_connection);
            continue;
        }

        // The first thing the new connection should do is send a request for
        // a specific debug sink.  Wait for the name:
        // TODO: If expected number of connections increases drastically, spin this
        //       off as a separate thread so we don't stall on accepting connections.
        char name[MAX_SINK_NAME_LENGTH];
        if(recv(new_connection, name, MAX_SINK_NAME_LENGTH, MSG_WAITALL) != MAX_SINK_NAME_LENGTH) {
            // Connection closed or other error -> close it.
            closesocket(new_connection);
            continue;
        }

        // First check if this is a list request:
        if(strcmp(name, REQUEST_SINK_LIST) == 0) {
            // Transmit the number of names, followed by the complete list:
            g_sinks_mutex.lock();
            int name_count = g_sinks.size();
            send(new_connection, (char*)&name_count, sizeof(name_count), 0);
            for(g_sinks_type::iterator a = g_sinks.begin() ; a != g_sinks.end() ; ++a) {
                strcpy(name, a->first.c_str());
                send(new_connection, name, MAX_SINK_NAME_LENGTH, 0);
            }
            g_sinks_mutex.unlock();

            // Close socket and continue listening
            closesocket(new_connection);
            continue;
        }

        // Otherwise, attempt to look up the name.  
        g_sinks_mutex.lock();

        g_sinks_type::iterator lookup = g_sinks.find(std::string(name));
        if(lookup == g_sinks.end()) {
            // Not found -> send back the "error" typecode and close the socket.
            size_t zero = 0;
            send(new_connection, (char*)&zero,           sizeof(zero), 0);
            send(new_connection, (char*)&TYPECODE_ERROR, sizeof(int),  0);
            closesocket(new_connection);
        } else {
            // Found -> send back the typecode and pass the socket on 
            // to the appropriate tcp_debug_sink.
            send(new_connection, (char*)&lookup->second.itemsize, sizeof(lookup->second.itemsize), 0);
            send(new_connection, (char*)&lookup->second.typecode, sizeof(lookup->second.typecode), 0);
            lookup->second.ptr->handoff_socket(new_connection);
        }

        g_sinks_mutex.unlock();
    }

    // Disconnect the server socket
    closesocket(server_socket);

    // Unlock the thread-lock, to indicate thread exit.
    g_server_running_mutex.unlock();
    return 0; // Normal exit
}
