#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <tcp_debug_sink.h>
#include "tcp_debug_server.h"
#include <gr_io_signature.h>
#include <iostream>

tcp_debug_sink::tcp_debug_sink(const char* name, size_t itemsize, int typecode)
    : gr_sync_block ("tcp_debug_sink",
                     gr_make_io_signature(1, 1, itemsize), // One input of specified word-size
                     gr_make_io_signature(0, 0, 0))        // No outputs
    , d_socket(NULL)
    , d_count(0)
    , d_itemsize(itemsize)
{
    tcp_debug_server_register(name, itemsize, typecode, this);
}

tcp_debug_sink_sptr
make_tcp_debug_sink(const char* name, size_t itemsize, int typecode)
{
    return tcp_debug_sink_sptr (new tcp_debug_sink(name, itemsize, typecode));
}

tcp_debug_sink::~tcp_debug_sink ()
{
    // Unregister this debug_sink with the main server:
    tcp_debug_server_unregister(this);

    // If socket is connected, close it
    d_socket_mutex.lock(); // Acquire lock, and never release it.
    if(d_socket)
        close_socket();
}


void tcp_debug_sink::handoff_socket(SOCKET sock)
{
    d_socket_mutex.lock();

    // If socket already in use, close the old connection.
    if(d_socket)
        close_socket();

    // Use the new connection from now on.  Start off in PAUSED mode.
    d_socket = sock;
    d_count  = 0;

    // Set the socket to use non-blocking mode:
    u_long iMode = 1;
    ioctlsocket(d_socket, FIONBIO, &iMode);

    d_socket_mutex.release();
}


void tcp_debug_sink::close_socket()
{
    d_socket_mutex.lock();

    if(d_socket) {
        closesocket(d_socket);
        d_socket = NULL;
    }

    d_socket_mutex.release();
}


// Forward or discard
int tcp_debug_sink::work (int noutput_items,
			    gr_vector_const_void_star &input_items,
			    gr_vector_void_star &output_items)
{ 
    char *iptr = (char*) input_items[0];

    if(!d_socket)
        return noutput_items;  // Discard all data if no connection present

    omni_mutex_lock l(d_socket_mutex); // Acquire mutex, auto-release on function return

    // If client has sent a new request, receive it.
    fd_set fds; FD_ZERO(&fds); FD_SET(d_socket, &fds);
    TIMEVAL timeout; timeout.tv_sec = 0; timeout.tv_usec = 0;
    int selval = select(0, &fds, NULL, NULL, &timeout);
    if(selval != SOCKET_ERROR && fds.fd_count != 0) {
        int req;
        if(recv(d_socket, (char*)&req, sizeof(int), 0) == sizeof(int)) {
            d_count = req;
        }
    }

    // Do we have a pending request for data?  (including the latest request)
    if(d_count == 0)
        return noutput_items; // No client requests -> discard all data.
    // Do we have enough data to service the current request?
    if(d_count > 0 && noutput_items < d_count)
        return 0;              // Wait until a full packet is available

    // Forward requested data:
    int max_words = (d_count > 0) ? d_count : noutput_items;        // If single mode, send requested size.  Otherwise send all available data.
    int bytes_sent = send(d_socket, iptr, max_words*d_itemsize, 0); // Send the data!  (non-blocking)
    if(bytes_sent > 0) {
        // Normal operation:
        int items_sent = bytes_sent / d_itemsize;
        if(d_count > 0)
            d_count -= items_sent;
        return items_sent;
    } else {
        // recv returned an error code:
        if(bytes_sent == WSAEWOULDBLOCK) {
            // If send() call would block, that just means its buffer is full.
            // If we are in continuous mode, return 0 so we don't drop data.
            // Otherwise, feel free to drop data, so return noutput_items.
            if(d_count < 0)
                return 0;             // Keep data, hopefully send it next time around.
            else
                return noutput_items; // Discard all data
        } else {
            // Some other more serious error -> abort the connection
            close_socket();
            return noutput_items; // Discard all data
        }
    }
}
