
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gr_io_signature.h>
#include <gr_prefs.h>
#include <ar_clock_recovery_mm_cc.h>
#include <ari_mmse_fir_interpolator_cc.h>
#include <stdexcept>
#include <limits>

//#include <fstream>
//std::ofstream timing_debug("c:/temp/timing_debug.dat");
// Public constructor


ar_clock_recovery_mm_cc_sptr 
ar_make_clock_recovery_mm_cc(float omega, float gain_omega, float mu, float gain_mu,
			     float omega_relative_limit)
{
  return ar_clock_recovery_mm_cc_sptr (new ar_clock_recovery_mm_cc (omega, 
								    gain_omega, 
								    mu,
								    gain_mu,
								    omega_relative_limit));
}

ar_clock_recovery_mm_cc::ar_clock_recovery_mm_cc (float omega, float gain_omega, float mu, 
						  float gain_mu, float omega_relative_limit)
  : gr_block ("clock_recovery_mm_cc",
	          gr_make_io_signature (1, 1, sizeof (gr_complex)),
	          gr_make_io_signature (1, 2, sizeof (gr_complex))),
               d_mu (mu), d_omega(omega), d_gain_omega(gain_omega), 
               d_omega_relative_limit(omega_relative_limit), 
               d_gain_mu(gain_mu), d_last_sample(0), 
               d_interp(new ari_mmse_fir_interpolator_cc()),
               d_verbose(gr_prefs::singleton()->get_bool("clock_recovery_mm_cc", "verbose", false)),
               d_p_2T(0), d_p_1T(0), d_p_0T(0), d_c_2T(0), d_c_1T(0), d_c_0T(0)
{
  if (omega <= 0.0)
    throw std::out_of_range ("clock rate must be > 0");
  if (gain_mu <  0  || gain_omega < 0)
    throw std::out_of_range ("Gains must be non-negative");

  set_omega(omega);			// also sets min and max omega
  set_relative_rate (1.0 / omega);
  set_history(3);			// ensure 2 extra input sample is available
}

ar_clock_recovery_mm_cc::~ar_clock_recovery_mm_cc ()
{
  delete d_interp;
}

void
ar_clock_recovery_mm_cc::forecast(int noutput_items, gr_vector_int &ninput_items_required)
{
  unsigned ninputs = ninput_items_required.size();
  for (unsigned i=0; i < ninputs; i++)
    ninput_items_required[i] =
      (int) ceil((noutput_items * d_omega) + d_interp->ntaps());
}

gr_complex
ar_clock_recovery_mm_cc::slicer_0deg (gr_complex sample)
{
  float real=0, imag=0;

  if(sample.real() > 0)
    real = 1;
  if(sample.imag() > 0)
    imag = 1;
  return gr_complex(real,imag);
}

gr_complex
ar_clock_recovery_mm_cc::slicer_45deg (gr_complex sample)
{
  float real= -1, imag = -1;
  if(sample.real() > 0)
    real=1;
  if(sample.imag() > 0)
    imag = 1;
  return gr_complex(real,imag);
}

/*
  Modified Mueller and Muller clock recovery circuit
  Based:
     G. R. Danesfahani, T.G. Jeans, "Optimisation of modified Mueller and Muller 
     algorithm,"  Electronics Letters, Vol. 31, no. 13,  22 June 1995, pp. 1032 - 1033.
*/

static const int FUDGE = 16;

int
ar_clock_recovery_mm_cc::general_work (int noutput_items,
				       gr_vector_int &ninput_items,
				       gr_vector_const_void_star &input_items,
				       gr_vector_void_star &output_items)
{
  const gr_complex *in   = (const gr_complex *) input_items[0];
  gr_complex *out        = (gr_complex *) output_items[0];
  gr_complex *foptr      = (gr_complex *) output_items[1];

  bool write_foptr       = output_items.size() >= 2;
  
  int  ii = 0;				// input index
  int  oo = 0;				// output index
  int  ni = ninput_items[0] - d_interp->ntaps() - FUDGE;  // don't use more input than this

  assert(d_mu >= 0.0);
  assert(d_mu <= 1.0);

  float mm_val=0;
  gr_complex u, x, y;

  // This loop writes the error to the second output, if it exists
     if (write_foptr) {
          while(oo < noutput_items && ii < ni) {               
               // Update Registers: DATA
               d_p_2T    = d_p_1T;
               d_p_1T    = d_p_0T;
               d_p_0T    = d_interp->interpolate (&in[ii], d_mu);
               // Update Registers: Decision
               d_c_2T    = d_c_1T;
               d_c_1T    = d_c_0T; 
               d_c_0T    = slicer_0deg(d_p_0T);
           
               x         = (d_c_0T - d_c_2T) * conj(d_p_1T);
               y         = (d_p_0T - d_p_2T) * conj(d_c_1T);
               u         = y - x;
               mm_val    = u.real();
               out[oo++] = d_p_0T;
           
               // limit mm_val
               /*
               if (mm_val > 1.0)
	               mm_val = 1.0;
               else if (mm_val < -1.0)
	               mm_val = -1.0;
               else if (mm_val == std::numeric_limits<float>::quiet_NaN())
                    mm_val = 0.0;
               */      
               d_omega = d_omega + d_gain_omega * mm_val;
               if (d_omega > d_max_omega)
	               d_omega = d_max_omega;
               else if (d_omega < d_min_omega)
	               d_omega = d_min_omega;
           
               d_mu      = d_mu + d_omega + d_gain_mu * mm_val;
               ii        += (int)floor(d_mu);
               d_mu      -= floor(d_mu);
           
               #if 0
                    printf("%f\t%f\n", d_omega, d_mu);
               #endif
           
               // write the error signal to the second output
               //   foptr[oo-1] = gr_complex(d_mu,0);
               //EV
               foptr[oo-1] = gr_complex(mm_val,d_mu);

               if (ii < 0)	// clamp it.  This should only happen with bogus input
	               ii = 0;
               }
          }// end while
     // This loop does not write to the second output (ugly, but faster)
     else {
          while(oo < noutput_items && ii < ni) {
               d_p_2T = d_p_1T;
               d_p_1T = d_p_0T;
               d_p_0T = d_interp->interpolate (&in[ii], d_mu);

               d_c_2T = d_c_1T;
               d_c_1T = d_c_0T;
               d_c_0T = slicer_0deg(d_p_0T);
           
               x = (d_c_0T - d_c_2T) * conj(d_p_1T);
               y = (d_p_0T - d_p_2T) * conj(d_c_1T);
               u = y - x;
               mm_val = u.real();
               out[oo++] = d_p_0T;
           
               // limit mm_val
               /*
               if (mm_val > 1.0)
	               mm_val = 1.0;
               else if (mm_val < -1.0)
	               mm_val = -1.0;
               */      
               if (mm_val == std::numeric_limits<float>::quiet_NaN()) mm_val = 0.0;

               d_omega = d_omega + d_gain_omega * mm_val;
               if (d_omega > d_max_omega)
	               d_omega = d_max_omega;
               else if (d_omega < d_min_omega)
	               d_omega = d_min_omega;
           
               d_mu = d_mu + d_omega + d_gain_mu * mm_val;
               ii += (int)floor(d_mu);
               d_mu -= floor(d_mu);
           
               if(d_verbose) {
	               printf("%f\t%f\n", d_omega, d_mu);
               }
                 
               //timing_debug << d_omega << " " << d_mu << " " << mm_val << "\n";
               if (ii < 0)	// clamp it.  This should only happen with bogus input
	          ii = 0;
          }
          }// end while

          if (ii > 0){
               if (ii > ninput_items[0]){
                    fprintf(stderr, "ar_clock_recovery_mm_cc: ii > ninput_items[0] (%d > %d)\n",
	                    ii, ninput_items[0]);
               assert(0);
          }
          consume_each (ii);
    }// end else

  return oo;
}
