/* -*- c++ -*- */
/*
 * Copyright 2004,206 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifndef INCLUDED_GR_PLL_CARRIERTRACKING_CC_H
#define INCLUDED_GR_PLL_CARRIERTRACKING_CC_H

#include <gr_sync_block.h>

class gr_pll_carriertracking_cc;
typedef boost::shared_ptr<gr_pll_carriertracking_cc> gr_pll_carriertracking_cc_sptr;

gr_pll_carriertracking_cc_sptr gr_make_pll_carriertracking_cc (float alpha, float beta,
					       float max_freq, float min_freq);

#define GR_PLL_ERR_HISTORY_LEN (1<<13)

/*!
 * \brief Implements a PLL which locks to the input frequency and outputs the 
 * input signal mixed with that carrier.
 * \ingroup block
 *
 * input: stream of complex; output: stream of complex
 *
 * This PLL locks onto a [possibly noisy] reference carrier on
 * the input and outputs that signal, downconverted to DC
 * 
 * All settings max_freq and min_freq are in terms of radians per sample, 
 * NOT HERTZ.  Alpha is the phase gain (first order, units of radians per radian) 
 * and beta is the frequency gain (second order, units of radians per sample per radian)
 * \sa gr_pll_freqdet_cf, gr_pll_carriertracking_cc
 * 
 * The PLL lock detector is defined by abs(mean(imag) / mean(real)), where the
 * mean is computed over GR_PLL_ERR_HISTORY_LEN samples.  The lock detector is
 * updated once every GR_PLL_ERR_HISTORY_LEN samples.  The lock threshold should
 * be around 0.1, indicating that the real part is at least 10x larger than the
 * imaginary part.
 */
class gr_pll_carriertracking_cc : public gr_sync_block
{
  friend gr_pll_carriertracking_cc_sptr gr_make_pll_carriertracking_cc (float alpha, float beta,
							float max_freq, float min_freq);

  float d_alpha,d_beta,d_max_freq,d_min_freq,d_phase,d_freq,d_locksig,d_lock_threshold;
  bool d_squelch_enable;
  float d_lock_re, d_lock_im;
  float d_offset_est;
  gr_complex d_errors[GR_PLL_ERR_HISTORY_LEN];
  int   d_errors_count;

  gr_pll_carriertracking_cc (float alpha, float beta, float max_freq, float min_freq);

  int work (int noutput_items,
	    gr_vector_const_void_star &input_items,
	    gr_vector_void_star &output_items);
private:
  float mod_2pi (float in);
  float phase_detector(gr_complex sample,float ref_phase);
public:
  bool lock_detector(void);
  bool  squelch_enable(bool);
  //! Set the lock threshold.  The lock estimator is given by:
  //! (Abs[Mean[Re[x]]]+Abs[Mean[Im[x]]])/(Mean[Abs[Re[x]]]+Mean[Abs[Im[x]]])
  //! The system is locked when the detector value is closer to 1.0, unlocked
  //! when close to 0.0
  float set_lock_threshold(float);
  //! Get the estimated frequency offset
  float get_freq_offset() {return d_offset_est;}
  //! Set the coefficients for the second order loop.  Loop gain is 1.0
  //! Set alpha and beta to zero to reset and disable the loop
  void set_loop_coefs(float alpha, float beta)
  {
    d_alpha = alpha; d_beta = beta;
    if (d_alpha == 0 && d_beta == 0) {d_phase = 0; d_freq = 0;}
  }
  //! Set the initial frequency offset.  Typically from the acquisition block
  //! /param freq Frequency offset in radians/sample
  void set_freq_offset(float freq)
  {
      if (freq > d_max_freq || freq < d_min_freq) 
          printf("gr_pll_carriertracking: freq outside min/max range\n");
      d_freq = freq;
  }
};

#endif
