#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <ar_sig_source_c.h>
#include <algorithm>
#include <gr_io_signature.h>
#include <stdexcept>
#include <gr_complex.h>
#include <ipp.h>


gr_sig_source_c::gr_sig_source_c (double sampling_freq, gr_waveform_t waveform,
		double frequency, double ampl, gr_complex offset)
  : gr_sync_block ("sig_source_c",
		   gr_make_io_signature (0, 0, 0),
		   gr_make_io_signature (1, 1, sizeof (gr_complex))),
    d_sampling_freq (sampling_freq), d_waveform (waveform), d_frequency (frequency),
    d_ampl (ampl), d_offset (offset)
{
    set_frequency(d_frequency);
}

gr_sig_source_c_sptr
gr_make_sig_source_c (double sampling_freq, gr_waveform_t waveform,
		     double frequency, double ampl, gr_complex offset)
{
  return gr_sig_source_c_sptr (new gr_sig_source_c (sampling_freq, waveform, frequency, ampl, offset));
}

int
gr_sig_source_c::work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items)
{
  gr_complex *optr = (gr_complex *) output_items[0];
  gr_complex t;

  switch (d_waveform){

  case GR_CONST_WAVE:
    t = (gr_complex) d_ampl + d_offset;
    ippsSet_32fc(*((Ipp32fc*)&t), (Ipp32fc*) optr, noutput_items);
    break;
    
  case GR_SIN_WAVE:
  case GR_COS_WAVE:
    // The ippsTone function uses freq [0 .. 1], with freq [0.5 .. 1] negative
    ippsTone_Direct_32fc((Ipp32fc*) optr, noutput_items, d_ampl, d_freq, 
                         &d_phase, ippAlgHintFast);

    if (d_offset == gr_complex(0,0))
      break;

    ippsAddC_32fc_I(*((Ipp32fc*)&d_offset), (Ipp32fc*) optr, noutput_items);
    break;
    
  /* Implements a real square wave high from -PI to 0.  
  * The imaginary square wave leads by 90 deg. Implemented as a triangle
  * wave followed by clipping and scaling
  */
  case GR_SQR_WAVE:	
    // The ipps function uses freq [0 .. 1], with freq [0.5 .. 1] negative
    ippsTriangle_Direct_32fc((Ipp32fc*) optr, noutput_items, d_ampl, d_freq, 
                             0.0, &d_phase);

    // The complex version of this function operates on the magnitude of the
    // samples, which is NOT what we want.  Used the real version, with
    // twice as many samples.
    ippsThreshold_LTValGTVal_32f_I((Ipp32f*) optr, 2*noutput_items, 0.0,
                                   0.0, 0.0, d_ampl);    

    if (d_offset == gr_complex(0,0))
      break;

    ippsAddC_32fc_I(*((Ipp32fc*)&d_offset), (Ipp32fc*) optr, noutput_items);
    break;
	
  /* Implements a real triangle wave rising from -PI to 0 and  
  * falling from 0 to PI. The imaginary triangle wave leads by 90 deg.
  */
  case GR_TRI_WAVE:	
    // The ipps function uses freq [0 .. 1], with freq [0.5 .. 1] negative
    ippsTriangle_Direct_32fc((Ipp32fc*) optr, noutput_items, d_ampl, d_freq, 
                             0.0, &d_phase);

    if (d_offset == gr_complex(0,0))
      break;

    ippsAddC_32fc_I(*((Ipp32fc*)&d_offset), (Ipp32fc*) optr, noutput_items);
    break;
	
  /* Implements a real saw tooth wave rising from -PI to PI. 
  * The imaginary saw tooth wave leads by 90 deg.  Implemented as a triangle
  * wave with an asymmetry parameter of -pi
  */
  case GR_SAW_WAVE:	
    // The ipps function uses freq [0 .. 1], with freq [0.5 .. 1] negative
    ippsTriangle_Direct_32fc((Ipp32fc*) optr, noutput_items, d_ampl, d_freq, 
                             -M_PI, &d_phase);

    if (d_offset == gr_complex(0,0))
      break;

    ippsAddC_32fc_I(*((Ipp32fc*)&d_offset), (Ipp32fc*) optr, noutput_items);
    break;

  default:
    throw std::runtime_error ("gr_sig_source: invalid waveform");
  }

  return noutput_items;
}

void
gr_sig_source_c::set_sampling_freq (double sampling_freq)
{
  d_sampling_freq = sampling_freq;
  set_frequency(d_frequency);
}

void
gr_sig_source_c::set_waveform (gr_waveform_t waveform)
{
  d_waveform = waveform;
}

void
gr_sig_source_c::set_frequency (double frequency)
{
  d_frequency = frequency;
  d_freq = d_frequency / d_sampling_freq / 2;
  if (d_frequency < 0) d_freq += 1.0;
}

void
gr_sig_source_c::set_amplitude (double ampl)
{
  d_ampl = ampl;
}

void
gr_sig_source_c::set_offset (gr_complex offset)
{
  d_offset = offset;
}

