#ifndef INCLUDED_GR_FREQ_ACQ_H
#define INCLUDED_GR_FREQ_ACQ_H

#include <gr_sync_block.h>
#include "mld_threads.h"

using namespace std;

class gr_freq_acq;
typedef boost::shared_ptr<gr_freq_acq> gr_freq_acq_sptr;

//! Frequency acquisition block.  This block examines an FFT input vector
//! and finds the peaks.  The input is typically:
//! gr_fft_vcc -> gr_complex_to_mag_squared -> gr_integrate_dump_ff
class gr_freq_acq : public gr_sync_block
{
  friend gr_freq_acq_sptr gr_make_freq_acq (size_t fft_size);

  gr_freq_acq (size_t sizeof_stream_item);
  
  int d_fft_size;
  vector<float> d_fft_result;
  bool d_result_ready;

  mld_mutex d_mutex;
 public:

  virtual int work (int noutput_items,
		    gr_vector_const_void_star &input_items,
		    gr_vector_void_star &output_items);

  //! Get the frequencies of the N highest peaks. Returns false if 
  //! the acquisition engine does not have an estimate ready yet.
  //! \param num_peaks Number of largest bins to return
  //! \param freqs Bin numbers for the largest peaks.
  //! \param powers Power in the bins relative to the average power.  The
  //!               average power does not include the num_peaks largest
  bool get_freqs(int num_peaks, vector<float> &freqs, vector<float> &powers);
  //! Get the center frequency of the peak (interpolated) Returns  
  //! false if the acquisition engine does not have an estimate ready yet.
  //! Computes a center of mass to get a precise frequency estimate
  //! \param freq Returns the estimated bin # (fractional)
  //! \param power Returns the power relative to the average
  bool get_peak_freq(float *freq, float *power);
  //! Get the FFT block size set in the constructor
  int get_fft_size() const {return d_fft_size;}
  //! Advanced acquisition based on a frequency domain mask/template.
  //! Peak detection is unreliable if the spectrum contains multiple
  //! features.  However, if the relative feature positions are known,
  //! this can be exploited by computing a crosscorellation between
  //! the known feature mask and the FFT output.  This function computed
  //! the crosscorrelation and returns the index of the peak.  The mask
  //! must be the same size as the FFT.
  //! \param mask Template/mask with the FFT features.  Typically 1.0 for bins
  //!             where power is expected, and 0 otherwise
  bool get_crosscorr_freq(const vector<float> &mask, int *max_idx,
            float *max_power, float *mean_power, vector<float> &fft);
};

//! Create the frequency acquisition class
//! \param fft_size Number of FFT bins (should be a power of 2)
gr_freq_acq_sptr
    gr_make_freq_acq (size_t fft_size);

#endif /* INCLUDED_GR_freq_acq_H */

